//-------------------------------------------------------------------------
/*
Copyright (C) 2010 EDuke32 developers and contributors

This file is part of EDuke32.

EDuke32 is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License version 2
as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
//-------------------------------------------------------------------------
#include "ns.h"	// Must come before everything else!

#include "compat.h"
#include "build.h"

#include "namesdyn.h"
#include "global.h"
#include "gamecontrol.h"

BEGIN_DUKE_NS

#define DVPTR(x) &x

int16_t DynamicTileMap[MAXTILES];

struct dynitem
{
    const char *str;
    int32_t *dynvalptr;
    const int16_t staticval;
    const int16_t staticval_rr;
};

static struct dynitem g_dynTileList[] =
{
    { "SECTOREFFECTOR", DVPTR(TILE_SECTOREFFECTOR), SECTOREFFECTOR__STATIC, SECTOREFFECTOR__STATICRR },
    { "ACTIVATOR", DVPTR(TILE_ACTIVATOR), ACTIVATOR__STATIC, ACTIVATOR__STATICRR },
    { "TOUCHPLATE", DVPTR(TILE_TOUCHPLATE), TOUCHPLATE__STATIC, TOUCHPLATE__STATICRR },
    { "ACTIVATORLOCKED", DVPTR(TILE_ACTIVATORLOCKED), ACTIVATORLOCKED__STATIC, ACTIVATORLOCKED__STATICRR },
    { "MUSICANDSFX", DVPTR(TILE_MUSICANDSFX), MUSICANDSFX__STATIC, MUSICANDSFX__STATICRR },
    { "LOCATORS", DVPTR(TILE_LOCATORS), LOCATORS__STATIC, LOCATORS__STATICRR },
    { "CYCLER", DVPTR(TILE_CYCLER), CYCLER__STATIC, CYCLER__STATICRR },
    { "MASTERSWITCH", DVPTR(TILE_MASTERSWITCH), MASTERSWITCH__STATIC, MASTERSWITCH__STATICRR },
    { "RESPAWN", DVPTR(TILE_RESPAWN), RESPAWN__STATIC, RESPAWN__STATICRR },
    { "GPSPEED", DVPTR(TILE_GPSPEED), GPSPEED__STATIC, GPSPEED__STATICRR },
    { "FOF", DVPTR(TILE_FOF), FOF__STATIC, FOF__STATICRR },
    { "ARROW", DVPTR(TILE_ARROW), ARROW__STATIC, ARROW__STATICRR },
    { "FIRSTGUNSPRITE", DVPTR(TILE_FIRSTGUNSPRITE), FIRSTGUNSPRITE__STATIC, FIRSTGUNSPRITE__STATICRR },
    { "CHAINGUNSPRITE", DVPTR(TILE_CHAINGUNSPRITE), CHAINGUNSPRITE__STATIC, CHAINGUNSPRITE__STATICRR },
    { "RPGSPRITE", DVPTR(TILE_RPGSPRITE), RPGSPRITE__STATIC, RPGSPRITE__STATICRR },
    { "FREEZESPRITE", DVPTR(TILE_FREEZESPRITE), FREEZESPRITE__STATIC, FREEZESPRITE__STATICRR },
    { "SHRINKERSPRITE", DVPTR(TILE_SHRINKERSPRITE), SHRINKERSPRITE__STATIC, SHRINKERSPRITE__STATICRR },
    { "HEAVYHBOMB", DVPTR(TILE_HEAVYHBOMB), HEAVYHBOMB__STATIC, HEAVYHBOMB__STATICRR },
    { "TRIPBOMBSPRITE", DVPTR(TILE_TRIPBOMBSPRITE), TRIPBOMBSPRITE__STATIC, TRIPBOMBSPRITE__STATICRR },
    { "SHOTGUNSPRITE", DVPTR(TILE_SHOTGUNSPRITE), SHOTGUNSPRITE__STATIC, SHOTGUNSPRITE__STATICRR },
    { "DEVISTATORSPRITE", DVPTR(TILE_DEVISTATORSPRITE), DEVISTATORSPRITE__STATIC, DEVISTATORSPRITE__STATICRR },
    { "HEALTHBOX", DVPTR(TILE_HEALTHBOX), HEALTHBOX__STATIC, HEALTHBOX__STATICRR },
    { "AMMOBOX", DVPTR(TILE_AMMOBOX), AMMOBOX__STATIC, AMMOBOX__STATICRR },
    { "GROWSPRITEICON", DVPTR(TILE_GROWSPRITEICON), GROWSPRITEICON__STATIC, GROWSPRITEICON__STATICRR },
    { "INVENTORYBOX", DVPTR(TILE_INVENTORYBOX), INVENTORYBOX__STATIC, INVENTORYBOX__STATICRR },
    { "FREEZEAMMO", DVPTR(TILE_FREEZEAMMO), FREEZEAMMO__STATIC, FREEZEAMMO__STATICRR },
    { "AMMO", DVPTR(TILE_AMMO), AMMO__STATIC, AMMO__STATICRR },
    { "BATTERYAMMO", DVPTR(TILE_BATTERYAMMO), BATTERYAMMO__STATIC, BATTERYAMMO__STATICRR },
    { "DEVISTATORAMMO", DVPTR(TILE_DEVISTATORAMMO), DEVISTATORAMMO__STATIC, DEVISTATORAMMO__STATICRR },
    { "RPGAMMO", DVPTR(TILE_RPGAMMO), RPGAMMO__STATIC, RPGAMMO__STATICRR },
    { "GROWAMMO", DVPTR(TILE_GROWAMMO), GROWAMMO__STATIC, GROWAMMO__STATICRR },
    { "CRYSTALAMMO", DVPTR(TILE_CRYSTALAMMO), CRYSTALAMMO__STATIC, CRYSTALAMMO__STATICRR },
    { "HBOMBAMMO", DVPTR(TILE_HBOMBAMMO), HBOMBAMMO__STATIC, HBOMBAMMO__STATICRR },
    { "AMMOLOTS", DVPTR(TILE_AMMOLOTS), AMMOLOTS__STATIC, AMMOLOTS__STATICRR },
    { "SHOTGUNAMMO", DVPTR(TILE_SHOTGUNAMMO), SHOTGUNAMMO__STATIC, SHOTGUNAMMO__STATICRR },
    { "COLA", DVPTR(TILE_COLA), COLA__STATIC, COLA__STATICRR },
    { "SIXPAK", DVPTR(TILE_SIXPAK), SIXPAK__STATIC, SIXPAK__STATICRR },
    { "FIRSTAID", DVPTR(TILE_FIRSTAID), FIRSTAID__STATIC, FIRSTAID__STATICRR },
    { "SHIELD", DVPTR(TILE_SHIELD), SHIELD__STATIC, SHIELD__STATICRR },
    { "STEROIDS", DVPTR(TILE_STEROIDS), STEROIDS__STATIC, STEROIDS__STATICRR },
    { "AIRTANK", DVPTR(TILE_AIRTANK), AIRTANK__STATIC, AIRTANK__STATICRR },
    { "JETPACK", DVPTR(TILE_JETPACK), JETPACK__STATIC, JETPACK__STATICRR },
    { "HEATSENSOR", DVPTR(TILE_HEATSENSOR), HEATSENSOR__STATIC, HEATSENSOR__STATICRR },
    { "ACCESSCARD", DVPTR(TILE_ACCESSCARD), ACCESSCARD__STATIC, ACCESSCARD__STATICRR },
    { "BOOTS", DVPTR(TILE_BOOTS), BOOTS__STATIC, BOOTS__STATICRR },
    { "MIRRORBROKE", DVPTR(TILE_MIRRORBROKE), MIRRORBROKE__STATIC, MIRRORBROKE__STATICRR },
    { "CLOUDYOCEAN", DVPTR(TILE_CLOUDYOCEAN), CLOUDYOCEAN__STATIC, 0 },
    { "CLOUDYSKIES", DVPTR(TILE_CLOUDYSKIES), CLOUDYSKIES__STATIC, CLOUDYSKIES__STATICRR },
    { "MOONSKY1", DVPTR(TILE_MOONSKY1), MOONSKY1__STATIC, MOONSKY1__STATICRR },
    { "MOONSKY2", DVPTR(TILE_MOONSKY2), MOONSKY2__STATIC, MOONSKY2__STATICRR },
    { "MOONSKY3", DVPTR(TILE_MOONSKY3), MOONSKY3__STATIC, MOONSKY3__STATICRR },
    { "MOONSKY4", DVPTR(TILE_MOONSKY4), MOONSKY4__STATIC, MOONSKY4__STATICRR },
    { "BIGORBIT1", DVPTR(TILE_BIGORBIT1), BIGORBIT1__STATIC, BIGORBIT1__STATICRR },
    { "BIGORBIT2", DVPTR(TILE_BIGORBIT2), BIGORBIT2__STATIC, BIGORBIT2__STATICRR },
    { "BIGORBIT3", DVPTR(TILE_BIGORBIT3), BIGORBIT3__STATIC, BIGORBIT3__STATICRR },
    { "BIGORBIT4", DVPTR(TILE_BIGORBIT4), BIGORBIT4__STATIC, BIGORBIT4__STATICRR },
    { "BIGORBIT5", DVPTR(TILE_BIGORBIT5), BIGORBIT5__STATIC, BIGORBIT5__STATICRR },
    { "LA", DVPTR(TILE_LA), LA__STATIC, LA__STATICRR },
    { "REDSKY1", DVPTR(TILE_REDSKY1), REDSKY1__STATIC, REDSKY1__STATICRR },
    { "REDSKY2", DVPTR(TILE_REDSKY2), REDSKY2__STATIC, REDSKY2__STATICRR },
    { "ATOMICHEALTH", DVPTR(TILE_ATOMICHEALTH), ATOMICHEALTH__STATIC, ATOMICHEALTH__STATICRR },
    { "TECHLIGHT2", DVPTR(TILE_TECHLIGHT2), TECHLIGHT2__STATIC, TECHLIGHT2__STATICRR },
    { "TECHLIGHTBUST2", DVPTR(TILE_TECHLIGHTBUST2), TECHLIGHTBUST2__STATIC, TECHLIGHTBUST2__STATICRR },
    { "TECHLIGHT4", DVPTR(TILE_TECHLIGHT4), TECHLIGHT4__STATIC, TECHLIGHT4__STATICRR },
    { "TECHLIGHTBUST4", DVPTR(TILE_TECHLIGHTBUST4), TECHLIGHTBUST4__STATIC, TECHLIGHTBUST4__STATICRR },
    { "WALLLIGHT4", DVPTR(TILE_WALLLIGHT4), WALLLIGHT4__STATIC, WALLLIGHT4__STATICRR },
    { "WALLLIGHTBUST4", DVPTR(TILE_WALLLIGHTBUST4), WALLLIGHTBUST4__STATIC, WALLLIGHTBUST4__STATICRR },
    { "ACCESSSWITCH", DVPTR(TILE_ACCESSSWITCH), ACCESSSWITCH__STATIC, ACCESSSWITCH__STATICRR },
    { "SLOTDOOR", DVPTR(TILE_SLOTDOOR), SLOTDOOR__STATIC, SLOTDOOR__STATICRR },
    { "LIGHTSWITCH", DVPTR(TILE_LIGHTSWITCH), LIGHTSWITCH__STATIC, LIGHTSWITCH__STATICRR },
    { "SPACEDOORSWITCH", DVPTR(TILE_SPACEDOORSWITCH), SPACEDOORSWITCH__STATIC, SPACEDOORSWITCH__STATICRR },
    { "SPACELIGHTSWITCH", DVPTR(TILE_SPACELIGHTSWITCH), SPACELIGHTSWITCH__STATIC, SPACELIGHTSWITCH__STATICRR },
    { "FRANKENSTINESWITCH", DVPTR(TILE_FRANKENSTINESWITCH), FRANKENSTINESWITCH__STATIC, FRANKENSTINESWITCH__STATICRR },
    { "NUKEBUTTON", DVPTR(TILE_NUKEBUTTON), NUKEBUTTON__STATIC, NUKEBUTTON__STATICRR },
    { "MULTISWITCH", DVPTR(TILE_MULTISWITCH), MULTISWITCH__STATIC, MULTISWITCH__STATICRR },
    { "DOORTILE5", DVPTR(TILE_DOORTILE5), DOORTILE5__STATIC, DOORTILE5__STATICRR },
    { "DOORTILE6", DVPTR(TILE_DOORTILE6), DOORTILE6__STATIC, DOORTILE6__STATICRR },
    { "DOORTILE1", DVPTR(TILE_DOORTILE1), DOORTILE1__STATIC, DOORTILE1__STATICRR },
    { "DOORTILE2", DVPTR(TILE_DOORTILE2), DOORTILE2__STATIC, DOORTILE2__STATICRR },
    { "DOORTILE3", DVPTR(TILE_DOORTILE3), DOORTILE3__STATIC, DOORTILE3__STATICRR },
    { "DOORTILE4", DVPTR(TILE_DOORTILE4), DOORTILE4__STATIC, DOORTILE4__STATICRR },
    { "DOORTILE7", DVPTR(TILE_DOORTILE7), DOORTILE7__STATIC, DOORTILE7__STATICRR },
    { "DOORTILE8", DVPTR(TILE_DOORTILE8), DOORTILE8__STATIC, DOORTILE8__STATICRR },
    { "DOORTILE9", DVPTR(TILE_DOORTILE9), DOORTILE9__STATIC, DOORTILE9__STATICRR },
    { "DOORTILE10", DVPTR(TILE_DOORTILE10), DOORTILE10__STATIC, DOORTILE10__STATICRR },
    { "DOORSHOCK", DVPTR(TILE_DOORSHOCK), DOORSHOCK__STATIC, DOORSHOCK__STATICRR },
    { "DIPSWITCH", DVPTR(TILE_DIPSWITCH), DIPSWITCH__STATIC, DIPSWITCH__STATICRR },
    { "DIPSWITCH2", DVPTR(TILE_DIPSWITCH2), DIPSWITCH2__STATIC, DIPSWITCH2__STATICRR },
    { "TECHSWITCH", DVPTR(TILE_TECHSWITCH), TECHSWITCH__STATIC, TECHSWITCH__STATICRR },
    { "DIPSWITCH3", DVPTR(TILE_DIPSWITCH3), DIPSWITCH3__STATIC, DIPSWITCH3__STATICRR },
    { "ACCESSSWITCH2", DVPTR(TILE_ACCESSSWITCH2), ACCESSSWITCH2__STATIC, ACCESSSWITCH2__STATICRR },
    { "REFLECTWATERTILE", DVPTR(TILE_REFLECTWATERTILE), REFLECTWATERTILE__STATIC, REFLECTWATERTILE__STATICRR },
    { "FLOORSLIME", DVPTR(TILE_FLOORSLIME), FLOORSLIME__STATIC, FLOORSLIME__STATICRR },
    { "BIGFORCE", DVPTR(TILE_BIGFORCE), BIGFORCE__STATIC, BIGFORCE__STATICRR },
    { "EPISODE", DVPTR(TILE_EPISODE), EPISODE__STATIC, EPISODE__STATICRR },
    { "MASKWALL9", DVPTR(TILE_MASKWALL9), MASKWALL9__STATIC, MASKWALL9__STATICRR },
    { "W_LIGHT", DVPTR(TILE_W_LIGHT), W_LIGHT__STATIC, W_LIGHT__STATICRR },
    { "SCREENBREAK1", DVPTR(TILE_SCREENBREAK1), SCREENBREAK1__STATIC, SCREENBREAK1__STATICRR },
    { "SCREENBREAK2", DVPTR(TILE_SCREENBREAK2), SCREENBREAK2__STATIC, SCREENBREAK2__STATICRR },
    { "SCREENBREAK3", DVPTR(TILE_SCREENBREAK3), SCREENBREAK3__STATIC, SCREENBREAK3__STATICRR },
    { "SCREENBREAK4", DVPTR(TILE_SCREENBREAK4), SCREENBREAK4__STATIC, SCREENBREAK4__STATICRR },
    { "SCREENBREAK5", DVPTR(TILE_SCREENBREAK5), SCREENBREAK5__STATIC, SCREENBREAK5__STATICRR },
    { "SCREENBREAK6", DVPTR(TILE_SCREENBREAK6), SCREENBREAK6__STATIC, SCREENBREAK6__STATICRR },
    { "SCREENBREAK7", DVPTR(TILE_SCREENBREAK7), SCREENBREAK7__STATIC, SCREENBREAK7__STATICRR },
    { "SCREENBREAK8", DVPTR(TILE_SCREENBREAK8), SCREENBREAK8__STATIC, SCREENBREAK8__STATICRR },
    { "SCREENBREAK9", DVPTR(TILE_SCREENBREAK9), SCREENBREAK9__STATIC, SCREENBREAK9__STATICRR },
    { "SCREENBREAK10", DVPTR(TILE_SCREENBREAK10), SCREENBREAK10__STATIC, SCREENBREAK10__STATICRR },
    { "SCREENBREAK11", DVPTR(TILE_SCREENBREAK11), SCREENBREAK11__STATIC, SCREENBREAK11__STATICRR },
    { "SCREENBREAK12", DVPTR(TILE_SCREENBREAK12), SCREENBREAK12__STATIC, SCREENBREAK12__STATICRR },
    { "SCREENBREAK13", DVPTR(TILE_SCREENBREAK13), SCREENBREAK13__STATIC, SCREENBREAK13__STATICRR },
    { "MASKWALL1", DVPTR(TILE_MASKWALL1), MASKWALL1__STATIC, MASKWALL1__STATICRR },
    { "W_TECHWALL1", DVPTR(TILE_W_TECHWALL1), W_TECHWALL1__STATIC, W_TECHWALL1__STATICRR },
    { "W_TECHWALL2", DVPTR(TILE_W_TECHWALL2), W_TECHWALL2__STATIC, W_TECHWALL2__STATICRR },
    { "W_TECHWALL15", DVPTR(TILE_W_TECHWALL15), W_TECHWALL15__STATIC, W_TECHWALL15__STATICRR },
    { "W_TECHWALL3", DVPTR(TILE_W_TECHWALL3), W_TECHWALL3__STATIC, W_TECHWALL3__STATICRR },
    { "W_TECHWALL4", DVPTR(TILE_W_TECHWALL4), W_TECHWALL4__STATIC, W_TECHWALL4__STATICRR },
    { "W_TECHWALL10", DVPTR(TILE_W_TECHWALL10), W_TECHWALL10__STATIC, W_TECHWALL10__STATICRR },
    { "W_TECHWALL16", DVPTR(TILE_W_TECHWALL16), W_TECHWALL16__STATIC, W_TECHWALL16__STATICRR },
    { "WATERTILE2", DVPTR(TILE_WATERTILE2), WATERTILE2__STATIC, WATERTILE2__STATICRR },
    { "BPANNEL1", DVPTR(TILE_BPANNEL1), BPANNEL1__STATIC, BPANNEL1__STATICRR },
    { "PANNEL1", DVPTR(TILE_PANNEL1), PANNEL1__STATIC, PANNEL1__STATICRR },
    { "PANNEL2", DVPTR(TILE_PANNEL2), PANNEL2__STATIC, PANNEL2__STATICRR },
    { "WATERTILE", DVPTR(TILE_WATERTILE), WATERTILE__STATIC, WATERTILE__STATICRR },
    { "STATIC", DVPTR(TILE_STATIC), STATIC__STATIC, STATIC__STATICRR },
    { "W_SCREENBREAK", DVPTR(TILE_W_SCREENBREAK), W_SCREENBREAK__STATIC, W_SCREENBREAK__STATICRR },
    { "W_HITTECHWALL3", DVPTR(TILE_W_HITTECHWALL3), W_HITTECHWALL3__STATIC, W_HITTECHWALL3__STATICRR },
    { "W_HITTECHWALL4", DVPTR(TILE_W_HITTECHWALL4), W_HITTECHWALL4__STATIC, W_HITTECHWALL4__STATICRR },
    { "W_HITTECHWALL2", DVPTR(TILE_W_HITTECHWALL2), W_HITTECHWALL2__STATIC, W_HITTECHWALL2__STATICRR },
    { "W_HITTECHWALL1", DVPTR(TILE_W_HITTECHWALL1), W_HITTECHWALL1__STATIC, W_HITTECHWALL1__STATICRR },
    { "MASKWALL10", DVPTR(TILE_MASKWALL10), MASKWALL10__STATIC, MASKWALL10__STATICRR },
    { "MASKWALL11", DVPTR(TILE_MASKWALL11), MASKWALL11__STATIC, MASKWALL11__STATICRR },
    { "DOORTILE22", DVPTR(TILE_DOORTILE22), DOORTILE22__STATIC, DOORTILE22__STATICRR },
    { "FANSPRITE", DVPTR(TILE_FANSPRITE), FANSPRITE__STATIC, FANSPRITE__STATICRR },
    { "FANSPRITEBROKE", DVPTR(TILE_FANSPRITEBROKE), FANSPRITEBROKE__STATIC, FANSPRITEBROKE__STATICRR },
    { "FANSHADOW", DVPTR(TILE_FANSHADOW), FANSHADOW__STATIC, FANSHADOW__STATICRR },
    { "FANSHADOWBROKE", DVPTR(TILE_FANSHADOWBROKE), FANSHADOWBROKE__STATIC, FANSHADOWBROKE__STATICRR },
    { "DOORTILE18", DVPTR(TILE_DOORTILE18), DOORTILE18__STATIC, DOORTILE18__STATICRR },
    { "DOORTILE19", DVPTR(TILE_DOORTILE19), DOORTILE19__STATIC, DOORTILE19__STATICRR },
    { "DOORTILE20", DVPTR(TILE_DOORTILE20), DOORTILE20__STATIC, DOORTILE20__STATICRR },
    { "SATELLITE", DVPTR(TILE_SATELLITE), SATELLITE__STATIC, SATELLITE__STATICRR },
    { "VIEWSCREEN2", DVPTR(TILE_VIEWSCREEN2), VIEWSCREEN2__STATIC, VIEWSCREEN2__STATICRR },
    { "VIEWSCREENBROKE", DVPTR(TILE_VIEWSCREENBROKE), VIEWSCREENBROKE__STATIC, VIEWSCREENBROKE__STATICRR },
    { "VIEWSCREEN", DVPTR(TILE_VIEWSCREEN), VIEWSCREEN__STATIC, VIEWSCREEN__STATICRR },
    { "GLASS", DVPTR(TILE_GLASS), GLASS__STATIC, GLASS__STATICRR },
    { "GLASS2", DVPTR(TILE_GLASS2), GLASS2__STATIC, GLASS2__STATICRR },
    { "STAINGLASS1", DVPTR(TILE_STAINGLASS1), STAINGLASS1__STATIC, STAINGLASS1__STATICRR },
    { "MASKWALL5", DVPTR(TILE_MASKWALL5), MASKWALL5__STATIC, MASKWALL5__STATICRR },
    { "SATELITE", DVPTR(TILE_SATELITE), SATELITE__STATIC, SATELITE__STATICRR },
    { "FUELPOD", DVPTR(TILE_FUELPOD), FUELPOD__STATIC, FUELPOD__STATICRR },
    { "SLIMEPIPE", DVPTR(TILE_SLIMEPIPE), SLIMEPIPE__STATIC, SLIMEPIPE__STATICRR },
    { "CRACK1", DVPTR(TILE_CRACK1), CRACK1__STATIC, CRACK1__STATICRR },
    { "CRACK2", DVPTR(TILE_CRACK2), CRACK2__STATIC, CRACK2__STATICRR },
    { "CRACK3", DVPTR(TILE_CRACK3), CRACK3__STATIC, CRACK3__STATICRR },
    { "CRACK4", DVPTR(TILE_CRACK4), CRACK4__STATIC, CRACK4__STATICRR },
    { "FOOTPRINTS", DVPTR(TILE_FOOTPRINTS), FOOTPRINTS__STATIC, FOOTPRINTS__STATICRR },
    { "DOMELITE", DVPTR(TILE_DOMELITE), DOMELITE__STATIC, DOMELITE__STATICRR },
    { "CAMERAPOLE", DVPTR(TILE_CAMERAPOLE), CAMERAPOLE__STATIC, CAMERAPOLE__STATICRR },
    { "CHAIR1", DVPTR(TILE_CHAIR1), CHAIR1__STATIC, CHAIR1__STATICRR },
    { "CHAIR2", DVPTR(TILE_CHAIR2), CHAIR2__STATIC, CHAIR2__STATICRR },
    { "BROKENCHAIR", DVPTR(TILE_BROKENCHAIR), BROKENCHAIR__STATIC, BROKENCHAIR__STATICRR },
    { "MIRROR", DVPTR(TILE_MIRROR), MIRROR__STATIC, MIRROR__STATICRR },
    { "WATERFOUNTAIN", DVPTR(TILE_WATERFOUNTAIN), WATERFOUNTAIN__STATIC, WATERFOUNTAIN__STATICRR },
    { "WATERFOUNTAINBROKE", DVPTR(TILE_WATERFOUNTAINBROKE), WATERFOUNTAINBROKE__STATIC, WATERFOUNTAINBROKE__STATICRR },
    { "FEMMAG1", DVPTR(TILE_FEMMAG1), FEMMAG1__STATIC, FEMMAG1__STATICRR },
    { "TOILET", DVPTR(TILE_TOILET), TOILET__STATIC, TOILET__STATICRR },
    { "STALL", DVPTR(TILE_STALL), STALL__STATIC, STALL__STATICRR },
    { "STALLBROKE", DVPTR(TILE_STALLBROKE), STALLBROKE__STATIC, STALLBROKE__STATICRR },
    { "FEMMAG2", DVPTR(TILE_FEMMAG2), FEMMAG2__STATIC, FEMMAG2__STATICRR },
    { "REACTOR2", DVPTR(TILE_REACTOR2), REACTOR2__STATIC, REACTOR2__STATICRR },
    { "REACTOR2BURNT", DVPTR(TILE_REACTOR2BURNT), REACTOR2BURNT__STATIC, REACTOR2BURNT__STATICRR },
    { "REACTOR2SPARK", DVPTR(TILE_REACTOR2SPARK), REACTOR2SPARK__STATIC, REACTOR2SPARK__STATICRR },
    { "GRATE1", DVPTR(TILE_GRATE1), GRATE1__STATIC, GRATE1__STATICRR },
    { "BGRATE1", DVPTR(TILE_BGRATE1), BGRATE1__STATIC, BGRATE1__STATICRR },
    { "SOLARPANNEL", DVPTR(TILE_SOLARPANNEL), SOLARPANNEL__STATIC, SOLARPANNEL__STATICRR },
    { "NAKED1", DVPTR(TILE_NAKED1), NAKED1__STATIC, NAKED1__STATICRR },
    { "ANTENNA", DVPTR(TILE_ANTENNA), ANTENNA__STATIC, ANTENNA__STATICRR },
    { "MASKWALL12", DVPTR(TILE_MASKWALL12), MASKWALL12__STATIC, MASKWALL12__STATICRR },
    { "TOILETBROKE", DVPTR(TILE_TOILETBROKE), TOILETBROKE__STATIC, TOILETBROKE__STATICRR },
    { "PIPE2", DVPTR(TILE_PIPE2), PIPE2__STATIC, PIPE2__STATICRR },
    { "PIPE1B", DVPTR(TILE_PIPE1B), PIPE1B__STATIC, PIPE1B__STATICRR },
    { "PIPE3", DVPTR(TILE_PIPE3), PIPE3__STATIC, PIPE3__STATICRR },
    { "PIPE1", DVPTR(TILE_PIPE1), PIPE1__STATIC, PIPE1__STATICRR },
    { "CAMERA1", DVPTR(TILE_CAMERA1), CAMERA1__STATIC, CAMERA1__STATICRR },
    { "BRICK", DVPTR(TILE_BRICK), BRICK__STATIC, BRICK__STATICRR },
    { "SPLINTERWOOD", DVPTR(TILE_SPLINTERWOOD), SPLINTERWOOD__STATIC, SPLINTERWOOD__STATICRR },
    { "PIPE2B", DVPTR(TILE_PIPE2B), PIPE2B__STATIC, PIPE2B__STATICRR },
    { "BOLT1", DVPTR(TILE_BOLT1), BOLT1__STATIC, BOLT1__STATICRR },
    { "W_NUMBERS", DVPTR(TILE_W_NUMBERS), W_NUMBERS__STATIC, W_NUMBERS__STATICRR },
    { "WATERDRIP", DVPTR(TILE_WATERDRIP), WATERDRIP__STATIC, WATERDRIP__STATICRR },
    { "WATERBUBBLE", DVPTR(TILE_WATERBUBBLE), WATERBUBBLE__STATIC, WATERBUBBLE__STATICRR },
    { "WATERBUBBLEMAKER", DVPTR(TILE_WATERBUBBLEMAKER), WATERBUBBLEMAKER__STATIC, WATERBUBBLEMAKER__STATICRR },
    { "W_FORCEFIELD", DVPTR(TILE_W_FORCEFIELD), W_FORCEFIELD__STATIC, W_FORCEFIELD__STATICRR },
    { "VACUUM", DVPTR(TILE_VACUUM), VACUUM__STATIC, VACUUM__STATICRR },
    { "FOOTPRINTS2", DVPTR(TILE_FOOTPRINTS2), FOOTPRINTS2__STATIC, FOOTPRINTS2__STATICRR },
    { "FOOTPRINTS3", DVPTR(TILE_FOOTPRINTS3), FOOTPRINTS3__STATIC, FOOTPRINTS3__STATICRR },
    { "FOOTPRINTS4", DVPTR(TILE_FOOTPRINTS4), FOOTPRINTS4__STATIC, FOOTPRINTS4__STATICRR },
    { "EGG", DVPTR(TILE_EGG), EGG__STATIC, EGG__STATICRR },
    { "SCALE", DVPTR(TILE_SCALE), SCALE__STATIC, SCALE__STATICRR },
    { "CHAIR3", DVPTR(TILE_CHAIR3), CHAIR3__STATIC, CHAIR3__STATICRR },
    { "CAMERALIGHT", DVPTR(TILE_CAMERALIGHT), CAMERALIGHT__STATIC, CAMERALIGHT__STATICRR },
    { "MOVIECAMERA", DVPTR(TILE_MOVIECAMERA), MOVIECAMERA__STATIC, MOVIECAMERA__STATICRR },
    { "IVUNIT", DVPTR(TILE_IVUNIT), IVUNIT__STATIC, IVUNIT__STATICRR },
    { "POT1", DVPTR(TILE_POT1), POT1__STATIC, POT1__STATICRR },
    { "POT2", DVPTR(TILE_POT2), POT2__STATIC, POT2__STATICRR },
    { "POT3", DVPTR(TILE_POT3), POT3__STATIC, POT3__STATICRR },
    { "PIPE3B", DVPTR(TILE_PIPE3B), PIPE3B__STATIC, PIPE3B__STATICRR },
    { "WALLLIGHT3", DVPTR(TILE_WALLLIGHT3), WALLLIGHT3__STATIC, WALLLIGHT3__STATICRR },
    { "WALLLIGHTBUST3", DVPTR(TILE_WALLLIGHTBUST3), WALLLIGHTBUST3__STATIC, WALLLIGHTBUST3__STATICRR },
    { "WALLLIGHT1", DVPTR(TILE_WALLLIGHT1), WALLLIGHT1__STATIC, WALLLIGHT1__STATICRR },
    { "WALLLIGHTBUST1", DVPTR(TILE_WALLLIGHTBUST1), WALLLIGHTBUST1__STATIC, WALLLIGHTBUST1__STATICRR },
    { "WALLLIGHT2", DVPTR(TILE_WALLLIGHT2), WALLLIGHT2__STATIC, WALLLIGHT2__STATICRR },
    { "WALLLIGHTBUST2", DVPTR(TILE_WALLLIGHTBUST2), WALLLIGHTBUST2__STATIC, WALLLIGHTBUST2__STATICRR },
    { "LIGHTSWITCH2", DVPTR(TILE_LIGHTSWITCH2), LIGHTSWITCH2__STATIC, LIGHTSWITCH2__STATICRR },
    { "WAITTOBESEATED", DVPTR(TILE_WAITTOBESEATED), WAITTOBESEATED__STATIC, WAITTOBESEATED__STATICRR },
    { "DOORTILE14", DVPTR(TILE_DOORTILE14), DOORTILE14__STATIC, DOORTILE14__STATICRR },
    { "STATUE", DVPTR(TILE_STATUE), STATUE__STATIC, STATUE__STATICRR },
    { "MIKE", DVPTR(TILE_MIKE), MIKE__STATIC, MIKE__STATICRR },
    { "VASE", DVPTR(TILE_VASE), VASE__STATIC, VASE__STATICRR },
    { "SUSHIPLATE1", DVPTR(TILE_SUSHIPLATE1), SUSHIPLATE1__STATIC, SUSHIPLATE1__STATICRR },
    { "SUSHIPLATE2", DVPTR(TILE_SUSHIPLATE2), SUSHIPLATE2__STATIC, SUSHIPLATE2__STATICRR },
    { "SUSHIPLATE3", DVPTR(TILE_SUSHIPLATE3), SUSHIPLATE3__STATIC, SUSHIPLATE3__STATICRR },
    { "SUSHIPLATE4", DVPTR(TILE_SUSHIPLATE4), SUSHIPLATE4__STATIC, SUSHIPLATE4__STATICRR },
    { "DOORTILE16", DVPTR(TILE_DOORTILE16), DOORTILE16__STATIC, DOORTILE16__STATICRR },
    { "SUSHIPLATE5", DVPTR(TILE_SUSHIPLATE5), SUSHIPLATE5__STATIC, SUSHIPLATE5__STATICRR },
    { "OJ", DVPTR(TILE_OJ), OJ__STATIC, OJ__STATICRR },
    { "MASKWALL13", DVPTR(TILE_MASKWALL13), MASKWALL13__STATIC, MASKWALL13__STATICRR },
    { "HURTRAIL", DVPTR(TILE_HURTRAIL), HURTRAIL__STATIC, HURTRAIL__STATICRR },
    { "POWERSWITCH1", DVPTR(TILE_POWERSWITCH1), POWERSWITCH1__STATIC, POWERSWITCH1__STATICRR },
    { "LOCKSWITCH1", DVPTR(TILE_LOCKSWITCH1), LOCKSWITCH1__STATIC, LOCKSWITCH1__STATICRR },
    { "POWERSWITCH2", DVPTR(TILE_POWERSWITCH2), POWERSWITCH2__STATIC, POWERSWITCH2__STATICRR },
    { "ATM", DVPTR(TILE_ATM), ATM__STATIC, ATM__STATICRR },
    { "STATUEFLASH", DVPTR(TILE_STATUEFLASH), STATUEFLASH__STATIC, STATUEFLASH__STATICRR },
    { "ATMBROKE", DVPTR(TILE_ATMBROKE), ATMBROKE__STATIC, ATMBROKE__STATICRR },
    { "BIGHOLE2", DVPTR(TILE_BIGHOLE2), BIGHOLE2__STATIC, BIGHOLE2__STATICRR },
    { "STRIPEBALL", DVPTR(TILE_STRIPEBALL), STRIPEBALL__STATIC, STRIPEBALL__STATICRR },
    { "QUEBALL", DVPTR(TILE_QUEBALL), QUEBALL__STATIC, QUEBALL__STATICRR },
    { "POCKET", DVPTR(TILE_POCKET), POCKET__STATIC, POCKET__STATICRR },
    { "WOODENHORSE", DVPTR(TILE_WOODENHORSE), WOODENHORSE__STATIC, WOODENHORSE__STATICRR },
    { "TREE1", DVPTR(TILE_TREE1), TREE1__STATIC, TREE1__STATICRR },
    { "TREE2", DVPTR(TILE_TREE2), TREE2__STATIC, TREE2__STATICRR },
    { "CACTUS", DVPTR(TILE_CACTUS), CACTUS__STATIC, CACTUS__STATICRR },
    { "MASKWALL2", DVPTR(TILE_MASKWALL2), MASKWALL2__STATIC, MASKWALL2__STATICRR },
    { "MASKWALL3", DVPTR(TILE_MASKWALL3), MASKWALL3__STATIC, MASKWALL3__STATICRR },
    { "MASKWALL4", DVPTR(TILE_MASKWALL4), MASKWALL4__STATIC, MASKWALL4__STATICRR },
    { "FIREEXT", DVPTR(TILE_FIREEXT), FIREEXT__STATIC, FIREEXT__STATICRR },
    { "TOILETWATER", DVPTR(TILE_TOILETWATER), TOILETWATER__STATIC, TOILETWATER__STATICRR },
    { "NEON1", DVPTR(TILE_NEON1), NEON1__STATIC, NEON1__STATICRR },
    { "NEON2", DVPTR(TILE_NEON2), NEON2__STATIC, NEON2__STATICRR },
    { "CACTUSBROKE", DVPTR(TILE_CACTUSBROKE), CACTUSBROKE__STATIC, CACTUSBROKE__STATICRR },
    { "BOUNCEMINE", DVPTR(TILE_BOUNCEMINE), BOUNCEMINE__STATIC, BOUNCEMINE__STATICRR },
    { "BROKEFIREHYDRENT", DVPTR(TILE_BROKEFIREHYDRENT), BROKEFIREHYDRENT__STATIC, BROKEFIREHYDRENT__STATICRR },
    { "BOX", DVPTR(TILE_BOX), BOX__STATIC, BOX__STATICRR },
    { "BULLETHOLE", DVPTR(TILE_BULLETHOLE), BULLETHOLE__STATIC, BULLETHOLE__STATICRR },
    { "BOTTLE1", DVPTR(TILE_BOTTLE1), BOTTLE1__STATIC, BOTTLE1__STATICRR },
    { "BOTTLE2", DVPTR(TILE_BOTTLE2), BOTTLE2__STATIC, BOTTLE2__STATICRR },
    { "BOTTLE3", DVPTR(TILE_BOTTLE3), BOTTLE3__STATIC, BOTTLE3__STATICRR },
    { "BOTTLE4", DVPTR(TILE_BOTTLE4), BOTTLE4__STATIC, BOTTLE4__STATICRR },
    { "FEMPIC5", DVPTR(TILE_FEMPIC5), FEMPIC5__STATIC, FEMPIC5__STATICRR },
    { "FEMPIC6", DVPTR(TILE_FEMPIC6), FEMPIC6__STATIC, FEMPIC6__STATICRR },
    { "FEMPIC7", DVPTR(TILE_FEMPIC7), FEMPIC7__STATIC, FEMPIC7__STATICRR },
    { "HYDROPLANT", DVPTR(TILE_HYDROPLANT), HYDROPLANT__STATIC, HYDROPLANT__STATICRR },
    { "OCEANSPRITE1", DVPTR(TILE_OCEANSPRITE1), OCEANSPRITE1__STATIC, OCEANSPRITE1__STATICRR },
    { "OCEANSPRITE2", DVPTR(TILE_OCEANSPRITE2), OCEANSPRITE2__STATIC, OCEANSPRITE2__STATICRR },
    { "OCEANSPRITE3", DVPTR(TILE_OCEANSPRITE3), OCEANSPRITE3__STATIC, OCEANSPRITE3__STATICRR },
    { "OCEANSPRITE4", DVPTR(TILE_OCEANSPRITE4), OCEANSPRITE4__STATIC, OCEANSPRITE4__STATICRR },
    { "OCEANSPRITE5", DVPTR(TILE_OCEANSPRITE5), OCEANSPRITE5__STATIC, OCEANSPRITE5__STATICRR },
    { "GENERICPOLE", DVPTR(TILE_GENERICPOLE), GENERICPOLE__STATIC, GENERICPOLE__STATICRR },
    { "CONE", DVPTR(TILE_CONE), CONE__STATIC, CONE__STATICRR },
    { "HANGLIGHT", DVPTR(TILE_HANGLIGHT), HANGLIGHT__STATIC, HANGLIGHT__STATICRR },
    { "HYDRENT", DVPTR(TILE_HYDRENT), HYDRENT__STATIC, HYDRENT__STATICRR },
    { "MASKWALL14", DVPTR(TILE_MASKWALL14), MASKWALL14__STATIC, MASKWALL14__STATICRR },
    { "TIRE", DVPTR(TILE_TIRE), TIRE__STATIC, TIRE__STATICRR },
    { "PIPE5", DVPTR(TILE_PIPE5), PIPE5__STATIC, PIPE5__STATICRR },
    { "PIPE6", DVPTR(TILE_PIPE6), PIPE6__STATIC, PIPE6__STATICRR },
    { "PIPE4", DVPTR(TILE_PIPE4), PIPE4__STATIC, PIPE4__STATICRR },
    { "PIPE4B", DVPTR(TILE_PIPE4B), PIPE4B__STATIC, PIPE4B__STATICRR },
    { "BROKEHYDROPLANT", DVPTR(TILE_BROKEHYDROPLANT), BROKEHYDROPLANT__STATIC, BROKEHYDROPLANT__STATICRR },
    { "PIPE5B", DVPTR(TILE_PIPE5B), PIPE5B__STATIC, PIPE5B__STATICRR },
    { "NEON3", DVPTR(TILE_NEON3), NEON3__STATIC, NEON3__STATICRR },
    { "NEON4", DVPTR(TILE_NEON4), NEON4__STATIC, NEON4__STATICRR },
    { "NEON5", DVPTR(TILE_NEON5), NEON5__STATIC, NEON5__STATICRR },
    { "BOTTLE5", DVPTR(TILE_BOTTLE5), BOTTLE5__STATIC, BOTTLE5__STATICRR },
    { "BOTTLE6", DVPTR(TILE_BOTTLE6), BOTTLE6__STATIC, BOTTLE6__STATICRR },
    { "BOTTLE8", DVPTR(TILE_BOTTLE8), BOTTLE8__STATIC, BOTTLE8__STATICRR },
    { "SPOTLITE", DVPTR(TILE_SPOTLITE), SPOTLITE__STATIC, SPOTLITE__STATICRR },
    { "HANGOOZ", DVPTR(TILE_HANGOOZ), HANGOOZ__STATIC, HANGOOZ__STATICRR },
    { "MASKWALL15", DVPTR(TILE_MASKWALL15), MASKWALL15__STATIC, MASKWALL15__STATICRR },
    { "BOTTLE7", DVPTR(TILE_BOTTLE7), BOTTLE7__STATIC, BOTTLE7__STATICRR },
    { "HORSEONSIDE", DVPTR(TILE_HORSEONSIDE), HORSEONSIDE__STATIC, HORSEONSIDE__STATICRR },
    { "GLASSPIECES", DVPTR(TILE_GLASSPIECES), GLASSPIECES__STATIC, GLASSPIECES__STATICRR },
    { "HORSELITE", DVPTR(TILE_HORSELITE), HORSELITE__STATIC, HORSELITE__STATICRR },
    { "DONUTS", DVPTR(TILE_DONUTS), DONUTS__STATIC, DONUTS__STATICRR },
    { "NEON6", DVPTR(TILE_NEON6), NEON6__STATIC, NEON6__STATICRR },
    { "MASKWALL6", DVPTR(TILE_MASKWALL6), MASKWALL6__STATIC, MASKWALL6__STATICRR },
    { "CLOCK", DVPTR(TILE_CLOCK), CLOCK__STATIC, CLOCK__STATICRR },
    { "RUBBERCAN", DVPTR(TILE_RUBBERCAN), RUBBERCAN__STATIC, RUBBERCAN__STATICRR },
    { "BROKENCLOCK", DVPTR(TILE_BROKENCLOCK), BROKENCLOCK__STATIC, BROKENCLOCK__STATICRR },
    { "PLUG", DVPTR(TILE_PLUG), PLUG__STATIC, PLUG__STATICRR },
    { "OOZFILTER", DVPTR(TILE_OOZFILTER), OOZFILTER__STATIC, OOZFILTER__STATICRR },
    { "FLOORPLASMA", DVPTR(TILE_FLOORPLASMA), FLOORPLASMA__STATIC, FLOORPLASMA__STATICRR },
    { "REACTOR", DVPTR(TILE_REACTOR), REACTOR__STATIC, REACTOR__STATICRR },
    { "REACTORSPARK", DVPTR(TILE_REACTORSPARK), REACTORSPARK__STATIC, REACTORSPARK__STATICRR },
    { "REACTORBURNT", DVPTR(TILE_REACTORBURNT), REACTORBURNT__STATIC, REACTORBURNT__STATICRR },
    { "DOORTILE15", DVPTR(TILE_DOORTILE15), DOORTILE15__STATIC, DOORTILE15__STATICRR },
    { "HANDSWITCH", DVPTR(TILE_HANDSWITCH), HANDSWITCH__STATIC, HANDSWITCH__STATICRR },
    { "CIRCLEPANNEL", DVPTR(TILE_CIRCLEPANNEL), CIRCLEPANNEL__STATIC, CIRCLEPANNEL__STATICRR },
    { "CIRCLEPANNELBROKE", DVPTR(TILE_CIRCLEPANNELBROKE), CIRCLEPANNELBROKE__STATIC, CIRCLEPANNELBROKE__STATICRR },
    { "PULLSWITCH", DVPTR(TILE_PULLSWITCH), PULLSWITCH__STATIC, PULLSWITCH__STATICRR },
    { "MASKWALL8", DVPTR(TILE_MASKWALL8), MASKWALL8__STATIC, MASKWALL8__STATICRR },
    { "BIGHOLE", DVPTR(TILE_BIGHOLE), BIGHOLE__STATIC, BIGHOLE__STATICRR },
    { "ALIENSWITCH", DVPTR(TILE_ALIENSWITCH), ALIENSWITCH__STATIC, ALIENSWITCH__STATICRR },
    { "DOORTILE21", DVPTR(TILE_DOORTILE21), DOORTILE21__STATIC, DOORTILE21__STATICRR },
    { "HANDPRINTSWITCH", DVPTR(TILE_HANDPRINTSWITCH), HANDPRINTSWITCH__STATIC, HANDPRINTSWITCH__STATICRR },
    { "BOTTLE10", DVPTR(TILE_BOTTLE10), BOTTLE10__STATIC, BOTTLE10__STATICRR },
    { "BOTTLE11", DVPTR(TILE_BOTTLE11), BOTTLE11__STATIC, BOTTLE11__STATICRR },
    { "BOTTLE12", DVPTR(TILE_BOTTLE12), BOTTLE12__STATIC, BOTTLE12__STATICRR },
    { "BOTTLE13", DVPTR(TILE_BOTTLE13), BOTTLE13__STATIC, BOTTLE13__STATICRR },
    { "BOTTLE14", DVPTR(TILE_BOTTLE14), BOTTLE14__STATIC, BOTTLE14__STATICRR },
    { "BOTTLE15", DVPTR(TILE_BOTTLE15), BOTTLE15__STATIC, BOTTLE15__STATICRR },
    { "BOTTLE16", DVPTR(TILE_BOTTLE16), BOTTLE16__STATIC, BOTTLE16__STATICRR },
    { "BOTTLE17", DVPTR(TILE_BOTTLE17), BOTTLE17__STATIC, BOTTLE17__STATICRR },
    { "BOTTLE18", DVPTR(TILE_BOTTLE18), BOTTLE18__STATIC, BOTTLE18__STATICRR },
    { "BOTTLE19", DVPTR(TILE_BOTTLE19), BOTTLE19__STATIC, BOTTLE19__STATICRR },
    { "DOORTILE17", DVPTR(TILE_DOORTILE17), DOORTILE17__STATIC, DOORTILE17__STATICRR },
    { "MASKWALL7", DVPTR(TILE_MASKWALL7), MASKWALL7__STATIC, MASKWALL7__STATICRR },
    { "JAILBARBREAK", DVPTR(TILE_JAILBARBREAK), JAILBARBREAK__STATIC, JAILBARBREAK__STATICRR },
    { "DOORTILE11", DVPTR(TILE_DOORTILE11), DOORTILE11__STATIC, DOORTILE11__STATICRR },
    { "DOORTILE12", DVPTR(TILE_DOORTILE12), DOORTILE12__STATIC, DOORTILE12__STATICRR },
    { "VENDMACHINE", DVPTR(TILE_VENDMACHINE), VENDMACHINE__STATIC, VENDMACHINE__STATICRR },
    { "VENDMACHINEBROKE", DVPTR(TILE_VENDMACHINEBROKE), VENDMACHINEBROKE__STATIC, VENDMACHINEBROKE__STATICRR },
    { "COLAMACHINE", DVPTR(TILE_COLAMACHINE), COLAMACHINE__STATIC, COLAMACHINE__STATICRR },
    { "COLAMACHINEBROKE", DVPTR(TILE_COLAMACHINEBROKE), COLAMACHINEBROKE__STATIC, COLAMACHINEBROKE__STATICRR },
    { "CRANEPOLE", DVPTR(TILE_CRANEPOLE), CRANEPOLE__STATIC, CRANEPOLE__STATICRR },
    { "CRANE", DVPTR(TILE_CRANE), CRANE__STATIC, CRANE__STATICRR },
    { "BARBROKE", DVPTR(TILE_BARBROKE), BARBROKE__STATIC, BARBROKE__STATICRR },
    { "BLOODPOOL", DVPTR(TILE_BLOODPOOL), BLOODPOOL__STATIC, BLOODPOOL__STATICRR },
    { "NUKEBARREL", DVPTR(TILE_NUKEBARREL), NUKEBARREL__STATIC, NUKEBARREL__STATICRR },
    { "NUKEBARRELDENTED", DVPTR(TILE_NUKEBARRELDENTED), NUKEBARRELDENTED__STATIC, NUKEBARRELDENTED__STATICRR },
    { "NUKEBARRELLEAKED", DVPTR(TILE_NUKEBARRELLEAKED), NUKEBARRELLEAKED__STATIC, NUKEBARRELLEAKED__STATICRR },
    { "CANWITHSOMETHING", DVPTR(TILE_CANWITHSOMETHING), CANWITHSOMETHING__STATIC, CANWITHSOMETHING__STATICRR },
    { "MONEY", DVPTR(TILE_MONEY), MONEY__STATIC, MONEY__STATICRR },
    { "BANNER", DVPTR(TILE_BANNER), BANNER__STATIC, BANNER__STATICRR },
    { "EXPLODINGBARREL", DVPTR(TILE_EXPLODINGBARREL), EXPLODINGBARREL__STATIC, EXPLODINGBARREL__STATICRR },
    { "EXPLODINGBARREL2", DVPTR(TILE_EXPLODINGBARREL2), EXPLODINGBARREL2__STATIC, EXPLODINGBARREL2__STATICRR },
    { "FIREBARREL", DVPTR(TILE_FIREBARREL), FIREBARREL__STATIC, FIREBARREL__STATICRR },
    { "SEENINE", DVPTR(TILE_SEENINE), SEENINE__STATIC, SEENINE__STATICRR },
    { "SEENINEDEAD", DVPTR(TILE_SEENINEDEAD), SEENINEDEAD__STATIC, SEENINEDEAD__STATICRR },
    { "STEAM", DVPTR(TILE_STEAM), STEAM__STATIC, STEAM__STATICRR },
    { "CEILINGSTEAM", DVPTR(TILE_CEILINGSTEAM), CEILINGSTEAM__STATIC, CEILINGSTEAM__STATICRR },
    { "PIPE6B", DVPTR(TILE_PIPE6B), PIPE6B__STATIC, PIPE6B__STATICRR },
    { "TRANSPORTERBEAM", DVPTR(TILE_TRANSPORTERBEAM), TRANSPORTERBEAM__STATIC, TRANSPORTERBEAM__STATICRR },
    { "RAT", DVPTR(TILE_RAT), RAT__STATIC, RAT__STATICRR },
    { "TRASH", DVPTR(TILE_TRASH), TRASH__STATIC, TRASH__STATICRR },
    { "FEMPIC1", DVPTR(TILE_FEMPIC1), FEMPIC1__STATIC, FEMPIC1__STATICRR },
    { "FEMPIC2", DVPTR(TILE_FEMPIC2), FEMPIC2__STATIC, FEMPIC2__STATICRR },
    { "BLANKSCREEN", DVPTR(TILE_BLANKSCREEN), BLANKSCREEN__STATIC, BLANKSCREEN__STATICRR },
    { "PODFEM1", DVPTR(TILE_PODFEM1), PODFEM1__STATIC, PODFEM1__STATICRR },
    { "FEMPIC3", DVPTR(TILE_FEMPIC3), FEMPIC3__STATIC, FEMPIC3__STATICRR },
    { "FEMPIC4", DVPTR(TILE_FEMPIC4), FEMPIC4__STATIC, FEMPIC4__STATICRR },
    { "FEM1", DVPTR(TILE_FEM1), FEM1__STATIC, FEM1__STATICRR },
    { "FEM2", DVPTR(TILE_FEM2), FEM2__STATIC, FEM2__STATICRR },
    { "FEM3", DVPTR(TILE_FEM3), FEM3__STATIC, FEM3__STATICRR },
    { "FEM5", DVPTR(TILE_FEM5), FEM5__STATIC, FEM5__STATICRR },
    { "BLOODYPOLE", DVPTR(TILE_BLOODYPOLE), BLOODYPOLE__STATIC, BLOODYPOLE__STATICRR },
    { "FEM4", DVPTR(TILE_FEM4), FEM4__STATIC, FEM4__STATICRR },
    { "FEM6", DVPTR(TILE_FEM6), FEM6__STATIC, FEM6__STATICRR },
    { "FEM6PAD", DVPTR(TILE_FEM6PAD), FEM6PAD__STATIC, FEM6PAD__STATICRR },
    { "FEM8", DVPTR(TILE_FEM8), FEM8__STATIC, FEM8__STATICRR },
    { "HELECOPT", DVPTR(TILE_HELECOPT), HELECOPT__STATIC, HELECOPT__STATICRR },
    { "FETUSJIB", DVPTR(TILE_FETUSJIB), FETUSJIB__STATIC, FETUSJIB__STATICRR },
    { "HOLODUKE", DVPTR(TILE_HOLODUKE), HOLODUKE__STATIC, HOLODUKE__STATICRR },
    { "SPACEMARINE", DVPTR(TILE_SPACEMARINE), SPACEMARINE__STATIC, SPACEMARINE__STATICRR },
    { "INDY", DVPTR(TILE_INDY), INDY__STATIC, INDY__STATICRR },
    { "FETUS", DVPTR(TILE_FETUS), FETUS__STATIC, FETUS__STATICRR },
    { "FETUSBROKE", DVPTR(TILE_FETUSBROKE), FETUSBROKE__STATIC, FETUSBROKE__STATICRR },
    { "MONK", DVPTR(TILE_MONK), MONK__STATIC, MONK__STATICRR },
    { "LUKE", DVPTR(TILE_LUKE), LUKE__STATIC, LUKE__STATICRR },
    { "COOLEXPLOSION1", DVPTR(TILE_COOLEXPLOSION1), COOLEXPLOSION1__STATIC, COOLEXPLOSION1__STATICRR },
    { "WATERSPLASH2", DVPTR(TILE_WATERSPLASH2), WATERSPLASH2__STATIC, WATERSPLASH2__STATICRR },
    { "FIREVASE", DVPTR(TILE_FIREVASE), FIREVASE__STATIC, FIREVASE__STATICRR },
    { "SCRATCH", DVPTR(TILE_SCRATCH), SCRATCH__STATIC, SCRATCH__STATICRR },
    { "FEM7", DVPTR(TILE_FEM7), FEM7__STATIC, FEM7__STATICRR },
    { "APLAYERTOP", DVPTR(TILE_APLAYERTOP), APLAYERTOP__STATIC, APLAYERTOP__STATICRR },
    { "APLAYER", DVPTR(TILE_APLAYER), APLAYER__STATIC, APLAYER__STATICRR },
    { "PLAYERONWATER", DVPTR(TILE_PLAYERONWATER), PLAYERONWATER__STATIC, PLAYERONWATER__STATICRR },
    { "DUKELYINGDEAD", DVPTR(TILE_DUKELYINGDEAD), DUKELYINGDEAD__STATIC, DUKELYINGDEAD__STATICRR },
    { "DUKETORSO", DVPTR(TILE_DUKETORSO), DUKETORSO__STATIC, DUKETORSO__STATICRR },
    { "DUKEGUN", DVPTR(TILE_DUKEGUN), DUKEGUN__STATIC, DUKEGUN__STATICRR },
    { "DUKELEG", DVPTR(TILE_DUKELEG), DUKELEG__STATIC, DUKELEG__STATICRR },
    { "SHARK", DVPTR(TILE_SHARK), SHARK__STATIC, SHARK__STATICRR },
    { "BLOOD", DVPTR(TILE_BLOOD), BLOOD__STATIC, BLOOD__STATICRR },
    { "FIRELASER", DVPTR(TILE_FIRELASER), FIRELASER__STATIC, FIRELASER__STATICRR },
    { "TRANSPORTERSTAR", DVPTR(TILE_TRANSPORTERSTAR), TRANSPORTERSTAR__STATIC, TRANSPORTERSTAR__STATICRR },
    { "SPIT", DVPTR(TILE_SPIT), SPIT__STATIC, SPIT__STATICRR },
    { "LOOGIE", DVPTR(TILE_LOOGIE), LOOGIE__STATIC, LOOGIE__STATICRR },
    { "FIST", DVPTR(TILE_FIST), FIST__STATIC, FIST__STATICRR },
    { "FREEZEBLAST", DVPTR(TILE_FREEZEBLAST), FREEZEBLAST__STATIC, FREEZEBLAST__STATICRR },
    { "DEVISTATORBLAST", DVPTR(TILE_DEVISTATORBLAST), DEVISTATORBLAST__STATIC, DEVISTATORBLAST__STATICRR },
    { "SHRINKSPARK", DVPTR(TILE_SHRINKSPARK), SHRINKSPARK__STATIC, SHRINKSPARK__STATICRR },
    { "TONGUE", DVPTR(TILE_TONGUE), TONGUE__STATIC, TONGUE__STATICRR },
    { "MORTER", DVPTR(TILE_MORTER), MORTER__STATIC, MORTER__STATICRR },
    { "SHRINKEREXPLOSION", DVPTR(TILE_SHRINKEREXPLOSION), SHRINKEREXPLOSION__STATIC, SHRINKEREXPLOSION__STATICRR },
    { "RADIUSEXPLOSION", DVPTR(TILE_RADIUSEXPLOSION), RADIUSEXPLOSION__STATIC, RADIUSEXPLOSION__STATICRR },
    { "FORCERIPPLE", DVPTR(TILE_FORCERIPPLE), FORCERIPPLE__STATIC, FORCERIPPLE__STATICRR },
    { "LIZTROOP", DVPTR(TILE_LIZTROOP), LIZTROOP__STATIC, 0 },
    { "LIZTROOPRUNNING", DVPTR(TILE_LIZTROOPRUNNING), LIZTROOPRUNNING__STATIC, 0 },
    { "LIZTROOPSTAYPUT", DVPTR(TILE_LIZTROOPSTAYPUT), LIZTROOPSTAYPUT__STATIC, 0 },
    { "LIZTOP", DVPTR(TILE_LIZTOP), LIZTOP__STATIC, 0 },
    { "LIZTROOPSHOOT", DVPTR(TILE_LIZTROOPSHOOT), LIZTROOPSHOOT__STATIC, 0 },
    { "LIZTROOPJETPACK", DVPTR(TILE_LIZTROOPJETPACK), LIZTROOPJETPACK__STATIC, 0 },
    { "LIZTROOPDSPRITE", DVPTR(TILE_LIZTROOPDSPRITE), LIZTROOPDSPRITE__STATIC, 0 },
    { "LIZTROOPONTOILET", DVPTR(TILE_LIZTROOPONTOILET), LIZTROOPONTOILET__STATIC, 0 },
    { "LIZTROOPJUSTSIT", DVPTR(TILE_LIZTROOPJUSTSIT), LIZTROOPJUSTSIT__STATIC, 0 },
    { "LIZTROOPDUCKING", DVPTR(TILE_LIZTROOPDUCKING), LIZTROOPDUCKING__STATIC, 0 },
    { "HEADJIB1", DVPTR(TILE_HEADJIB1), HEADJIB1__STATIC, 0 },
    { "ARMJIB1", DVPTR(TILE_ARMJIB1), ARMJIB1__STATIC, 0 },
    { "LEGJIB1", DVPTR(TILE_LEGJIB1), LEGJIB1__STATIC, 0 },
    { "CANNONBALL", DVPTR(TILE_CANNONBALL), CANNONBALL__STATIC, CANNONBALL__STATICRR },
    { "OCTABRAIN", DVPTR(TILE_OCTABRAIN), OCTABRAIN__STATIC, 0 },
    { "OCTABRAINSTAYPUT", DVPTR(TILE_OCTABRAINSTAYPUT), OCTABRAINSTAYPUT__STATIC, 0 },
    { "OCTATOP", DVPTR(TILE_OCTATOP), OCTATOP__STATIC, 0 },
    { "OCTADEADSPRITE", DVPTR(TILE_OCTADEADSPRITE), OCTADEADSPRITE__STATIC, 0 },
    { "INNERJAW", DVPTR(TILE_INNERJAW), INNERJAW__STATIC, INNERJAW__STATICRR },
    { "DRONE", DVPTR(TILE_DRONE), DRONE__STATIC, DRONE__STATICRR },
    { "EXPLOSION2", DVPTR(TILE_EXPLOSION2), EXPLOSION2__STATIC, EXPLOSION2__STATICRR },
    { "COMMANDER", DVPTR(TILE_COMMANDER), COMMANDER__STATIC, 0 },
    { "COMMANDERSTAYPUT", DVPTR(TILE_COMMANDERSTAYPUT), COMMANDERSTAYPUT__STATIC, 0 },
    { "RECON", DVPTR(TILE_RECON), RECON__STATIC, RECON__STATICRR },
    { "TANK", DVPTR(TILE_TANK), TANK__STATIC, 0 },
    { "PIGCOP", DVPTR(TILE_PIGCOP), PIGCOP__STATIC, 0 },
    { "PIGCOPSTAYPUT", DVPTR(TILE_PIGCOPSTAYPUT), PIGCOPSTAYPUT__STATIC, 0 },
    { "PIGCOPDIVE", DVPTR(TILE_PIGCOPDIVE), PIGCOPDIVE__STATIC, 0 },
    { "PIGCOPDEADSPRITE", DVPTR(TILE_PIGCOPDEADSPRITE), PIGCOPDEADSPRITE__STATIC, 0 },
    { "PIGTOP", DVPTR(TILE_PIGTOP), PIGTOP__STATIC, 0 },
    { "LIZMAN", DVPTR(TILE_LIZMAN), LIZMAN__STATIC, 0 },
    { "LIZMANSTAYPUT", DVPTR(TILE_LIZMANSTAYPUT), LIZMANSTAYPUT__STATIC, 0 },
    { "LIZMANSPITTING", DVPTR(TILE_LIZMANSPITTING), LIZMANSPITTING__STATIC, 0 },
    { "LIZMANFEEDING", DVPTR(TILE_LIZMANFEEDING), LIZMANFEEDING__STATIC, 0 },
    { "LIZMANJUMP", DVPTR(TILE_LIZMANJUMP), LIZMANJUMP__STATIC, 0 },
    { "LIZMANDEADSPRITE", DVPTR(TILE_LIZMANDEADSPRITE), LIZMANDEADSPRITE__STATIC, 0 },
    { "FECES", DVPTR(TILE_FECES), FECES__STATIC, FECES__STATICRR },
    { "LIZMANHEAD1", DVPTR(TILE_LIZMANHEAD1), LIZMANHEAD1__STATIC, 0 },
    { "LIZMANARM1", DVPTR(TILE_LIZMANARM1), LIZMANARM1__STATIC, 0 },
    { "LIZMANLEG1", DVPTR(TILE_LIZMANLEG1), LIZMANLEG1__STATIC, 0 },
    { "EXPLOSION2BOT", DVPTR(TILE_EXPLOSION2BOT), EXPLOSION2BOT__STATIC, EXPLOSION2BOT__STATICRR },
    { "USERWEAPON", DVPTR(TILE_USERWEAPON), USERWEAPON__STATIC, USERWEAPON__STATICRR },
    { "HEADERBAR", DVPTR(TILE_HEADERBAR), HEADERBAR__STATIC, HEADERBAR__STATICRR },
    { "JIBS1", DVPTR(TILE_JIBS1), JIBS1__STATIC, JIBS1__STATICRR },
    { "JIBS2", DVPTR(TILE_JIBS2), JIBS2__STATIC, JIBS2__STATICRR },
    { "JIBS3", DVPTR(TILE_JIBS3), JIBS3__STATIC, JIBS3__STATICRR },
    { "JIBS4", DVPTR(TILE_JIBS4), JIBS4__STATIC, JIBS4__STATICRR },
    { "JIBS5", DVPTR(TILE_JIBS5), JIBS5__STATIC, JIBS5__STATICRR },
    { "BURNING", DVPTR(TILE_BURNING), BURNING__STATIC, BURNING__STATICRR },
    { "FIRE", DVPTR(TILE_FIRE), FIRE__STATIC, FIRE__STATICRR },
    { "JIBS6", DVPTR(TILE_JIBS6), JIBS6__STATIC, JIBS6__STATICRR },
    { "BLOODSPLAT1", DVPTR(TILE_BLOODSPLAT1), BLOODSPLAT1__STATIC, BLOODSPLAT1__STATICRR },
    { "BLOODSPLAT3", DVPTR(TILE_BLOODSPLAT3), BLOODSPLAT3__STATIC, BLOODSPLAT3__STATICRR },
    { "BLOODSPLAT2", DVPTR(TILE_BLOODSPLAT2), BLOODSPLAT2__STATIC, BLOODSPLAT2__STATICRR },
    { "BLOODSPLAT4", DVPTR(TILE_BLOODSPLAT4), BLOODSPLAT4__STATIC, BLOODSPLAT4__STATICRR },
    { "OOZ", DVPTR(TILE_OOZ), OOZ__STATIC, OOZ__STATICRR },
    { "OOZ2", DVPTR(TILE_OOZ2), OOZ2__STATIC, OOZ2__STATICRR },
    { "WALLBLOOD1", DVPTR(TILE_WALLBLOOD1), WALLBLOOD1__STATIC, WALLBLOOD1__STATICRR },
    { "WALLBLOOD2", DVPTR(TILE_WALLBLOOD2), WALLBLOOD2__STATIC, WALLBLOOD2__STATICRR },
    { "WALLBLOOD3", DVPTR(TILE_WALLBLOOD3), WALLBLOOD3__STATIC, WALLBLOOD3__STATICRR },
    { "WALLBLOOD4", DVPTR(TILE_WALLBLOOD4), WALLBLOOD4__STATIC, WALLBLOOD4__STATICRR },
    { "WALLBLOOD5", DVPTR(TILE_WALLBLOOD5), WALLBLOOD5__STATIC, WALLBLOOD5__STATICRR },
    { "WALLBLOOD6", DVPTR(TILE_WALLBLOOD6), WALLBLOOD6__STATIC, WALLBLOOD6__STATICRR },
    { "WALLBLOOD7", DVPTR(TILE_WALLBLOOD7), WALLBLOOD7__STATIC, WALLBLOOD7__STATICRR },
    { "WALLBLOOD8", DVPTR(TILE_WALLBLOOD8), WALLBLOOD8__STATIC, WALLBLOOD8__STATICRR },
    { "BURNING2", DVPTR(TILE_BURNING2), BURNING2__STATIC, BURNING2__STATICRR },
    { "FIRE2", DVPTR(TILE_FIRE2), FIRE2__STATIC, FIRE2__STATICRR },
    { "CRACKKNUCKLES", DVPTR(TILE_CRACKKNUCKLES), CRACKKNUCKLES__STATIC, CRACKKNUCKLES__STATICRR },
    { "SMALLSMOKE", DVPTR(TILE_SMALLSMOKE), SMALLSMOKE__STATIC, SMALLSMOKE__STATICRR },
    { "SMALLSMOKEMAKER", DVPTR(TILE_SMALLSMOKEMAKER), SMALLSMOKEMAKER__STATIC, SMALLSMOKEMAKER__STATICRR },
    { "FLOORFLAME", DVPTR(TILE_FLOORFLAME), FLOORFLAME__STATIC, FLOORFLAME__STATICRR },
    { "ROTATEGUN", DVPTR(TILE_ROTATEGUN), ROTATEGUN__STATIC, ROTATEGUN__STATICRR },
    { "GREENSLIME", DVPTR(TILE_GREENSLIME), GREENSLIME__STATIC, GREENSLIME__STATICRR },
    { "WATERDRIPSPLASH", DVPTR(TILE_WATERDRIPSPLASH), WATERDRIPSPLASH__STATIC, WATERDRIPSPLASH__STATICRR },
    { "SCRAP6", DVPTR(TILE_SCRAP6), SCRAP6__STATIC, SCRAP6__STATICRR },
    { "SCRAP1", DVPTR(TILE_SCRAP1), SCRAP1__STATIC, SCRAP1__STATICRR },
    { "SCRAP2", DVPTR(TILE_SCRAP2), SCRAP2__STATIC, SCRAP2__STATICRR },
    { "SCRAP3", DVPTR(TILE_SCRAP3), SCRAP3__STATIC, SCRAP3__STATICRR },
    { "SCRAP4", DVPTR(TILE_SCRAP4), SCRAP4__STATIC, SCRAP4__STATICRR },
    { "SCRAP5", DVPTR(TILE_SCRAP5), SCRAP5__STATIC, SCRAP5__STATICRR },
    { "ORGANTIC", DVPTR(TILE_ORGANTIC), ORGANTIC__STATIC, ORGANTIC__STATICRR },
    { "BETAVERSION", DVPTR(TILE_BETAVERSION), BETAVERSION__STATIC, BETAVERSION__STATICRR },
    { "PLAYERISHERE", DVPTR(TILE_PLAYERISHERE), PLAYERISHERE__STATIC, PLAYERISHERE__STATICRR },
    { "PLAYERWASHERE", DVPTR(TILE_PLAYERWASHERE), PLAYERWASHERE__STATIC, PLAYERWASHERE__STATICRR },
    { "SELECTDIR", DVPTR(TILE_SELECTDIR), SELECTDIR__STATIC, SELECTDIR__STATICRR },
    { "F1HELP", DVPTR(TILE_F1HELP), F1HELP__STATIC, F1HELP__STATICRR },
    { "NOTCHON", DVPTR(TILE_NOTCHON), NOTCHON__STATIC, NOTCHON__STATICRR },
    { "NOTCHOFF", DVPTR(TILE_NOTCHOFF), NOTCHOFF__STATIC, NOTCHOFF__STATICRR },
    { "GROWSPARK", DVPTR(TILE_GROWSPARK), GROWSPARK__STATIC, GROWSPARK__STATICRR },
    { "DUKEICON", DVPTR(TILE_DUKEICON), DUKEICON__STATIC, DUKEICON__STATICRR },
    { "BADGUYICON", DVPTR(TILE_BADGUYICON), BADGUYICON__STATIC, BADGUYICON__STATICRR },
    { "FOODICON", DVPTR(TILE_FOODICON), FOODICON__STATIC, FOODICON__STATICRR },
    { "GETICON", DVPTR(TILE_GETICON), GETICON__STATIC, GETICON__STATICRR },
    { "MENUSCREEN", DVPTR(TILE_MENUSCREEN), MENUSCREEN__STATIC, MENUSCREEN__STATICRR },
    { "MENUBAR", DVPTR(TILE_MENUBAR), MENUBAR__STATIC, MENUBAR__STATICRR },
    { "KILLSICON", DVPTR(TILE_KILLSICON), KILLSICON__STATIC, KILLSICON__STATICRR },
    { "FIRSTAID_ICON", DVPTR(TILE_FIRSTAID_ICON), FIRSTAID_ICON__STATIC, FIRSTAID_ICON__STATICRR },
    { "HEAT_ICON", DVPTR(TILE_HEAT_ICON), HEAT_ICON__STATIC, HEAT_ICON__STATICRR },
    { "BOTTOMSTATUSBAR", DVPTR(TILE_BOTTOMSTATUSBAR), BOTTOMSTATUSBAR__STATIC, BOTTOMSTATUSBAR__STATICRR },
    { "BOOT_ICON", DVPTR(TILE_BOOT_ICON), BOOT_ICON__STATIC, BOOT_ICON__STATICRR },
    { "FRAGBAR", DVPTR(TILE_FRAGBAR), FRAGBAR__STATIC, FRAGBAR__STATICRR },
    { "JETPACK_ICON", DVPTR(TILE_JETPACK_ICON), JETPACK_ICON__STATIC, JETPACK_ICON__STATICRR },
    { "AIRTANK_ICON", DVPTR(TILE_AIRTANK_ICON), AIRTANK_ICON__STATIC, AIRTANK_ICON__STATICRR },
    { "STEROIDS_ICON", DVPTR(TILE_STEROIDS_ICON), STEROIDS_ICON__STATIC, STEROIDS_ICON__STATICRR },
    { "HOLODUKE_ICON", DVPTR(TILE_HOLODUKE_ICON), HOLODUKE_ICON__STATIC, HOLODUKE_ICON__STATICRR },
    { "ACCESS_ICON", DVPTR(TILE_ACCESS_ICON), ACCESS_ICON__STATIC, ACCESS_ICON__STATICRR },
    { "DIGITALNUM", DVPTR(TILE_DIGITALNUM), DIGITALNUM__STATIC, DIGITALNUM__STATICRR },
    { "DUKECAR", DVPTR(TILE_DUKECAR), DUKECAR__STATIC, DUKECAR__STATICRR },
    { "CAMCORNER", DVPTR(TILE_CAMCORNER), CAMCORNER__STATIC, CAMCORNER__STATICRR },
    { "CAMLIGHT", DVPTR(TILE_CAMLIGHT), CAMLIGHT__STATIC, CAMLIGHT__STATICRR },
    { "LOGO", DVPTR(TILE_LOGO), LOGO__STATIC, LOGO__STATICRR },
    { "TITLE", DVPTR(TILE_TITLE), TITLE__STATIC, TITLE__STATICRR },
    { "NUKEWARNINGICON", DVPTR(TILE_NUKEWARNINGICON), NUKEWARNINGICON__STATIC, NUKEWARNINGICON__STATICRR },
    { "MOUSECURSOR", DVPTR(TILE_MOUSECURSOR), MOUSECURSOR__STATIC, MOUSECURSOR__STATICRR },
    { "SLIDEBAR", DVPTR(TILE_SLIDEBAR), SLIDEBAR__STATIC, SLIDEBAR__STATICRR },
    { "DREALMS", DVPTR(TILE_DREALMS), DREALMS__STATIC, DREALMS__STATICRR },
    { "BETASCREEN", DVPTR(TILE_BETASCREEN), BETASCREEN__STATIC, BETASCREEN__STATICRR },
    { "WINDOWBORDER1", DVPTR(TILE_WINDOWBORDER1), WINDOWBORDER1__STATIC, WINDOWBORDER1__STATICRR },
    { "TEXTBOX", DVPTR(TILE_TEXTBOX), TEXTBOX__STATIC, TEXTBOX__STATICRR },
    { "WINDOWBORDER2", DVPTR(TILE_WINDOWBORDER2), WINDOWBORDER2__STATIC, WINDOWBORDER2__STATICRR },
    { "DUKENUKEM", DVPTR(TILE_DUKENUKEM), DUKENUKEM__STATIC, DUKENUKEM__STATICRR },
    { "THREEDEE", DVPTR(TILE_THREEDEE), THREEDEE__STATIC, THREEDEE__STATICRR },
    { "INGAMEDUKETHREEDEE", DVPTR(TILE_INGAMEDUKETHREEDEE), INGAMEDUKETHREEDEE__STATIC, INGAMEDUKETHREEDEE__STATICRR },
    { "TENSCREEN", DVPTR(TILE_TENSCREEN), TENSCREEN__STATIC, TENSCREEN__STATICRR },
    { "PLUTOPAKSPRITE", DVPTR(TILE_PLUTOPAKSPRITE), PLUTOPAKSPRITE__STATIC, PLUTOPAKSPRITE__STATICRR },
    { "DEVISTATOR", DVPTR(TILE_DEVISTATOR), DEVISTATOR__STATIC, DEVISTATOR__STATICRR },
    { "KNEE", DVPTR(TILE_KNEE), KNEE__STATIC, KNEE__STATICRR },
    { "CROSSHAIR", DVPTR(TILE_CROSSHAIR), CROSSHAIR__STATIC, CROSSHAIR__STATICRR },
    { "FIRSTGUN", DVPTR(TILE_FIRSTGUN), FIRSTGUN__STATIC, FIRSTGUN__STATICRR },
    { "FIRSTGUNRELOAD", DVPTR(TILE_FIRSTGUNRELOAD), FIRSTGUNRELOAD__STATIC, FIRSTGUNRELOAD__STATICRR },
    { "FALLINGCLIP", DVPTR(TILE_FALLINGCLIP), FALLINGCLIP__STATIC, FALLINGCLIP__STATICRR },
    { "CLIPINHAND", DVPTR(TILE_CLIPINHAND), CLIPINHAND__STATIC, CLIPINHAND__STATICRR },
    { "HAND", DVPTR(TILE_HAND), HAND__STATIC, HAND__STATICRR },
    { "SHELL", DVPTR(TILE_SHELL), SHELL__STATIC, SHELL__STATICRR },
    { "SHOTGUNSHELL", DVPTR(TILE_SHOTGUNSHELL), SHOTGUNSHELL__STATIC, SHOTGUNSHELL__STATICRR },
    { "CHAINGUN", DVPTR(TILE_CHAINGUN), CHAINGUN__STATIC, CHAINGUN__STATICRR },
    { "RPGGUN", DVPTR(TILE_RPGGUN), RPGGUN__STATIC, RPGGUN__STATICRR },
    { "RPGMUZZLEFLASH", DVPTR(TILE_RPGMUZZLEFLASH), RPGMUZZLEFLASH__STATIC, RPGMUZZLEFLASH__STATICRR },
    { "FREEZE", DVPTR(TILE_FREEZE), FREEZE__STATIC, FREEZE__STATICRR },
    { "CATLITE", DVPTR(TILE_CATLITE), CATLITE__STATIC, CATLITE__STATICRR },
    { "SHRINKER", DVPTR(TILE_SHRINKER), SHRINKER__STATIC, SHRINKER__STATICRR },
    { "HANDHOLDINGLASER", DVPTR(TILE_HANDHOLDINGLASER), HANDHOLDINGLASER__STATIC, HANDHOLDINGLASER__STATICRR },
    { "TRIPBOMB", DVPTR(TILE_TRIPBOMB), TRIPBOMB__STATIC, TRIPBOMB__STATICRR },
    { "LASERLINE", DVPTR(TILE_LASERLINE), LASERLINE__STATIC, LASERLINE__STATICRR },
    { "HANDHOLDINGACCESS", DVPTR(TILE_HANDHOLDINGACCESS), HANDHOLDINGACCESS__STATIC, HANDHOLDINGACCESS__STATICRR },
    { "HANDREMOTE", DVPTR(TILE_HANDREMOTE), HANDREMOTE__STATIC, HANDREMOTE__STATICRR },
    { "HANDTHROW", DVPTR(TILE_HANDTHROW), HANDTHROW__STATIC, HANDTHROW__STATICRR },
    { "TIP", DVPTR(TILE_TIP), TIP__STATIC, TIP__STATICRR },
    { "GLAIR", DVPTR(TILE_GLAIR), GLAIR__STATIC, GLAIR__STATICRR },
    { "SCUBAMASK", DVPTR(TILE_SCUBAMASK), SCUBAMASK__STATIC, SCUBAMASK__STATICRR },
    { "SPACEMASK", DVPTR(TILE_SPACEMASK), SPACEMASK__STATIC, SPACEMASK__STATICRR },
    { "FORCESPHERE", DVPTR(TILE_FORCESPHERE), FORCESPHERE__STATIC, FORCESPHERE__STATICRR },
    { "SHOTSPARK1", DVPTR(TILE_SHOTSPARK1), SHOTSPARK1__STATIC, SHOTSPARK1__STATICRR },
    { "RPG", DVPTR(TILE_RPG), RPG__STATIC, RPG__STATICRR },
    { "LASERSITE", DVPTR(TILE_LASERSITE), LASERSITE__STATIC, 0 },
    { "SHOTGUN", DVPTR(TILE_SHOTGUN), SHOTGUN__STATIC, SHOTGUN__STATICRR },
    { "BOSS1", DVPTR(TILE_BOSS1), BOSS1__STATIC, BOSS1__STATICRR },
    { "BOSS1STAYPUT", DVPTR(TILE_BOSS1STAYPUT), BOSS1STAYPUT__STATIC, 0 },
    { "BOSS1SHOOT", DVPTR(TILE_BOSS1SHOOT), BOSS1SHOOT__STATIC, 0 },
    { "BOSS1LOB", DVPTR(TILE_BOSS1LOB), BOSS1LOB__STATIC, 0 },
    { "BOSSTOP", DVPTR(TILE_BOSSTOP), BOSSTOP__STATIC, 0 },
    { "BOSS2", DVPTR(TILE_BOSS2), BOSS2__STATIC, BOSS2__STATICRR },
    { "BOSS3", DVPTR(TILE_BOSS3), BOSS3__STATIC, BOSS3__STATICRR },
    { "SPINNINGNUKEICON", DVPTR(TILE_SPINNINGNUKEICON), SPINNINGNUKEICON__STATIC, SPINNINGNUKEICON__STATICRR },
    { "BIGFNTCURSOR", DVPTR(TILE_BIGFNTCURSOR), BIGFNTCURSOR__STATIC, BIGFNTCURSOR__STATICRR },
    { "SMALLFNTCURSOR", DVPTR(TILE_SMALLFNTCURSOR), SMALLFNTCURSOR__STATIC, SMALLFNTCURSOR__STATICRR },
    { "STARTALPHANUM", DVPTR(TILE_STARTALPHANUM), STARTALPHANUM__STATIC, STARTALPHANUM__STATICRR },
    { "ENDALPHANUM", DVPTR(TILE_ENDALPHANUM), ENDALPHANUM__STATIC, ENDALPHANUM__STATICRR },
    { "BIGALPHANUM", DVPTR(TILE_BIGALPHANUM), BIGALPHANUM__STATIC, BIGALPHANUM__STATICRR },
    { "BIGPERIOD", DVPTR(TILE_BIGPERIOD), BIGPERIOD__STATIC, BIGPERIOD__STATICRR },
    { "BIGCOMMA", DVPTR(TILE_BIGCOMMA), BIGCOMMA__STATIC, BIGCOMMA__STATICRR },
    { "BIGX", DVPTR(TILE_BIGX_), BIGX__STATIC, BIGX__STATICRR },
    { "BIGQ", DVPTR(TILE_BIGQ), BIGQ__STATIC, BIGQ__STATICRR },
    { "BIGSEMI", DVPTR(TILE_BIGSEMI), BIGSEMI__STATIC, BIGSEMI__STATICRR },
    { "BIGCOLIN", DVPTR(TILE_BIGCOLIN), BIGCOLIN__STATIC, BIGCOLIN__STATICRR },
    { "THREEBYFIVE", DVPTR(TILE_THREEBYFIVE), THREEBYFIVE__STATIC, THREEBYFIVE__STATICRR },
    { "BIGAPPOS", DVPTR(TILE_BIGAPPOS), BIGAPPOS__STATIC, BIGAPPOS__STATICRR },
    { "BLANK", DVPTR(TILE_BLANK), BLANK__STATIC, BLANK__STATICRR },
    { "MINIFONT", DVPTR(TILE_MINIFONT), MINIFONT__STATIC, MINIFONT__STATICRR },
    { "BUTTON1", DVPTR(TILE_BUTTON1), BUTTON1__STATIC, BUTTON1__STATICRR },
    { "GLASS3", DVPTR(TILE_GLASS3), GLASS3__STATIC, GLASS3__STATICRR },
    { "RESPAWNMARKERRED", DVPTR(TILE_RESPAWNMARKERRED), RESPAWNMARKERRED__STATIC, RESPAWNMARKERRED__STATICRR },
    { "RESPAWNMARKERYELLOW", DVPTR(TILE_RESPAWNMARKERYELLOW), RESPAWNMARKERYELLOW__STATIC, RESPAWNMARKERYELLOW__STATICRR },
    { "RESPAWNMARKERGREEN", DVPTR(TILE_RESPAWNMARKERGREEN), RESPAWNMARKERGREEN__STATIC, RESPAWNMARKERGREEN__STATICRR },
    { "BONUSSCREEN", DVPTR(TILE_BONUSSCREEN), BONUSSCREEN__STATIC, BONUSSCREEN__STATICRR },
    { "VIEWBORDER", DVPTR(TILE_VIEWBORDER), VIEWBORDER__STATIC, VIEWBORDER__STATICRR },
    { "VICTORY1", DVPTR(TILE_VICTORY1), VICTORY1__STATIC, VICTORY1__STATICRR },
    { "ORDERING", DVPTR(TILE_ORDERING), ORDERING__STATIC, ORDERING__STATICRR },
    { "TEXTSTORY", DVPTR(TILE_TEXTSTORY), TEXTSTORY__STATIC, TEXTSTORY__STATICRR },
    { "LOADSCREEN", DVPTR(TILE_LOADSCREEN), LOADSCREEN__STATIC, LOADSCREEN__STATICRR },
    { "BORNTOBEWILDSCREEN", DVPTR(TILE_BORNTOBEWILDSCREEN), BORNTOBEWILDSCREEN__STATIC, BORNTOBEWILDSCREEN__STATICRR },
    { "BLIMP", DVPTR(TILE_BLIMP), BLIMP__STATIC, BLIMP__STATICRR },
    { "FEM9", DVPTR(TILE_FEM9), FEM9__STATIC, FEM9__STATICRR },
    { "FOOTPRINT", DVPTR(TILE_FOOTPRINT), FOOTPRINT__STATIC, FOOTPRINT__STATICRR },
    { "POOP", DVPTR(TILE_POOP), POOP__STATIC, POOP__STATICRR },
    { "FRAMEEFFECT1", DVPTR(TILE_FRAMEEFFECT1), FRAMEEFFECT1__STATIC, FRAMEEFFECT1__STATICRR },
    { "PANNEL3", DVPTR(TILE_PANNEL3), PANNEL3__STATIC, PANNEL3__STATICRR },
    { "SCREENBREAK14", DVPTR(TILE_SCREENBREAK14), SCREENBREAK14__STATIC, SCREENBREAK14__STATICRR },
    { "SCREENBREAK15", DVPTR(TILE_SCREENBREAK15), SCREENBREAK15__STATIC, SCREENBREAK15__STATICRR },
    { "SCREENBREAK19", DVPTR(TILE_SCREENBREAK19), SCREENBREAK19__STATIC, SCREENBREAK19__STATICRR },
    { "SCREENBREAK16", DVPTR(TILE_SCREENBREAK16), SCREENBREAK16__STATIC, SCREENBREAK16__STATICRR },
    { "SCREENBREAK17", DVPTR(TILE_SCREENBREAK17), SCREENBREAK17__STATIC, SCREENBREAK17__STATICRR },
    { "SCREENBREAK18", DVPTR(TILE_SCREENBREAK18), SCREENBREAK18__STATIC, SCREENBREAK18__STATICRR },
    { "W_TECHWALL11", DVPTR(TILE_W_TECHWALL11), W_TECHWALL11__STATIC, W_TECHWALL11__STATICRR },
    { "W_TECHWALL12", DVPTR(TILE_W_TECHWALL12), W_TECHWALL12__STATIC, W_TECHWALL12__STATICRR },
    { "W_TECHWALL13", DVPTR(TILE_W_TECHWALL13), W_TECHWALL13__STATIC, W_TECHWALL13__STATICRR },
    { "W_TECHWALL14", DVPTR(TILE_W_TECHWALL14), W_TECHWALL14__STATIC, W_TECHWALL14__STATICRR },
    { "W_TECHWALL5", DVPTR(TILE_W_TECHWALL5), W_TECHWALL5__STATIC, W_TECHWALL5__STATICRR },
    { "W_TECHWALL6", DVPTR(TILE_W_TECHWALL6), W_TECHWALL6__STATIC, W_TECHWALL6__STATICRR },
    { "W_TECHWALL7", DVPTR(TILE_W_TECHWALL7), W_TECHWALL7__STATIC, W_TECHWALL7__STATICRR },
    { "W_TECHWALL8", DVPTR(TILE_W_TECHWALL8), W_TECHWALL8__STATIC, W_TECHWALL8__STATICRR },
    { "W_TECHWALL9", DVPTR(TILE_W_TECHWALL9), W_TECHWALL9__STATIC, W_TECHWALL9__STATICRR },
    { "BPANNEL3", DVPTR(TILE_BPANNEL3), BPANNEL3__STATIC, BPANNEL3__STATICRR },
    { "W_HITTECHWALL16", DVPTR(TILE_W_HITTECHWALL16), W_HITTECHWALL16__STATIC, W_HITTECHWALL16__STATICRR },
    { "W_HITTECHWALL10", DVPTR(TILE_W_HITTECHWALL10), W_HITTECHWALL10__STATIC, W_HITTECHWALL10__STATICRR },
    { "W_HITTECHWALL15", DVPTR(TILE_W_HITTECHWALL15), W_HITTECHWALL15__STATIC, W_HITTECHWALL15__STATICRR },
    { "W_MILKSHELF", DVPTR(TILE_W_MILKSHELF), W_MILKSHELF__STATIC, W_MILKSHELF__STATICRR },
    { "W_MILKSHELFBROKE", DVPTR(TILE_W_MILKSHELFBROKE), W_MILKSHELFBROKE__STATIC, W_MILKSHELFBROKE__STATICRR },
    { "PURPLELAVA", DVPTR(TILE_PURPLELAVA), PURPLELAVA__STATIC, PURPLELAVA__STATICRR },
    { "LAVABUBBLE", DVPTR(TILE_LAVABUBBLE), LAVABUBBLE__STATIC, LAVABUBBLE__STATICRR },
    { "DUKECUTOUT", DVPTR(TILE_DUKECUTOUT), DUKECUTOUT__STATIC, DUKECUTOUT__STATICRR },
    { "TARGET", DVPTR(TILE_TARGET), TARGET__STATIC, TARGET__STATICRR },
    { "GUNPOWDERBARREL", DVPTR(TILE_GUNPOWDERBARREL), GUNPOWDERBARREL__STATIC, GUNPOWDERBARREL__STATICRR },
    { "DUCK", DVPTR(TILE_DUCK), DUCK__STATIC, DUCK__STATICRR },
    { "HATRACK", DVPTR(TILE_HATRACK), HATRACK__STATIC, HATRACK__STATICRR },
    { "DESKLAMP", DVPTR(TILE_DESKLAMP), DESKLAMP__STATIC, DESKLAMP__STATICRR },
    { "COFFEEMACHINE", DVPTR(TILE_COFFEEMACHINE), COFFEEMACHINE__STATIC, COFFEEMACHINE__STATICRR },
    { "CUPS", DVPTR(TILE_CUPS), CUPS__STATIC, CUPS__STATICRR },
    { "GAVALS", DVPTR(TILE_GAVALS), GAVALS__STATIC, GAVALS__STATICRR },
    { "GAVALS2", DVPTR(TILE_GAVALS2), GAVALS2__STATIC, GAVALS2__STATICRR },
    { "POLICELIGHTPOLE", DVPTR(TILE_POLICELIGHTPOLE), POLICELIGHTPOLE__STATIC, POLICELIGHTPOLE__STATICRR },
    { "FLOORBASKET", DVPTR(TILE_FLOORBASKET), FLOORBASKET__STATIC, FLOORBASKET__STATICRR },
    { "PUKE", DVPTR(TILE_PUKE), PUKE__STATIC, PUKE__STATICRR },
    { "DOORTILE23", DVPTR(TILE_DOORTILE23), DOORTILE23__STATIC, DOORTILE23__STATICRR },
    { "TOPSECRET", DVPTR(TILE_TOPSECRET), TOPSECRET__STATIC, TOPSECRET__STATICRR },
    { "SPEAKER", DVPTR(TILE_SPEAKER), SPEAKER__STATIC, SPEAKER__STATICRR },
    { "TEDDYBEAR", DVPTR(TILE_TEDDYBEAR), TEDDYBEAR__STATIC, TEDDYBEAR__STATICRR },
    { "ROBOTDOG", DVPTR(TILE_ROBOTDOG), ROBOTDOG__STATIC, ROBOTDOG__STATICRR },
    { "ROBOTPIRATE", DVPTR(TILE_ROBOTPIRATE), ROBOTPIRATE__STATIC, ROBOTPIRATE__STATICRR },
    { "ROBOTMOUSE", DVPTR(TILE_ROBOTMOUSE), ROBOTMOUSE__STATIC, ROBOTMOUSE__STATICRR },
    { "MAIL", DVPTR(TILE_MAIL), MAIL__STATIC, MAIL__STATICRR },
    { "MAILBAG", DVPTR(TILE_MAILBAG), MAILBAG__STATIC, MAILBAG__STATICRR },
    { "HOTMEAT", DVPTR(TILE_HOTMEAT), HOTMEAT__STATIC, HOTMEAT__STATICRR },
    { "COFFEEMUG", DVPTR(TILE_COFFEEMUG), COFFEEMUG__STATIC, COFFEEMUG__STATICRR },
    { "DONUTS2", DVPTR(TILE_DONUTS2), DONUTS2__STATIC, DONUTS2__STATICRR },
    { "TRIPODCAMERA", DVPTR(TILE_TRIPODCAMERA), TRIPODCAMERA__STATIC, TRIPODCAMERA__STATICRR },
    { "METER", DVPTR(TILE_METER), METER__STATIC, METER__STATICRR },
    { "DESKPHONE", DVPTR(TILE_DESKPHONE), DESKPHONE__STATIC, DESKPHONE__STATICRR },
    { "GUMBALLMACHINE", DVPTR(TILE_GUMBALLMACHINE), GUMBALLMACHINE__STATIC, GUMBALLMACHINE__STATICRR },
    { "GUMBALLMACHINEBROKE", DVPTR(TILE_GUMBALLMACHINEBROKE), GUMBALLMACHINEBROKE__STATIC, GUMBALLMACHINEBROKE__STATICRR },
    { "PAPER", DVPTR(TILE_PAPER), PAPER__STATIC, PAPER__STATICRR },
    { "MACE", DVPTR(TILE_MACE), MACE__STATIC, MACE__STATICRR },
    { "GENERICPOLE2", DVPTR(TILE_GENERICPOLE2), GENERICPOLE2__STATIC, GENERICPOLE2__STATICRR },
    { "XXXSTACY", DVPTR(TILE_XXXSTACY), XXXSTACY__STATIC, XXXSTACY__STATICRR },
    { "WETFLOOR", DVPTR(TILE_WETFLOOR), WETFLOOR__STATIC, WETFLOOR__STATICRR },
    { "BROOM", DVPTR(TILE_BROOM), BROOM__STATIC, BROOM__STATICRR },
    { "MOP", DVPTR(TILE_MOP), MOP__STATIC, MOP__STATICRR },
    { "LETTER", DVPTR(TILE_LETTER), LETTER__STATIC, 0 },
    { "PIRATE1A", DVPTR(TILE_PIRATE1A), PIRATE1A__STATIC, PIRATE1A__STATICRR },
    { "PIRATE4A", DVPTR(TILE_PIRATE4A), PIRATE4A__STATIC, PIRATE4A__STATICRR },
    { "PIRATE2A", DVPTR(TILE_PIRATE2A), PIRATE2A__STATIC, PIRATE2A__STATICRR },
    { "PIRATE5A", DVPTR(TILE_PIRATE5A), PIRATE5A__STATIC, PIRATE5A__STATICRR },
    { "PIRATE3A", DVPTR(TILE_PIRATE3A), PIRATE3A__STATIC, PIRATE3A__STATICRR },
    { "PIRATE6A", DVPTR(TILE_PIRATE6A), PIRATE6A__STATIC, PIRATE6A__STATICRR },
    { "PIRATEHALF", DVPTR(TILE_PIRATEHALF), PIRATEHALF__STATIC, PIRATEHALF__STATICRR },
    { "CHESTOFGOLD", DVPTR(TILE_CHESTOFGOLD), CHESTOFGOLD__STATIC, CHESTOFGOLD__STATICRR },
    { "SIDEBOLT1", DVPTR(TILE_SIDEBOLT1), SIDEBOLT1__STATIC, SIDEBOLT1__STATICRR },
    { "FOODOBJECT1", DVPTR(TILE_FOODOBJECT1), FOODOBJECT1__STATIC, FOODOBJECT1__STATICRR },
    { "FOODOBJECT2", DVPTR(TILE_FOODOBJECT2), FOODOBJECT2__STATIC, FOODOBJECT2__STATICRR },
    { "FOODOBJECT3", DVPTR(TILE_FOODOBJECT3), FOODOBJECT3__STATIC, FOODOBJECT3__STATICRR },
    { "FOODOBJECT4", DVPTR(TILE_FOODOBJECT4), FOODOBJECT4__STATIC, FOODOBJECT4__STATICRR },
    { "FOODOBJECT5", DVPTR(TILE_FOODOBJECT5), FOODOBJECT5__STATIC, FOODOBJECT5__STATICRR },
    { "FOODOBJECT6", DVPTR(TILE_FOODOBJECT6), FOODOBJECT6__STATIC, FOODOBJECT6__STATICRR },
    { "FOODOBJECT7", DVPTR(TILE_FOODOBJECT7), FOODOBJECT7__STATIC, FOODOBJECT7__STATICRR },
    { "FOODOBJECT8", DVPTR(TILE_FOODOBJECT8), FOODOBJECT8__STATIC, FOODOBJECT8__STATICRR },
    { "FOODOBJECT9", DVPTR(TILE_FOODOBJECT9), FOODOBJECT9__STATIC, FOODOBJECT9__STATICRR },
    { "FOODOBJECT10", DVPTR(TILE_FOODOBJECT10), FOODOBJECT10__STATIC, FOODOBJECT10__STATICRR },
    { "FOODOBJECT11", DVPTR(TILE_FOODOBJECT11), FOODOBJECT11__STATIC, FOODOBJECT11__STATICRR },
    { "FOODOBJECT12", DVPTR(TILE_FOODOBJECT12), FOODOBJECT12__STATIC, FOODOBJECT12__STATICRR },
    { "FOODOBJECT13", DVPTR(TILE_FOODOBJECT13), FOODOBJECT13__STATIC, FOODOBJECT13__STATICRR },
    { "FOODOBJECT14", DVPTR(TILE_FOODOBJECT14), FOODOBJECT14__STATIC, FOODOBJECT14__STATICRR },
    { "FOODOBJECT15", DVPTR(TILE_FOODOBJECT15), FOODOBJECT15__STATIC, FOODOBJECT15__STATICRR },
    { "FOODOBJECT16", DVPTR(TILE_FOODOBJECT16), FOODOBJECT16__STATIC, FOODOBJECT16__STATICRR },
    { "FOODOBJECT17", DVPTR(TILE_FOODOBJECT17), FOODOBJECT17__STATIC, FOODOBJECT17__STATICRR },
    { "FOODOBJECT18", DVPTR(TILE_FOODOBJECT18), FOODOBJECT18__STATIC, FOODOBJECT18__STATICRR },
    { "FOODOBJECT19", DVPTR(TILE_FOODOBJECT19), FOODOBJECT19__STATIC, FOODOBJECT19__STATICRR },
    { "FOODOBJECT20", DVPTR(TILE_FOODOBJECT20), FOODOBJECT20__STATIC, FOODOBJECT20__STATICRR },
    { "HEADLAMP", DVPTR(TILE_HEADLAMP), HEADLAMP__STATIC, HEADLAMP__STATICRR },
    { "TAMPON", DVPTR(TILE_TAMPON), TAMPON__STATIC, TAMPON__STATICRR },
    { "SKINNEDCHICKEN", DVPTR(TILE_SKINNEDCHICKEN), SKINNEDCHICKEN__STATIC, SKINNEDCHICKEN__STATICRR },
    { "FEATHEREDCHICKEN", DVPTR(TILE_FEATHEREDCHICKEN), FEATHEREDCHICKEN__STATIC, FEATHEREDCHICKEN__STATICRR },
    { "ROBOTDOG2", DVPTR(TILE_ROBOTDOG2), ROBOTDOG2__STATIC, ROBOTDOG2__STATICRR },
    { "JOLLYMEAL", DVPTR(TILE_JOLLYMEAL), JOLLYMEAL__STATIC, JOLLYMEAL__STATICRR },
    { "DUKEBURGER", DVPTR(TILE_DUKEBURGER), DUKEBURGER__STATIC, DUKEBURGER__STATICRR },
    { "SHOPPINGCART", DVPTR(TILE_SHOPPINGCART), SHOPPINGCART__STATIC, SHOPPINGCART__STATICRR },
    { "CANWITHSOMETHING2", DVPTR(TILE_CANWITHSOMETHING2), CANWITHSOMETHING2__STATIC, CANWITHSOMETHING2__STATICRR },
    { "CANWITHSOMETHING3", DVPTR(TILE_CANWITHSOMETHING3), CANWITHSOMETHING3__STATIC, CANWITHSOMETHING3__STATICRR },
    { "CANWITHSOMETHING4", DVPTR(TILE_CANWITHSOMETHING4), CANWITHSOMETHING4__STATIC, CANWITHSOMETHING4__STATICRR },
    { "SNAKEP", DVPTR(TILE_SNAKEP), SNAKEP__STATIC, SNAKEP__STATICRR },
    { "DOLPHIN1", DVPTR(TILE_DOLPHIN1), DOLPHIN1__STATIC, DOLPHIN1__STATICRR },
    { "DOLPHIN2", DVPTR(TILE_DOLPHIN2), DOLPHIN2__STATIC, DOLPHIN2__STATICRR },
    { "NEWBEAST", DVPTR(TILE_NEWBEAST), NEWBEAST__STATIC, 0 },
    { "NEWBEASTSTAYPUT", DVPTR(TILE_NEWBEASTSTAYPUT), NEWBEASTSTAYPUT__STATIC, 0 },
    { "NEWBEASTJUMP", DVPTR(TILE_NEWBEASTJUMP), NEWBEASTJUMP__STATIC, 0 },
    { "NEWBEASTHANG", DVPTR(TILE_NEWBEASTHANG), NEWBEASTHANG__STATIC, 0 },
    { "NEWBEASTHANGDEAD", DVPTR(TILE_NEWBEASTHANGDEAD), NEWBEASTHANGDEAD__STATIC, 0 },
    { "BOSS4", DVPTR(TILE_BOSS4), BOSS4__STATIC, BOSS4__STATICRR },
    { "BOSS4STAYPUT", DVPTR(TILE_BOSS4STAYPUT), BOSS4STAYPUT__STATIC, 0 },
    { "FEM10", DVPTR(TILE_FEM10), FEM10__STATIC, FEM10__STATICRR },
    { "TOUGHGAL", DVPTR(TILE_TOUGHGAL), TOUGHGAL__STATIC, TOUGHGAL__STATICRR },
    { "MAN", DVPTR(TILE_MAN), MAN__STATIC, MAN__STATICRR },
    { "MAN2", DVPTR(TILE_MAN2), MAN2__STATIC, MAN2__STATICRR },
    { "WOMAN", DVPTR(TILE_WOMAN), WOMAN__STATIC, WOMAN__STATICRR },
    { "PLEASEWAIT", DVPTR(TILE_PLEASEWAIT), PLEASEWAIT__STATIC, PLEASEWAIT__STATICRR },
    { "NATURALLIGHTNING", DVPTR(TILE_NATURALLIGHTNING), NATURALLIGHTNING__STATIC, 0 },
    { "WEATHERWARN", DVPTR(TILE_WEATHERWARN), WEATHERWARN__STATIC, 0 },
    { "DUKETAG", DVPTR(TILE_DUKETAG), DUKETAG__STATIC, DUKETAG__STATICRR },
    { "SIGN1", DVPTR(TILE_SIGN1), SIGN1__STATIC, SIGN1__STATICRR },
    { "SIGN2", DVPTR(TILE_SIGN2), SIGN2__STATIC, SIGN2__STATICRR },
    { "JURYGUY", DVPTR(TILE_JURYGUY), JURYGUY__STATIC, JURYGUY__STATICRR },
    { "RRTILE11", DVPTR(TILE_RRTILE11), 0, RRTILE11__STATICRR },
    { "RPG2SPRITE", DVPTR(TILE_RPG2SPRITE), 0, RPG2SPRITE__STATICRR },
    { "RRTILE18", DVPTR(TILE_RRTILE18), 0, RRTILE18__STATICRR },
    { "RRTILE19", DVPTR(TILE_RRTILE19), 0, RRTILE19__STATICRR },
    { "RRTILE34", DVPTR(TILE_RRTILE34), 0, RRTILE34__STATICRR },
    { "RRTILE35", DVPTR(TILE_RRTILE35), 0, RRTILE35__STATICRR },
    { "DESTRUCTO", DVPTR(TILE_DESTRUCTO), 0, DESTRUCTO__STATICRR },
    { "RRTILE38", DVPTR(TILE_RRTILE38), 0, RRTILE38__STATICRR },
    { "RRTILE43", DVPTR(TILE_RRTILE43), 0, RRTILE43__STATICRR },
    { "GUTMETER", DVPTR(TILE_GUTMETER), 0, GUTMETER__STATICRR },
    { "RRTILE63", DVPTR(TILE_RRTILE63), 0, RRTILE63__STATICRR },
    { "RRTILE64", DVPTR(TILE_RRTILE64), 0, RRTILE64__STATICRR },
    { "RRTILE65", DVPTR(TILE_RRTILE65), 0, RRTILE65__STATICRR },
    { "RRTILE66", DVPTR(TILE_RRTILE66), 0, RRTILE66__STATICRR },
    { "RRTILE67", DVPTR(TILE_RRTILE67), 0, RRTILE67__STATICRR },
    { "RRTILE68", DVPTR(TILE_RRTILE68), 0, RRTILE68__STATICRR },
    { "SOUNDFX", DVPTR(TILE_SOUNDFX), 0, SOUNDFX__STATICRR },
    { "MOTOAMMO", DVPTR(TILE_MOTOAMMO), 0, MOTOAMMO__STATICRR },
    { "UFOBEAM", DVPTR(TILE_UFOBEAM), 0, UFOBEAM__STATICRR },
    { "RRTILE280", DVPTR(TILE_RRTILE280), 0, RRTILE280__STATICRR },
    { "RRTILE281", DVPTR(TILE_RRTILE281), 0, RRTILE281__STATICRR },
    { "RRTILE282", DVPTR(TILE_RRTILE282), 0, RRTILE282__STATICRR },
    { "RRTILE283", DVPTR(TILE_RRTILE283), 0, RRTILE283__STATICRR },
    { "RRTILE285", DVPTR(TILE_RRTILE285), 0, RRTILE285__STATICRR },
    { "RRTILE286", DVPTR(TILE_RRTILE286), 0, RRTILE286__STATICRR },
    { "RRTILE287", DVPTR(TILE_RRTILE287), 0, RRTILE287__STATICRR },
    { "RRTILE288", DVPTR(TILE_RRTILE288), 0, RRTILE288__STATICRR },
    { "RRTILE289", DVPTR(TILE_RRTILE289), 0, RRTILE289__STATICRR },
    { "RRTILE290", DVPTR(TILE_RRTILE290), 0, RRTILE290__STATICRR },
    { "RRTILE291", DVPTR(TILE_RRTILE291), 0, RRTILE291__STATICRR },
    { "RRTILE292", DVPTR(TILE_RRTILE292), 0, RRTILE292__STATICRR },
    { "RRTILE293", DVPTR(TILE_RRTILE293), 0, RRTILE293__STATICRR },
    { "RRTILE295", DVPTR(TILE_RRTILE295), 0, RRTILE295__STATICRR },
    { "RRTILE296", DVPTR(TILE_RRTILE296), 0, RRTILE296__STATICRR },
    { "RRTILE297", DVPTR(TILE_RRTILE297), 0, RRTILE297__STATICRR },
    { "CDPLAYER", DVPTR(TILE_CDPLAYER), 0, CDPLAYER__STATICRR },
    { "RRTILE380", DVPTR(TILE_RRTILE380), 0, RRTILE380__STATICRR },
    { "RRTILE403", DVPTR(TILE_RRTILE403), 0, RRTILE403__STATICRR },
    { "RRTILE409", DVPTR(TILE_RRTILE409), 0, RRTILE409__STATICRR },
    { "GUTMETER_LIGHT1", DVPTR(TILE_GUTMETER_LIGHT1), 0, GUTMETER_LIGHT1__STATICRR },
    { "GUTMETER_LIGHT2", DVPTR(TILE_GUTMETER_LIGHT2), 0, GUTMETER_LIGHT2__STATICRR },
    { "GUTMETER_LIGHT3", DVPTR(TILE_GUTMETER_LIGHT3), 0, GUTMETER_LIGHT3__STATICRR },
    { "GUTMETER_LIGHT4", DVPTR(TILE_GUTMETER_LIGHT4), 0, GUTMETER_LIGHT4__STATICRR },
    { "AMMO_ICON", DVPTR(TILE_AMMO_ICON), 0, AMMO_ICON__STATICRR },
    { "RRTILE409", DVPTR(TILE_RRTILE1076), 0, RRTILE1076__STATICRR },
    { "MUD", DVPTR(TILE_MUD), 0, MUD__STATICRR },
    { "EXPLOSION3", DVPTR(TILE_EXPLOSION3), 0, EXPLOSION3__STATICRR },
    { "RRTILE1636", DVPTR(TILE_RRTILE1636), 0, RRTILE1636__STATICRR },
    { "WEAPONBAR", DVPTR(TILE_WEAPONBAR), 0, WEAPONBAR__STATICRR },
    { "RRTILE1752", DVPTR(TILE_RRTILE1752), 0, RRTILE1752__STATICRR },
    { "RPG2", DVPTR(TILE_RPG2), 0, RPG2__STATICRR },
    { "RRTILE1790", DVPTR(TILE_RRTILE1790), 0, RRTILE1790__STATICRR },
    { "RRTILE1792", DVPTR(TILE_RRTILE1792), 0, RRTILE1792__STATICRR },
    { "RRTILE1801", DVPTR(TILE_RRTILE1801), 0, RRTILE1801__STATICRR },
    { "RRTILE1805", DVPTR(TILE_RRTILE1805), 0, RRTILE1805__STATICRR },
    { "RRTILE1807", DVPTR(TILE_RRTILE1807), 0, RRTILE1807__STATICRR },
    { "RRTILE1808", DVPTR(TILE_RRTILE1808), 0, RRTILE1808__STATICRR },
    { "RRTILE1812", DVPTR(TILE_RRTILE1812), 0, RRTILE1812__STATICRR },
    { "RRTILE1814", DVPTR(TILE_RRTILE1814), 0, RRTILE1814__STATICRR },
    { "RRTILE1817", DVPTR(TILE_RRTILE1817), 0, RRTILE1817__STATICRR },
    { "RRTILE1821", DVPTR(TILE_RRTILE1821), 0, RRTILE1821__STATICRR },
    { "RRTILE1824", DVPTR(TILE_RRTILE1824), 0, RRTILE1824__STATICRR },
    { "RRTILE1826", DVPTR(TILE_RRTILE1826), 0, RRTILE1826__STATICRR },
    { "RRTILE1850", DVPTR(TILE_RRTILE1850), 0, RRTILE1850__STATICRR },
    { "RRTILE1851", DVPTR(TILE_RRTILE1851), 0, RRTILE1851__STATICRR },
    { "RRTILE1856", DVPTR(TILE_RRTILE1856), 0, RRTILE1856__STATICRR },
    { "RRTILE1877", DVPTR(TILE_RRTILE1877), 0, RRTILE1877__STATICRR },
    { "RRTILE1878", DVPTR(TILE_RRTILE1878), 0, RRTILE1878__STATICRR },
    { "RRTILE1938", DVPTR(TILE_RRTILE1938), 0, RRTILE1938__STATICRR },
    { "RRTILE1939", DVPTR(TILE_RRTILE1939), 0, RRTILE1939__STATICRR },
    { "RRTILE1942", DVPTR(TILE_RRTILE1942), 0, RRTILE1942__STATICRR },
    { "RRTILE1944", DVPTR(TILE_RRTILE1944), 0, RRTILE1944__STATICRR },
    { "RRTILE1945", DVPTR(TILE_RRTILE1945), 0, RRTILE1945__STATICRR },
    { "RRTILE1947", DVPTR(TILE_RRTILE1947), 0, RRTILE1947__STATICRR },
    { "RRTILE1951", DVPTR(TILE_RRTILE1951), 0, RRTILE1951__STATICRR },
    { "RRTILE1952", DVPTR(TILE_RRTILE1952), 0, RRTILE1952__STATICRR },
    { "RRTILE1953", DVPTR(TILE_RRTILE1953), 0, RRTILE1953__STATICRR },
    { "RRTILE1961", DVPTR(TILE_RRTILE1961), 0, RRTILE1961__STATICRR },
    { "RRTILE1964", DVPTR(TILE_RRTILE1964), 0, RRTILE1964__STATICRR },
    { "RRTILE1973", DVPTR(TILE_RRTILE1973), 0, RRTILE1973__STATICRR },
    { "RRTILE1985", DVPTR(TILE_RRTILE1985), 0, RRTILE1985__STATICRR },
    { "RRTILE1986", DVPTR(TILE_RRTILE1986), 0, RRTILE1986__STATICRR },
    { "RRTILE1987", DVPTR(TILE_RRTILE1987), 0, RRTILE1987__STATICRR },
    { "RRTILE1988", DVPTR(TILE_RRTILE1988), 0, RRTILE1988__STATICRR },
    { "RRTILE1990", DVPTR(TILE_RRTILE1990), 0, RRTILE1990__STATICRR },
    { "RRTILE1995", DVPTR(TILE_RRTILE1995), 0, RRTILE1995__STATICRR },
    { "RRTILE1996", DVPTR(TILE_RRTILE1996), 0, RRTILE1996__STATICRR },
    { "RRTILE2004", DVPTR(TILE_RRTILE2004), 0, RRTILE2004__STATICRR },
    { "RRTILE2005", DVPTR(TILE_RRTILE2005), 0, RRTILE2005__STATICRR },
    { "POPCORN", DVPTR(TILE_POPCORN), 0, POPCORN__STATICRR },
    { "RRTILE2022", DVPTR(TILE_RRTILE2022), 0, RRTILE2022__STATICRR },
    { "LANEPICS", DVPTR(TILE_LANEPICS), 0, LANEPICS__STATICRR },
    { "RRTILE2025", DVPTR(TILE_RRTILE2025), 0, RRTILE2025__STATICRR },
    { "RRTILE2026", DVPTR(TILE_RRTILE2026), 0, RRTILE2026__STATICRR },
    { "RRTILE2027", DVPTR(TILE_RRTILE2027), 0, RRTILE2027__STATICRR },
    { "RRTILE2028", DVPTR(TILE_RRTILE2028), 0, RRTILE2028__STATICRR },
    { "RRTILE2034", DVPTR(TILE_RRTILE2034), 0, RRTILE2034__STATICRR },
    { "RRTILE2050", DVPTR(TILE_RRTILE2050), 0, RRTILE2050__STATICRR },
    { "RRTILE2052", DVPTR(TILE_RRTILE2052), 0, RRTILE2052__STATICRR },
    { "RRTILE2053", DVPTR(TILE_RRTILE2053), 0, RRTILE2053__STATICRR },
    { "RRTILE2056", DVPTR(TILE_RRTILE2056), 0, RRTILE2056__STATICRR },
    { "RRTILE2060", DVPTR(TILE_RRTILE2060), 0, RRTILE2060__STATICRR },
    { "RRTILE2072", DVPTR(TILE_RRTILE2072), 0, RRTILE2072__STATICRR },
    { "RRTILE2074", DVPTR(TILE_RRTILE2074), 0, RRTILE2074__STATICRR },
    { "RRTILE2075", DVPTR(TILE_RRTILE2075), 0, RRTILE2075__STATICRR },
    { "RRTILE2083", DVPTR(TILE_RRTILE2083), 0, RRTILE2083__STATICRR },
    { "RRTILE2097", DVPTR(TILE_RRTILE2097), 0, RRTILE2097__STATICRR },
    { "RRTILE2121", DVPTR(TILE_RRTILE2121), 0, RRTILE2121__STATICRR },
    { "RRTILE2122", DVPTR(TILE_RRTILE2122), 0, RRTILE2122__STATICRR },
    { "RRTILE2123", DVPTR(TILE_RRTILE2123), 0, RRTILE2123__STATICRR },
    { "RRTILE2124", DVPTR(TILE_RRTILE2124), 0, RRTILE2124__STATICRR },
    { "RRTILE2125", DVPTR(TILE_RRTILE2125), 0, RRTILE2125__STATICRR },
    { "RRTILE2126", DVPTR(TILE_RRTILE2126), 0, RRTILE2126__STATICRR },
    { "RRTILE2137", DVPTR(TILE_RRTILE2137), 0, RRTILE2137__STATICRR },
    { "RRTILE2132", DVPTR(TILE_RRTILE2132), 0, RRTILE2132__STATICRR },
    { "RRTILE2136", DVPTR(TILE_RRTILE2136), 0, RRTILE2136__STATICRR },
    { "RRTILE2139", DVPTR(TILE_RRTILE2139), 0, RRTILE2139__STATICRR },
    { "RRTILE2150", DVPTR(TILE_RRTILE2150), 0, RRTILE2150__STATICRR },
    { "RRTILE2151", DVPTR(TILE_RRTILE2151), 0, RRTILE2151__STATICRR },
    { "RRTILE2152", DVPTR(TILE_RRTILE2152), 0, RRTILE2152__STATICRR },
    { "RRTILE2156", DVPTR(TILE_RRTILE2156), 0, RRTILE2156__STATICRR },
    { "RRTILE2157", DVPTR(TILE_RRTILE2157), 0, RRTILE2157__STATICRR },
    { "RRTILE2158", DVPTR(TILE_RRTILE2158), 0, RRTILE2158__STATICRR },
    { "RRTILE2159", DVPTR(TILE_RRTILE2159), 0, RRTILE2159__STATICRR },
    { "RRTILE2160", DVPTR(TILE_RRTILE2160), 0, RRTILE2160__STATICRR },
    { "RRTILE2161", DVPTR(TILE_RRTILE2161), 0, RRTILE2161__STATICRR },
    { "RRTILE2175", DVPTR(TILE_RRTILE2175), 0, RRTILE2175__STATICRR },
    { "RRTILE2176", DVPTR(TILE_RRTILE2176), 0, RRTILE2176__STATICRR },
    { "RRTILE2178", DVPTR(TILE_RRTILE2178), 0, RRTILE2178__STATICRR },
    { "RRTILE2186", DVPTR(TILE_RRTILE2186), 0, RRTILE2186__STATICRR },
    { "RRTILE2214", DVPTR(TILE_RRTILE2214), 0, RRTILE2214__STATICRR },
    { "RRTILE2319", DVPTR(TILE_RRTILE2319), 0, RRTILE2319__STATICRR },
    { "RRTILE2321", DVPTR(TILE_RRTILE2321), 0, RRTILE2321__STATICRR },
    { "RRTILE2326", DVPTR(TILE_RRTILE2326), 0, RRTILE2326__STATICRR },
    { "RRTILE2329", DVPTR(TILE_RRTILE2329), 0, RRTILE2329__STATICRR },
    { "RRTILE2357", DVPTR(TILE_RRTILE2357), 0, RRTILE2357__STATICRR },
    { "RRTILE2382", DVPTR(TILE_RRTILE2382), 0, RRTILE2382__STATICRR },
    { "RRTILE2430", DVPTR(TILE_RRTILE2430), 0, RRTILE2430__STATICRR },
    { "RRTILE2431", DVPTR(TILE_RRTILE2431), 0, RRTILE2431__STATICRR },
    { "RRTILE2432", DVPTR(TILE_RRTILE2432), 0, RRTILE2432__STATICRR },
    { "RRTILE2437", DVPTR(TILE_RRTILE2437), 0, RRTILE2437__STATICRR },
    { "RRTILE2443", DVPTR(TILE_RRTILE2443), 0, RRTILE2443__STATICRR },
    { "RRTILE2445", DVPTR(TILE_RRTILE2445), 0, RRTILE2445__STATICRR },
    { "RRTILE2446", DVPTR(TILE_RRTILE2446), 0, RRTILE2446__STATICRR },
    { "RRTILE2450", DVPTR(TILE_RRTILE2450), 0, RRTILE2450__STATICRR },
    { "RRTILE2451", DVPTR(TILE_RRTILE2451), 0, RRTILE2451__STATICRR },
    { "RRTILE2455", DVPTR(TILE_RRTILE2455), 0, RRTILE2455__STATICRR },
    { "RRTILE2460", DVPTR(TILE_RRTILE2460), 0, RRTILE2460__STATICRR },
    { "RRTILE2465", DVPTR(TILE_RRTILE2465), 0, RRTILE2465__STATICRR },
    { "RRTILE2560", DVPTR(TILE_RRTILE2560), 0, RRTILE2560__STATICRR },
    { "RRTILE2562", DVPTR(TILE_RRTILE2562), 0, RRTILE2562__STATICRR },
    { "RRTILE2564", DVPTR(TILE_RRTILE2564), 0, RRTILE2564__STATICRR },
    { "RRTILE2573", DVPTR(TILE_RRTILE2573), 0, RRTILE2573__STATICRR },
    { "RRTILE2574", DVPTR(TILE_RRTILE2574), 0, RRTILE2574__STATICRR },
    { "RRTILE2577", DVPTR(TILE_RRTILE2577), 0, RRTILE2577__STATICRR },
    { "RRTILE2578", DVPTR(TILE_RRTILE2578), 0, RRTILE2578__STATICRR },
    { "RRTILE2581", DVPTR(TILE_RRTILE2581), 0, RRTILE2581__STATICRR },
    { "RRTILE2583", DVPTR(TILE_RRTILE2583), 0, RRTILE2583__STATICRR },
    { "RRTILE2604", DVPTR(TILE_RRTILE2604), 0, RRTILE2604__STATICRR },
    { "RRTILE2610", DVPTR(TILE_RRTILE2610), 0, RRTILE2610__STATICRR },
    { "RRTILE2613", DVPTR(TILE_RRTILE2613), 0, RRTILE2613__STATICRR },
    { "RRTILE2621", DVPTR(TILE_RRTILE2621), 0, RRTILE2621__STATICRR },
    { "RRTILE2622", DVPTR(TILE_RRTILE2622), 0, RRTILE2622__STATICRR },
    { "RRTILE2636", DVPTR(TILE_RRTILE2636), 0, RRTILE2636__STATICRR },
    { "RRTILE2637", DVPTR(TILE_RRTILE2637), 0, RRTILE2637__STATICRR },
    { "RRTILE2654", DVPTR(TILE_RRTILE2654), 0, RRTILE2654__STATICRR },
    { "RRTILE2656", DVPTR(TILE_RRTILE2656), 0, RRTILE2656__STATICRR },
    { "RRTILE2676", DVPTR(TILE_RRTILE2676), 0, RRTILE2676__STATICRR },
    { "RRTILE2689", DVPTR(TILE_RRTILE2689), 0, RRTILE2689__STATICRR },
    { "RRTILE2697", DVPTR(TILE_RRTILE2697), 0, RRTILE2697__STATICRR },
    { "RRTILE2702", DVPTR(TILE_RRTILE2702), 0, RRTILE2702__STATICRR },
    { "RRTILE2707", DVPTR(TILE_RRTILE2707), 0, RRTILE2707__STATICRR },
    { "RRTILE2732", DVPTR(TILE_RRTILE2732), 0, RRTILE2732__STATICRR },
    { "RRTILE2030", DVPTR(TILE_RRTILE2030), 0, RRTILE2030__STATICRR },
    { "RRTILE2831", DVPTR(TILE_RRTILE2831), 0, RRTILE2831__STATICRR },
    { "RRTILE2832", DVPTR(TILE_RRTILE2832), 0, RRTILE2832__STATICRR },
    { "RRTILE2842", DVPTR(TILE_RRTILE2842), 0, RRTILE2842__STATICRR },
    { "RRTILE2859", DVPTR(TILE_RRTILE2859), 0, RRTILE2859__STATICRR },
    { "RRTILE2876", DVPTR(TILE_RRTILE2876), 0, RRTILE2876__STATICRR },
    { "RRTILE2878", DVPTR(TILE_RRTILE2878), 0, RRTILE2878__STATICRR },
    { "RRTILE2879", DVPTR(TILE_RRTILE2879), 0, RRTILE2879__STATICRR },
    { "RRTILE2893", DVPTR(TILE_RRTILE2893), 0, RRTILE2893__STATICRR },
    { "RRTILE2894", DVPTR(TILE_RRTILE2894), 0, RRTILE2894__STATICRR },
    { "RRTILE2898", DVPTR(TILE_RRTILE2898), 0, RRTILE2898__STATICRR },
    { "RRTILE2899", DVPTR(TILE_RRTILE2899), 0, RRTILE2899__STATICRR },
    { "RRTILE2915", DVPTR(TILE_RRTILE2915), 0, RRTILE2915__STATICRR },
    { "RRTILE2940", DVPTR(TILE_RRTILE2940), 0, RRTILE2940__STATICRR },
    { "RRTILE2944", DVPTR(TILE_RRTILE2944), 0, RRTILE2944__STATICRR },
    { "RRTILE2945", DVPTR(TILE_RRTILE2945), 0, RRTILE2945__STATICRR },
    { "RRTILE2946", DVPTR(TILE_RRTILE2946), 0, RRTILE2946__STATICRR },
    { "RRTILE2947", DVPTR(TILE_RRTILE2947), 0, RRTILE2947__STATICRR },
    { "RRTILE2948", DVPTR(TILE_RRTILE2948), 0, RRTILE2948__STATICRR },
    { "RRTILE2949", DVPTR(TILE_RRTILE2949), 0, RRTILE2949__STATICRR },
    { "RRTILE2961", DVPTR(TILE_RRTILE2961), 0, RRTILE2961__STATICRR },
    { "RRTILE2970", DVPTR(TILE_RRTILE2970), 0, RRTILE2970__STATICRR },
    { "RRTILE2977", DVPTR(TILE_RRTILE2977), 0, RRTILE2977__STATICRR },
    { "RRTILE2978", DVPTR(TILE_RRTILE2978), 0, RRTILE2978__STATICRR },
    { "RRTILE2990", DVPTR(TILE_RRTILE2990), 0, RRTILE2990__STATICRR },
    { "RRTILE3073", DVPTR(TILE_RRTILE3073), 0, RRTILE3073__STATICRR },
    { "RRTILE3083", DVPTR(TILE_RRTILE3083), 0, RRTILE3083__STATICRR },
    { "RRTILE3100", DVPTR(TILE_RRTILE3100), 0, RRTILE3100__STATICRR },
    { "RRTILE3114", DVPTR(TILE_RRTILE3114), 0, RRTILE3114__STATICRR },
    { "RRTILE3115", DVPTR(TILE_RRTILE3115), 0, RRTILE3115__STATICRR },
    { "RRTILE3116", DVPTR(TILE_RRTILE3116), 0, RRTILE3116__STATICRR },
    { "RRTILE3117", DVPTR(TILE_RRTILE3117), 0, RRTILE3117__STATICRR },
    { "RRTILE3120", DVPTR(TILE_RRTILE3120), 0, RRTILE3120__STATICRR },
    { "RRTILE3121", DVPTR(TILE_RRTILE3121), 0, RRTILE3121__STATICRR },
    { "RRTILE3122", DVPTR(TILE_RRTILE3122), 0, RRTILE3122__STATICRR },
    { "RRTILE3123", DVPTR(TILE_RRTILE3123), 0, RRTILE3123__STATICRR },
    { "RRTILE3124", DVPTR(TILE_RRTILE3124), 0, RRTILE3124__STATICRR },
    { "RRTILE3132", DVPTR(TILE_RRTILE3132), 0, RRTILE3132__STATICRR },
    { "RRTILE3139", DVPTR(TILE_RRTILE3139), 0, RRTILE3139__STATICRR },
    { "RRTILE3144", DVPTR(TILE_RRTILE3144), 0, RRTILE3144__STATICRR },
    { "RRTILE3152", DVPTR(TILE_RRTILE3152), 0, RRTILE3152__STATICRR },
    { "RRTILE3153", DVPTR(TILE_RRTILE3153), 0, RRTILE3153__STATICRR },
    { "RRTILE3155", DVPTR(TILE_RRTILE3155), 0, RRTILE3155__STATICRR },
    { "RRTILE3171", DVPTR(TILE_RRTILE3171), 0, RRTILE3171__STATICRR },
    { "RRTILE3172", DVPTR(TILE_RRTILE3172), 0, RRTILE3172__STATICRR },
    { "RRTILE3190", DVPTR(TILE_RRTILE3190), 0, RRTILE3190__STATICRR },
    { "RRTILE3191", DVPTR(TILE_RRTILE3191), 0, RRTILE3191__STATICRR },
    { "RRTILE3192", DVPTR(TILE_RRTILE3192), 0, RRTILE3192__STATICRR },
    { "RRTILE3195", DVPTR(TILE_RRTILE3195), 0, RRTILE3195__STATICRR },
    { "RRTILE3200", DVPTR(TILE_RRTILE3200), 0, RRTILE3200__STATICRR },
    { "RRTILE3201", DVPTR(TILE_RRTILE3201), 0, RRTILE3201__STATICRR },
    { "RRTILE3202", DVPTR(TILE_RRTILE3202), 0, RRTILE3202__STATICRR },
    { "RRTILE3203", DVPTR(TILE_RRTILE3203), 0, RRTILE3203__STATICRR },
    { "RRTILE3204", DVPTR(TILE_RRTILE3204), 0, RRTILE3204__STATICRR },
    { "RRTILE3205", DVPTR(TILE_RRTILE3205), 0, RRTILE3205__STATICRR },
    { "RRTILE3206", DVPTR(TILE_RRTILE3206), 0, RRTILE3206__STATICRR },
    { "RRTILE3207", DVPTR(TILE_RRTILE3207), 0, RRTILE3207__STATICRR },
    { "RRTILE3208", DVPTR(TILE_RRTILE3208), 0, RRTILE3208__STATICRR },
    { "RRTILE3209", DVPTR(TILE_RRTILE3209), 0, RRTILE3209__STATICRR },
    { "RRTILE3216", DVPTR(TILE_RRTILE3216), 0, RRTILE3216__STATICRR },
    { "RRTILE3218", DVPTR(TILE_RRTILE3218), 0, RRTILE3218__STATICRR },
    { "RRTILE3219", DVPTR(TILE_RRTILE3219), 0, RRTILE3219__STATICRR },
    { "RRTILE3232", DVPTR(TILE_RRTILE3232), 0, RRTILE3232__STATICRR },
    { "SHOTGUNSHELLS", DVPTR(TILE_SHOTGUNSHELLS), 0, SHOTGUNSHELLS__STATICRR },
    { "CIRCLESTUCK", DVPTR(TILE_CIRCLESTUCK), 0, CIRCLESTUCK__STATICRR },
    { "RRTILE3410", DVPTR(TILE_RRTILE3410), 0, RRTILE3410__STATICRR },
    { "LUMBERBLADE", DVPTR(TILE_LUMBERBLADE), 0, LUMBERBLADE__STATICRR },
    { "BOWLINGBALLH", DVPTR(TILE_BOWLINGBALLH), 0, BOWLINGBALLH__STATICRR },
    { "BOWLINGBALL", DVPTR(TILE_BOWLINGBALL), 0, BOWLINGBALL__STATICRR },
    { "BOWLINGBALLSPRITE", DVPTR(TILE_BOWLINGBALLSPRITE), 0, BOWLINGBALLSPRITE__STATICRR },
    { "POWDERH", DVPTR(TILE_POWDERH), 0, POWDERH__STATICRR },
    { "RRTILE3440", DVPTR(TILE_RRTILE3440), 0, RRTILE3440__STATICRR },
    { "RRTILE3462", DVPTR(TILE_RRTILE3462), 0, RRTILE3462__STATICRR },
    { "OWHIP", DVPTR(TILE_OWHIP), 0, OWHIP__STATICRR },
    { "UWHIP", DVPTR(TILE_UWHIP), 0, UWHIP__STATICRR },
    { "RPGGUN2", DVPTR(TILE_RPGGUN2), 0, RPGGUN2__STATICRR },
    { "RRTILE3497", DVPTR(TILE_RRTILE3497), 0, RRTILE3497__STATICRR },
    { "RRTILE3498", DVPTR(TILE_RRTILE3498), 0, RRTILE3498__STATICRR },
    { "RRTILE3499", DVPTR(TILE_RRTILE3499), 0, RRTILE3499__STATICRR },
    { "RRTILE3500", DVPTR(TILE_RRTILE3500), 0, RRTILE3500__STATICRR },
    { "SLINGBLADE", DVPTR(TILE_SLINGBLADE), 0, SLINGBLADE__STATICRR },
    { "RRTILE3584", DVPTR(TILE_RRTILE3584), 0, RRTILE3584__STATICRR },
    { "RRTILE3586", DVPTR(TILE_RRTILE3586), 0, RRTILE3586__STATICRR },
    { "RRTILE3587", DVPTR(TILE_RRTILE3587), 0, RRTILE3587__STATICRR },
    { "RRTILE3600", DVPTR(TILE_RRTILE3600), 0, RRTILE3600__STATICRR },
    { "RRTILE3631", DVPTR(TILE_RRTILE3631), 0, RRTILE3631__STATICRR },
    { "RRTILE3635", DVPTR(TILE_RRTILE3635), 0, RRTILE3635__STATICRR },
    { "RRTILE3637", DVPTR(TILE_RRTILE3637), 0, RRTILE3637__STATICRR },
    { "RRTILE3643", DVPTR(TILE_RRTILE3643), 0, RRTILE3643__STATICRR },
    { "RRTILE3647", DVPTR(TILE_RRTILE3647), 0, RRTILE3647__STATICRR },
    { "RRTILE3652", DVPTR(TILE_RRTILE3652), 0, RRTILE3652__STATICRR },
    { "RRTILE3653", DVPTR(TILE_RRTILE3653), 0, RRTILE3653__STATICRR },
    { "RRTILE3668", DVPTR(TILE_RRTILE3668), 0, RRTILE3668__STATICRR },
    { "RRTILE3671", DVPTR(TILE_RRTILE3671), 0, RRTILE3671__STATICRR },
    { "RRTILE3673", DVPTR(TILE_RRTILE3673), 0, RRTILE3673__STATICRR },
    { "RRTILE3677", DVPTR(TILE_RRTILE3677), 0, RRTILE3677__STATICRR },
    { "RRTILE3684", DVPTR(TILE_RRTILE3684), 0, RRTILE3684__STATICRR },
    { "RRTILE3708", DVPTR(TILE_RRTILE3708), 0, RRTILE3708__STATICRR },
    { "RRTILE3714", DVPTR(TILE_RRTILE3714), 0, RRTILE3714__STATICRR },
    { "RRTILE3716", DVPTR(TILE_RRTILE3716), 0, RRTILE3716__STATICRR },
    { "RRTILE3720", DVPTR(TILE_RRTILE3720), 0, RRTILE3720__STATICRR },
    { "RRTILE3723", DVPTR(TILE_RRTILE3723), 0, RRTILE3723__STATICRR },
    { "RRTILE3725", DVPTR(TILE_RRTILE3725), 0, RRTILE3725__STATICRR },
    { "RRTILE3737", DVPTR(TILE_RRTILE3737), 0, RRTILE3737__STATICRR },
    { "RRTILE3754", DVPTR(TILE_RRTILE3754), 0, RRTILE3754__STATICRR },
    { "RRTILE3762", DVPTR(TILE_RRTILE3762), 0, RRTILE3762__STATICRR },
    { "RRTILE3763", DVPTR(TILE_RRTILE3763), 0, RRTILE3763__STATICRR },
    { "RRTILE3764", DVPTR(TILE_RRTILE3764), 0, RRTILE3764__STATICRR },
    { "RRTILE3765", DVPTR(TILE_RRTILE3765), 0, RRTILE3765__STATICRR },
    { "RRTILE3767", DVPTR(TILE_RRTILE3767), 0, RRTILE3767__STATICRR },
    { "RRTILE3773", DVPTR(TILE_RRTILE3773), 0, RRTILE3773__STATICRR },
    { "RRTILE3774", DVPTR(TILE_RRTILE3774), 0, RRTILE3774__STATICRR },
    { "RRTILE3793", DVPTR(TILE_RRTILE3793), 0, RRTILE3793__STATICRR },
    { "RRTILE3795", DVPTR(TILE_RRTILE3795), 0, RRTILE3795__STATICRR },
    { "RRTILE3804", DVPTR(TILE_RRTILE3804), 0, RRTILE3804__STATICRR },
    { "RRTILE3814", DVPTR(TILE_RRTILE3814), 0, RRTILE3814__STATICRR },
    { "RRTILE3815", DVPTR(TILE_RRTILE3815), 0, RRTILE3815__STATICRR },
    { "RRTILE3819", DVPTR(TILE_RRTILE3819), 0, RRTILE3819__STATICRR },
    { "RRTILE3827", DVPTR(TILE_RRTILE3827), 0, RRTILE3827__STATICRR },
    { "RRTILE3837", DVPTR(TILE_RRTILE3837), 0, RRTILE3837__STATICRR },
    { "RRTILE5014", DVPTR(TILE_RRTILE5014), 0, RRTILE5014__STATICRR },
    { "RRTILE5016", DVPTR(TILE_RRTILE5016), 0, RRTILE5016__STATICRR },
    { "RRTILE5017", DVPTR(TILE_RRTILE5017), 0, RRTILE5017__STATICRR },
    { "RRTILE5018", DVPTR(TILE_RRTILE5018), 0, RRTILE5018__STATICRR },
    { "RRTILE5019", DVPTR(TILE_RRTILE5019), 0, RRTILE5019__STATICRR },
    { "RRTILE5020", DVPTR(TILE_RRTILE5020), 0, RRTILE5020__STATICRR },
    { "RRTILE5021", DVPTR(TILE_RRTILE5021), 0, RRTILE5021__STATICRR },
    { "RRTILE5022", DVPTR(TILE_RRTILE5022), 0, RRTILE5022__STATICRR },
    { "RRTILE5023", DVPTR(TILE_RRTILE5023), 0, RRTILE5023__STATICRR },
    { "RRTILE5024", DVPTR(TILE_RRTILE5024), 0, RRTILE5024__STATICRR },
    { "RRTILE5025", DVPTR(TILE_RRTILE5025), 0, RRTILE5025__STATICRR },
    { "RRTILE5026", DVPTR(TILE_RRTILE5026), 0, RRTILE5026__STATICRR },
    { "RRTILE5027", DVPTR(TILE_RRTILE5027), 0, RRTILE5027__STATICRR },
    { "RRTILE5029", DVPTR(TILE_RRTILE5029), 0, RRTILE5029__STATICRR },
    { "RRTILE5030", DVPTR(TILE_RRTILE5030), 0, RRTILE5030__STATICRR },
    { "RRTILE5031", DVPTR(TILE_RRTILE5031), 0, RRTILE5031__STATICRR },
    { "RRTILE5032", DVPTR(TILE_RRTILE5032), 0, RRTILE5032__STATICRR },
    { "RRTILE5033", DVPTR(TILE_RRTILE5033), 0, RRTILE5033__STATICRR },
    { "RRTILE5034", DVPTR(TILE_RRTILE5034), 0, RRTILE5034__STATICRR },
    { "RRTILE5035", DVPTR(TILE_RRTILE5035), 0, RRTILE5035__STATICRR },
    { "RRTILE5036", DVPTR(TILE_RRTILE5036), 0, RRTILE5036__STATICRR },
    { "RRTILE5037", DVPTR(TILE_RRTILE5037), 0, RRTILE5037__STATICRR },
    { "RRTILE5038", DVPTR(TILE_RRTILE5038), 0, RRTILE5038__STATICRR },
    { "RRTILE5039", DVPTR(TILE_RRTILE5039), 0, RRTILE5039__STATICRR },
    { "RRTILE5040", DVPTR(TILE_RRTILE5040), 0, RRTILE5040__STATICRR },
    { "RRTILE5041", DVPTR(TILE_RRTILE5041), 0, RRTILE5041__STATICRR },
    { "RRTILE5043", DVPTR(TILE_RRTILE5043), 0, RRTILE5043__STATICRR },
    { "RRTILE5044", DVPTR(TILE_RRTILE5044), 0, RRTILE5044__STATICRR },
    { "RRTILE5045", DVPTR(TILE_RRTILE5045), 0, RRTILE5045__STATICRR },
    { "RRTILE5046", DVPTR(TILE_RRTILE5046), 0, RRTILE5046__STATICRR },
    { "RRTILE5047", DVPTR(TILE_RRTILE5047), 0, RRTILE5047__STATICRR },
    { "RRTILE5048", DVPTR(TILE_RRTILE5048), 0, RRTILE5048__STATICRR },
    { "RRTILE5049", DVPTR(TILE_RRTILE5049), 0, RRTILE5049__STATICRR },
    { "RRTILE5050", DVPTR(TILE_RRTILE5050), 0, RRTILE5050__STATICRR },
    { "RRTILE5051", DVPTR(TILE_RRTILE5051), 0, RRTILE5051__STATICRR },
    { "RRTILE5052", DVPTR(TILE_RRTILE5052), 0, RRTILE5052__STATICRR },
    { "RRTILE5053", DVPTR(TILE_RRTILE5053), 0, RRTILE5053__STATICRR },
    { "RRTILE5054", DVPTR(TILE_RRTILE5054), 0, RRTILE5054__STATICRR },
    { "RRTILE5055", DVPTR(TILE_RRTILE5055), 0, RRTILE5055__STATICRR },
    { "RRTILE5056", DVPTR(TILE_RRTILE5056), 0, RRTILE5056__STATICRR },
    { "RRTILE5057", DVPTR(TILE_RRTILE5057), 0, RRTILE5057__STATICRR },
    { "RRTILE5058", DVPTR(TILE_RRTILE5058), 0, RRTILE5058__STATICRR },
    { "RRTILE5059", DVPTR(TILE_RRTILE5059), 0, RRTILE5059__STATICRR },
    { "RRTILE5061", DVPTR(TILE_RRTILE5061), 0, RRTILE5061__STATICRR },
    { "RRTILE5062", DVPTR(TILE_RRTILE5062), 0, RRTILE5062__STATICRR },
    { "RRTILE5063", DVPTR(TILE_RRTILE5063), 0, RRTILE5063__STATICRR },
    { "RRTILE5064", DVPTR(TILE_RRTILE5064), 0, RRTILE5064__STATICRR },
    { "RRTILE5065", DVPTR(TILE_RRTILE5065), 0, RRTILE5065__STATICRR },
    { "RRTILE5066", DVPTR(TILE_RRTILE5066), 0, RRTILE5066__STATICRR },
    { "RRTILE5067", DVPTR(TILE_RRTILE5067), 0, RRTILE5067__STATICRR },
    { "RRTILE5068", DVPTR(TILE_RRTILE5068), 0, RRTILE5068__STATICRR },
    { "RRTILE5069", DVPTR(TILE_RRTILE5069), 0, RRTILE5069__STATICRR },
    { "RRTILE5070", DVPTR(TILE_RRTILE5070), 0, RRTILE5070__STATICRR },
    { "RRTILE5071", DVPTR(TILE_RRTILE5071), 0, RRTILE5071__STATICRR },
    { "RRTILE5072", DVPTR(TILE_RRTILE5072), 0, RRTILE5072__STATICRR },
    { "RRTILE5073", DVPTR(TILE_RRTILE5073), 0, RRTILE5073__STATICRR },
    { "RRTILE5074", DVPTR(TILE_RRTILE5074), 0, RRTILE5074__STATICRR },
    { "RRTILE5075", DVPTR(TILE_RRTILE5075), 0, RRTILE5075__STATICRR },
    { "RRTILE5076", DVPTR(TILE_RRTILE5076), 0, RRTILE5076__STATICRR },
    { "RRTILE5077", DVPTR(TILE_RRTILE5077), 0, RRTILE5077__STATICRR },
    { "RRTILE5078", DVPTR(TILE_RRTILE5078), 0, RRTILE5078__STATICRR },
    { "RRTILE5079", DVPTR(TILE_RRTILE5079), 0, RRTILE5079__STATICRR },
    { "RRTILE5080", DVPTR(TILE_RRTILE5080), 0, RRTILE5080__STATICRR },
    { "RRTILE5081", DVPTR(TILE_RRTILE5081), 0, RRTILE5081__STATICRR },
    { "RRTILE5082", DVPTR(TILE_RRTILE5082), 0, RRTILE5082__STATICRR },
    { "RRTILE5083", DVPTR(TILE_RRTILE5083), 0, RRTILE5083__STATICRR },
    { "RRTILE5084", DVPTR(TILE_RRTILE5084), 0, RRTILE5084__STATICRR },
    { "RRTILE5085", DVPTR(TILE_RRTILE5085), 0, RRTILE5085__STATICRR },
    { "RRTILE5086", DVPTR(TILE_RRTILE5086), 0, RRTILE5086__STATICRR },
    { "RRTILE5087", DVPTR(TILE_RRTILE5087), 0, RRTILE5087__STATICRR },
    { "RRTILE5088", DVPTR(TILE_RRTILE5088), 0, RRTILE5088__STATICRR },
    { "RRTILE5090", DVPTR(TILE_RRTILE5090), 0, RRTILE5090__STATICRR },
    { "RRTILE6144", DVPTR(TILE_RRTILE6144), 0, RRTILE6144__STATICRR },
    { "RRTILE7010", DVPTR(TILE_RRTILE7110), 0, RRTILE7110__STATICRR },
    { "RRTILE7011", DVPTR(TILE_RRTILE7111), 0, RRTILE7111__STATICRR },
    { "RRTILE7012", DVPTR(TILE_RRTILE7112), 0, RRTILE7112__STATICRR },
    { "RRTILE7013", DVPTR(TILE_RRTILE7113), 0, RRTILE7113__STATICRR },
    { "MOTOGUN", DVPTR(TILE_MOTOGUN), 0, MOTOGUN__STATICRR },
    { "RRTILE7169", DVPTR(TILE_RRTILE7169), 0, RRTILE7169__STATICRR },
    { "MOTOHIT", DVPTR(TILE_MOTOHIT), 0, MOTOHIT__STATICRR },
    { "BOATHIT", DVPTR(TILE_BOATHIT), 0, BOATHIT__STATICRR },
    { "RRTILE7184", DVPTR(TILE_RRTILE7184), 0, RRTILE7184__STATICRR },
    { "RRTILE7190", DVPTR(TILE_RRTILE7190), 0, RRTILE7190__STATICRR },
    { "RRTILE7191", DVPTR(TILE_RRTILE7191), 0, RRTILE7191__STATICRR },
    { "RRTILE7213", DVPTR(TILE_RRTILE7213), 0, RRTILE7213__STATICRR },
    { "RRTILE7219", DVPTR(TILE_RRTILE7219), 0, RRTILE7219__STATICRR },
    { "EMPTYBIKE", DVPTR(TILE_EMPTYBIKE), 0, EMPTYBIKE__STATICRR },
    { "EMPTYBOAT", DVPTR(TILE_EMPTYBOAT), 0, EMPTYBOAT__STATICRR },
    { "RRTILE7424", DVPTR(TILE_RRTILE7424), 0, RRTILE7424__STATICRR },
    { "RRTILE7430", DVPTR(TILE_RRTILE7430), 0, RRTILE7430__STATICRR },
    { "RRTILE7433", DVPTR(TILE_RRTILE7433), 0, RRTILE7433__STATICRR },
    { "RRTILE7441", DVPTR(TILE_RRTILE7441), 0, RRTILE7441__STATICRR },
    { "RRTILE7547", DVPTR(TILE_RRTILE7547), 0, RRTILE7547__STATICRR },
    { "RRTILE7467", DVPTR(TILE_RRTILE7467), 0, RRTILE7467__STATICRR },
    { "RRTILE7469", DVPTR(TILE_RRTILE7469), 0, RRTILE7469__STATICRR },
    { "RRTILE7470", DVPTR(TILE_RRTILE7470), 0, RRTILE7470__STATICRR },
    { "RRTILE7475", DVPTR(TILE_RRTILE7475), 0, RRTILE7475__STATICRR },
    { "RRTILE7478", DVPTR(TILE_RRTILE7478), 0, RRTILE7478__STATICRR },
    { "RRTILE7505", DVPTR(TILE_RRTILE7505), 0, RRTILE7505__STATICRR },
    { "RRTILE7506", DVPTR(TILE_RRTILE7506), 0, RRTILE7506__STATICRR },
    { "RRTILE7534", DVPTR(TILE_RRTILE7534), 0, RRTILE7534__STATICRR },
    { "RRTILE7540", DVPTR(TILE_RRTILE7540), 0, RRTILE7540__STATICRR },
    { "RRTILE7533", DVPTR(TILE_RRTILE7533), 0, RRTILE7533__STATICRR },
    { "RRTILE7545", DVPTR(TILE_RRTILE7545), 0, RRTILE7545__STATICRR },
    { "RRTILE7552", DVPTR(TILE_RRTILE7552), 0, RRTILE7552__STATICRR },
    { "RRTILE7553", DVPTR(TILE_RRTILE7553), 0, RRTILE7553__STATICRR },
    { "RRTILE7554", DVPTR(TILE_RRTILE7554), 0, RRTILE7554__STATICRR },
    { "RRTILE7555", DVPTR(TILE_RRTILE7555), 0, RRTILE7555__STATICRR },
    { "RRTILE7557", DVPTR(TILE_RRTILE7557), 0, RRTILE7557__STATICRR },
    { "RRTILE7558", DVPTR(TILE_RRTILE7558), 0, RRTILE7558__STATICRR },
    { "RRTILE7559", DVPTR(TILE_RRTILE7559), 0, RRTILE7559__STATICRR },
    { "RRTILE7561", DVPTR(TILE_RRTILE7561), 0, RRTILE7561__STATICRR },
    { "RRTILE7566", DVPTR(TILE_RRTILE7566), 0, RRTILE7566__STATICRR },
    { "RRTILE7568", DVPTR(TILE_RRTILE7568), 0, RRTILE7568__STATICRR },
    { "RRTILE7574", DVPTR(TILE_RRTILE7574), 0, RRTILE7574__STATICRR },
    { "RRTILE7575", DVPTR(TILE_RRTILE7575), 0, RRTILE7575__STATICRR },
    { "RRTILE7576", DVPTR(TILE_RRTILE7576), 0, RRTILE7576__STATICRR },
    { "RRTILE7578", DVPTR(TILE_RRTILE7578), 0, RRTILE7578__STATICRR },
    { "RRTILE7579", DVPTR(TILE_RRTILE7579), 0, RRTILE7579__STATICRR },
    { "RRTILE7580", DVPTR(TILE_RRTILE7580), 0, RRTILE7580__STATICRR },
    { "RRTILE7595", DVPTR(TILE_RRTILE7595), 0, RRTILE7595__STATICRR },
    { "RRTILE7629", DVPTR(TILE_RRTILE7629), 0, RRTILE7629__STATICRR },
    { "RRTILE7636", DVPTR(TILE_RRTILE7636), 0, RRTILE7636__STATICRR },
    { "RRTILE7638", DVPTR(TILE_RRTILE7638), 0, RRTILE7638__STATICRR },
    { "RRTILE7640", DVPTR(TILE_RRTILE7640), 0, RRTILE7640__STATICRR },
    { "RRTILE7644", DVPTR(TILE_RRTILE7644), 0, RRTILE7644__STATICRR },
    { "RRTILE7646", DVPTR(TILE_RRTILE7646), 0, RRTILE7646__STATICRR },
    { "RRTILE7648", DVPTR(TILE_RRTILE7648), 0, RRTILE7648__STATICRR },
    { "RRTILE7650", DVPTR(TILE_RRTILE7650), 0, RRTILE7650__STATICRR },
    { "RRTILE7653", DVPTR(TILE_RRTILE7653), 0, RRTILE7653__STATICRR },
    { "RRTILE7655", DVPTR(TILE_RRTILE7655), 0, RRTILE7655__STATICRR },
    { "RRTILE7657", DVPTR(TILE_RRTILE7657), 0, RRTILE7657__STATICRR },
    { "RRTILE7659", DVPTR(TILE_RRTILE7659), 0, RRTILE7659__STATICRR },
    { "RRTILE7691", DVPTR(TILE_RRTILE7691), 0, RRTILE7691__STATICRR },
    { "RRTILE7694", DVPTR(TILE_RRTILE7694), 0, RRTILE7694__STATICRR },
    { "RRTILE7696", DVPTR(TILE_RRTILE7696), 0, RRTILE7696__STATICRR },
    { "RRTILE7697", DVPTR(TILE_RRTILE7697), 0, RRTILE7697__STATICRR },
    { "RRTILE7700", DVPTR(TILE_RRTILE7700), 0, RRTILE7700__STATICRR },
    { "RRTILE7702", DVPTR(TILE_RRTILE7702), 0, RRTILE7702__STATICRR },
    { "RRTILE7704", DVPTR(TILE_RRTILE7704), 0, RRTILE7704__STATICRR },
    { "RRTILE7705", DVPTR(TILE_RRTILE7705), 0, RRTILE7705__STATICRR },
    { "RRTILE7711", DVPTR(TILE_RRTILE7711), 0, RRTILE7711__STATICRR },
    { "RRTILE7716", DVPTR(TILE_RRTILE7716), 0, RRTILE7716__STATICRR },
    { "RRTILE7756", DVPTR(TILE_RRTILE7756), 0, RRTILE7756__STATICRR },
    { "RRTILE7768", DVPTR(TILE_RRTILE7768), 0, RRTILE7768__STATICRR },
    { "RRTILE7806", DVPTR(TILE_RRTILE7806), 0, RRTILE7806__STATICRR },
    { "RRTILE7820", DVPTR(TILE_RRTILE7820), 0, RRTILE7820__STATICRR },
    { "RRTILE7859", DVPTR(TILE_RRTILE7859), 0, RRTILE7859__STATICRR },
    { "RRTILE7870", DVPTR(TILE_RRTILE7870), 0, RRTILE7870__STATICRR },
    { "RRTILE7873", DVPTR(TILE_RRTILE7873), 0, RRTILE7873__STATICRR },
    { "RRTILE7875", DVPTR(TILE_RRTILE7875), 0, RRTILE7875__STATICRR },
    { "RRTILE7876", DVPTR(TILE_RRTILE7876), 0, RRTILE7876__STATICRR },
    { "RRTILE7879", DVPTR(TILE_RRTILE7879), 0, RRTILE7879__STATICRR },
    { "RRTILE7881", DVPTR(TILE_RRTILE7881), 0, RRTILE7881__STATICRR },
    { "RRTILE7883", DVPTR(TILE_RRTILE7883), 0, RRTILE7883__STATICRR },
    { "RRTILE7885", DVPTR(TILE_RRTILE7885), 0, RRTILE7885__STATICRR },
    { "RRTILE7886", DVPTR(TILE_RRTILE7886), 0, RRTILE7886__STATICRR },
    { "RRTILE7887", DVPTR(TILE_RRTILE7887), 0, RRTILE7887__STATICRR },
    { "RRTILE7888", DVPTR(TILE_RRTILE7888), 0, RRTILE7888__STATICRR },
    { "RRTILE7889", DVPTR(TILE_RRTILE7889), 0, RRTILE7889__STATICRR },
    { "RRTILE7890", DVPTR(TILE_RRTILE7890), 0, RRTILE7890__STATICRR },
    { "RRTILE7900", DVPTR(TILE_RRTILE7900), 0, RRTILE7900__STATICRR },
    { "RRTILE7901", DVPTR(TILE_RRTILE7901), 0, RRTILE7901__STATICRR },
    { "RRTILE7906", DVPTR(TILE_RRTILE7906), 0, RRTILE7906__STATICRR },
    { "RRTILE7912", DVPTR(TILE_RRTILE7912), 0, RRTILE7912__STATICRR },
    { "RRTILE7913", DVPTR(TILE_RRTILE7913), 0, RRTILE7913__STATICRR },
    { "RRTILE7936", DVPTR(TILE_RRTILE7936), 0, RRTILE7936__STATICRR },
    { "RRTILE8047", DVPTR(TILE_RRTILE8047), 0, RRTILE8047__STATICRR },
    { "MULTISWITCH2", DVPTR(TILE_MULTISWITCH2), 0, MULTISWITCH2__STATICRR },
    { "RRTILE8059", DVPTR(TILE_RRTILE8059), 0, RRTILE8059__STATICRR },
    { "RRTILE8060", DVPTR(TILE_RRTILE8060), 0, RRTILE8060__STATICRR },
    { "RRTILE8063", DVPTR(TILE_RRTILE8063), 0, RRTILE8063__STATICRR },
    { "RRTILE8067", DVPTR(TILE_RRTILE8067), 0, RRTILE8067__STATICRR },
    { "RRTILE8076", DVPTR(TILE_RRTILE8076), 0, RRTILE8076__STATICRR },
    { "RRTILE8094", DVPTR(TILE_RRTILE8094), 0, RRTILE8094__STATICRR },
    { "RRTILE8096", DVPTR(TILE_RRTILE8096), 0, RRTILE8096__STATICRR },
    { "RRTILE8099", DVPTR(TILE_RRTILE8099), 0, RRTILE8099__STATICRR },
    { "RRTILE8106", DVPTR(TILE_RRTILE8106), 0, RRTILE8106__STATICRR },
    { "RRTILE8162", DVPTR(TILE_RRTILE8162), 0, RRTILE8162__STATICRR },
    { "RRTILE8163", DVPTR(TILE_RRTILE8163), 0, RRTILE8163__STATICRR },
    { "RRTILE8164", DVPTR(TILE_RRTILE8164), 0, RRTILE8164__STATICRR },
    { "RRTILE8165", DVPTR(TILE_RRTILE8165), 0, RRTILE8165__STATICRR },
    { "RRTILE8166", DVPTR(TILE_RRTILE8166), 0, RRTILE8166__STATICRR },
    { "RRTILE8167", DVPTR(TILE_RRTILE8167), 0, RRTILE8167__STATICRR },
    { "RRTILE8168", DVPTR(TILE_RRTILE8168), 0, RRTILE8168__STATICRR },
    { "RRTILE8192", DVPTR(TILE_RRTILE8192), 0, RRTILE8192__STATICRR },
    { "RRTILE8193", DVPTR(TILE_RRTILE8193), 0, RRTILE8193__STATICRR },
    { "RRTILE8215", DVPTR(TILE_RRTILE8215), 0, RRTILE8215__STATICRR },
    { "RRTILE8216", DVPTR(TILE_RRTILE8216), 0, RRTILE8216__STATICRR },
    { "RRTILE8217", DVPTR(TILE_RRTILE8217), 0, RRTILE8217__STATICRR },
    { "RRTILE8218", DVPTR(TILE_RRTILE8218), 0, RRTILE8218__STATICRR },
    { "RRTILE8220", DVPTR(TILE_RRTILE8220), 0, RRTILE8220__STATICRR },
    { "RRTILE8221", DVPTR(TILE_RRTILE8221), 0, RRTILE8221__STATICRR },
    { "RRTILE8222", DVPTR(TILE_RRTILE8222), 0, RRTILE8222__STATICRR },
    { "RRTILE8223", DVPTR(TILE_RRTILE8223), 0, RRTILE8223__STATICRR },
    { "RRTILE8224", DVPTR(TILE_RRTILE8224), 0, RRTILE8224__STATICRR },
    { "RRTILE8227", DVPTR(TILE_RRTILE8227), 0, RRTILE8227__STATICRR },
    { "RRTILE8312", DVPTR(TILE_RRTILE8312), 0, RRTILE8312__STATICRR },
    { "RRTILE8370", DVPTR(TILE_RRTILE8370), 0, RRTILE8370__STATICRR },
    { "RRTILE8371", DVPTR(TILE_RRTILE8371), 0, RRTILE8371__STATICRR },
    { "RRTILE8372", DVPTR(TILE_RRTILE8372), 0, RRTILE8372__STATICRR },
    { "RRTILE8373", DVPTR(TILE_RRTILE8373), 0, RRTILE8373__STATICRR },
    { "RRTILE8379", DVPTR(TILE_RRTILE8379), 0, RRTILE8379__STATICRR },
    { "RRTILE8380", DVPTR(TILE_RRTILE8380), 0, RRTILE8380__STATICRR },
    { "RRTILE8385", DVPTR(TILE_RRTILE8385), 0, RRTILE8385__STATICRR },
    { "RRTILE8386", DVPTR(TILE_RRTILE8386), 0, RRTILE8386__STATICRR },
    { "RRTILE8387", DVPTR(TILE_RRTILE8387), 0, RRTILE8387__STATICRR },
    { "RRTILE8388", DVPTR(TILE_RRTILE8388), 0, RRTILE8388__STATICRR },
    { "RRTILE8389", DVPTR(TILE_RRTILE8389), 0, RRTILE8389__STATICRR },
    { "RRTILE8390", DVPTR(TILE_RRTILE8390), 0, RRTILE8390__STATICRR },
    { "RRTILE8391", DVPTR(TILE_RRTILE8391), 0, RRTILE8391__STATICRR },
    { "RRTILE8392", DVPTR(TILE_RRTILE8392), 0, RRTILE8392__STATICRR },
    { "RRTILE8394", DVPTR(TILE_RRTILE8394), 0, RRTILE8394__STATICRR },
    { "RRTILE8395", DVPTR(TILE_RRTILE8395), 0, RRTILE8395__STATICRR },
    { "RRTILE8396", DVPTR(TILE_RRTILE8396), 0, RRTILE8396__STATICRR },
    { "RRTILE8397", DVPTR(TILE_RRTILE8397), 0, RRTILE8397__STATICRR },
    { "RRTILE8398", DVPTR(TILE_RRTILE8398), 0, RRTILE8398__STATICRR },
    { "RRTILE8399", DVPTR(TILE_RRTILE8399), 0, RRTILE8399__STATICRR },
    { "RRTILE8423", DVPTR(TILE_RRTILE8423), 0, RRTILE8423__STATICRR },
    { "RRTILE8448", DVPTR(TILE_RRTILE8448), 0, RRTILE8448__STATICRR },
    { "RRTILE8450", DVPTR(TILE_RRTILE8450), 0, RRTILE8450__STATICRR },
    { "BOATAMMO", DVPTR(TILE_BOATAMMO), 0, BOATAMMO__STATICRR },
    { "RRTILE8461", DVPTR(TILE_RRTILE8461), 0, RRTILE8461__STATICRR },
    { "RRTILE8462", DVPTR(TILE_RRTILE8462), 0, RRTILE8462__STATICRR },
    { "RRTILE8464", DVPTR(TILE_RRTILE8464), 0, RRTILE8464__STATICRR },
    { "RRTILE8475", DVPTR(TILE_RRTILE8475), 0, RRTILE8475__STATICRR },
    { "RRTILE8487", DVPTR(TILE_RRTILE8487), 0, RRTILE8487__STATICRR },
    { "RRTILE8488", DVPTR(TILE_RRTILE8488), 0, RRTILE8488__STATICRR },
    { "RRTILE8489", DVPTR(TILE_RRTILE8489), 0, RRTILE8489__STATICRR },
    { "RRTILE8490", DVPTR(TILE_RRTILE8490), 0, RRTILE8490__STATICRR },
    { "RRTILE8496", DVPTR(TILE_RRTILE8496), 0, RRTILE8496__STATICRR },
    { "RRTILE8497", DVPTR(TILE_RRTILE8497), 0, RRTILE8497__STATICRR },
    { "RRTILE8498", DVPTR(TILE_RRTILE8498), 0, RRTILE8498__STATICRR },
    { "RRTILE8499", DVPTR(TILE_RRTILE8499), 0, RRTILE8499__STATICRR },
    { "RRTILE8503", DVPTR(TILE_RRTILE8503), 0, RRTILE8503__STATICRR },
    { "RRTILE8525", DVPTR(TILE_RRTILE8525), 0, RRTILE8525__STATICRR },
    { "RRTILE8537", DVPTR(TILE_RRTILE8537), 0, RRTILE8537__STATICRR },
    { "RRTILE8565", DVPTR(TILE_RRTILE8565), 0, RRTILE8565__STATICRR },
    { "RRTILE8567", DVPTR(TILE_RRTILE8567), 0, RRTILE8567__STATICRR },
    { "RRTILE8568", DVPTR(TILE_RRTILE8568), 0, RRTILE8568__STATICRR },
    { "RRTILE8569", DVPTR(TILE_RRTILE8569), 0, RRTILE8569__STATICRR },
    { "RRTILE8570", DVPTR(TILE_RRTILE8570), 0, RRTILE8570__STATICRR },
    { "RRTILE8571", DVPTR(TILE_RRTILE8571), 0, RRTILE8571__STATICRR },
    { "RRTILE8579", DVPTR(TILE_RRTILE8579), 0, RRTILE8579__STATICRR },
    { "RRTILE8588", DVPTR(TILE_RRTILE8588), 0, RRTILE8588__STATICRR },
    { "RRTILE8589", DVPTR(TILE_RRTILE8589), 0, RRTILE8589__STATICRR },
    { "RRTILE8590", DVPTR(TILE_RRTILE8590), 0, RRTILE8590__STATICRR },
    { "RRTILE8591", DVPTR(TILE_RRTILE8591), 0, RRTILE8591__STATICRR },
    { "RRTILE8592", DVPTR(TILE_RRTILE8592), 0, RRTILE8592__STATICRR },
    { "RRTILE8593", DVPTR(TILE_RRTILE8593), 0, RRTILE8593__STATICRR },
    { "RRTILE8594", DVPTR(TILE_RRTILE8594), 0, RRTILE8594__STATICRR },
    { "RRTILE8595", DVPTR(TILE_RRTILE8595), 0, RRTILE8595__STATICRR },
    { "RRTILE8596", DVPTR(TILE_RRTILE8596), 0, RRTILE8596__STATICRR },
    { "RRTILE8598", DVPTR(TILE_RRTILE8598), 0, RRTILE8598__STATICRR },
    { "RRTILE8605", DVPTR(TILE_RRTILE8605), 0, RRTILE8605__STATICRR },
    { "RRTILE8608", DVPTR(TILE_RRTILE8608), 0, RRTILE8608__STATICRR },
    { "RRTILE8609", DVPTR(TILE_RRTILE8609), 0, RRTILE8609__STATICRR },
    { "RRTILE8611", DVPTR(TILE_RRTILE8611), 0, RRTILE8611__STATICRR },
    { "RRTILE8617", DVPTR(TILE_RRTILE8617), 0, RRTILE8617__STATICRR },
    { "RRTILE8618", DVPTR(TILE_RRTILE8618), 0, RRTILE8618__STATICRR },
    { "RRTILE8620", DVPTR(TILE_RRTILE8620), 0, RRTILE8620__STATICRR },
    { "RRTILE8621", DVPTR(TILE_RRTILE8621), 0, RRTILE8621__STATICRR },
    { "RRTILE8622", DVPTR(TILE_RRTILE8622), 0, RRTILE8622__STATICRR },
    { "RRTILE8623", DVPTR(TILE_RRTILE8623), 0, RRTILE8623__STATICRR },
    { "RRTILE8624", DVPTR(TILE_RRTILE8624), 0, RRTILE8624__STATICRR },
    { "RRTILE8640", DVPTR(TILE_RRTILE8640), 0, RRTILE8640__STATICRR },
    { "RRTILE8651", DVPTR(TILE_RRTILE8651), 0, RRTILE8651__STATICRR },
    { "RRTILE8660", DVPTR(TILE_RRTILE8660), 0, RRTILE8660__STATICRR },
    { "RRTILE8677", DVPTR(TILE_RRTILE8677), 0, RRTILE8677__STATICRR },
    { "RRTILE8679", DVPTR(TILE_RRTILE8679), 0, RRTILE8679__STATICRR },
    { "RRTILE8680", DVPTR(TILE_RRTILE8680), 0, RRTILE8680__STATICRR },
    { "RRTILE8681", DVPTR(TILE_RRTILE8681), 0, RRTILE8681__STATICRR },
    { "RRTILE8682", DVPTR(TILE_RRTILE8682), 0, RRTILE8682__STATICRR },
    { "RRTILE8683", DVPTR(TILE_RRTILE8683), 0, RRTILE8683__STATICRR },
    { "RRTILE8704", DVPTR(TILE_RRTILE8704), 0, RRTILE8704__STATICRR },
    { "BOULDER", DVPTR(TILE_BOULDER), 0, BOULDER__STATICRR },
    { "BOULDER1", DVPTR(TILE_BOULDER1), 0, BOULDER1__STATICRR },
    { "TORNADO", DVPTR(TILE_TORNADO), 0, TORNADO__STATICRR },
    { "CHEERBOMB", DVPTR(TILE_CHEERBOMB), 0, CHEERBOMB__STATICRR },
    { "CHEERBLADE", DVPTR(TILE_CHEERBLADE), 0, CHEERBLADE__STATICRR },
    { "DOGATTACK", DVPTR(TILE_DOGATTACK), 0, DOGATTACK__STATICRR },
    { "BILLYWALK", DVPTR(TILE_BILLYWALK), 0, BILLYWALK__STATICRR },
    { "BILLYDIE", DVPTR(TILE_BILLYDIE), 0, BILLYDIE__STATICRR },
    { "BILLYCOCK", DVPTR(TILE_BILLYCOCK), 0, BILLYCOCK__STATICRR },
    { "BILLYRAY", DVPTR(TILE_BILLYRAY), 0, BILLYRAY__STATICRR },
    { "BILLYRAYSTAYPUT", DVPTR(TILE_BILLYRAYSTAYPUT), 0, BILLYRAYSTAYPUT__STATICRR },
    { "BILLYBUT", DVPTR(TILE_BILLYBUT), 0, BILLYBUT__STATICRR },
    { "BILLYSCRATCH", DVPTR(TILE_BILLYSCRATCH), 0, BILLYSCRATCH__STATICRR },
    { "BILLYSNIFF", DVPTR(TILE_BILLYSNIFF), 0, BILLYSNIFF__STATICRR },
    { "BILLYWOUND", DVPTR(TILE_BILLYWOUND), 0, BILLYWOUND__STATICRR },
    { "BILLYGORE", DVPTR(TILE_BILLYGORE), 0, BILLYGORE__STATICRR },
    { "BILLYJIBA", DVPTR(TILE_BILLYJIBA), 0, BILLYJIBA__STATICRR },
    { "BILLYJIBB", DVPTR(TILE_BILLYJIBB), 0, BILLYJIBB__STATICRR },
    { "BRAYSNIPER", DVPTR(TILE_BRAYSNIPER), 0, BRAYSNIPER__STATICRR },
    { "DOGRUN", DVPTR(TILE_DOGRUN), 0, DOGRUN__STATICRR },
    { "DOGDIE", DVPTR(TILE_DOGDIE), 0, DOGDIE__STATICRR },
    { "DOGDEAD", DVPTR(TILE_DOGDEAD), 0, DOGDEAD__STATICRR },
    { "DOGBARK", DVPTR(TILE_DOGBARK), 0, DOGBARK__STATICRR },
    { "LTH", DVPTR(TILE_LTH), 0, LTH__STATICRR },
    { "LTHSTRAFE", DVPTR(TILE_LTHSTRAFE), 0, LTHSTRAFE__STATICRR },
    { "HULKHANG", DVPTR(TILE_HULKHANG), 0, HULKHANG__STATICRR },
    { "HULKHANGDEAD", DVPTR(TILE_HULKHANGDEAD), 0, HULKHANGDEAD__STATICRR },
    { "HULKJUMP", DVPTR(TILE_HULKJUMP), 0, HULKJUMP__STATICRR },
    { "LTHLOAD", DVPTR(TILE_LTHLOAD), 0, LTHLOAD__STATICRR },
    { "LTHDIE", DVPTR(TILE_LTHDIE), 0, LTHDIE__STATICRR },
    { "BUBBASCRATCH", DVPTR(TILE_BUBBASCRATCH), 0, BUBBASCRATCH__STATICRR },
    { "BUBBANOSE", DVPTR(TILE_BUBBANOSE), 0, BUBBANOSE__STATICRR },
    { "BUBBAPISS", DVPTR(TILE_BUBBAPISS), 0, BUBBAPISS__STATICRR },
    { "BUBBASTAND", DVPTR(TILE_BUBBASTAND), 0, BUBBASTAND__STATICRR },
    { "BUBBAOUCH", DVPTR(TILE_BUBBAOUCH), 0, BUBBAOUCH__STATICRR },
    { "BUBBADIE", DVPTR(TILE_BUBBADIE), 0, BUBBADIE__STATICRR },
    { "BUBBADEAD", DVPTR(TILE_BUBBADEAD), 0, BUBBADEAD__STATICRR },
    { "HULK", DVPTR(TILE_HULK), 0, HULK__STATICRR },
    { "HULKSTAYPUT", DVPTR(TILE_HULKSTAYPUT), 0, HULKSTAYPUT__STATICRR },
    { "HULKA", DVPTR(TILE_HULKA), 0, HULKA__STATICRR },
    { "HULKB", DVPTR(TILE_HULKB), 0, HULKB__STATICRR },
    { "HULKC", DVPTR(TILE_HULKC), 0, HULKC__STATICRR },
    { "HULKJIBA", DVPTR(TILE_HULKJIBA), 0, HULKJIBA__STATICRR },
    { "HULKJIBB", DVPTR(TILE_HULKJIBB), 0, HULKJIBB__STATICRR },
    { "HULKJIBC", DVPTR(TILE_HULKJIBC), 0, HULKJIBC__STATICRR },
    { "SBSWIPE", DVPTR(TILE_SBSWIPE), 0, SBSWIPE__STATICRR },
    { "SBPAIN", DVPTR(TILE_SBPAIN), 0, SBPAIN__STATICRR },
    { "SBDIE", DVPTR(TILE_SBDIE), 0, SBDIE__STATICRR },
    { "HEN", DVPTR(TILE_HEN), 0, HEN__STATICRR },
    { "HENSTAYPUT", DVPTR(TILE_HENSTAYPUT), 0, HENSTAYPUT__STATICRR },
    { "HENSTAND", DVPTR(TILE_HENSTAND), 0, HENSTAND__STATICRR },
    { "PIG", DVPTR(TILE_PIG), 0, PIG__STATICRR },
    { "PIGSTAYPUT", DVPTR(TILE_PIGSTAYPUT), 0, PIGSTAYPUT__STATICRR },
    { "PIGEAT", DVPTR(TILE_PIGEAT), 0, PIGEAT__STATICRR },
    { "SBMOVE", DVPTR(TILE_SBMOVE), 0, SBMOVE__STATICRR },
    { "SBSPIT", DVPTR(TILE_SBSPIT), 0, SBSPIT__STATICRR },
    { "SBDIP", DVPTR(TILE_SBDIP), 0, SBDIP__STATICRR },
    { "MINION", DVPTR(TILE_MINION), 0, MINION__STATICRR },
    { "MINIONSTAYPUT", DVPTR(TILE_MINIONSTAYPUT), 0, MINIONSTAYPUT__STATICRR },
    { "UFO1", DVPTR(TILE_UFO1), 0, UFO1__STATICRR },
    { "UFO2", DVPTR(TILE_UFO2), 0, UFO2__STATICRR },
    { "UFO3", DVPTR(TILE_UFO3), 0, UFO3__STATICRR },
    { "UFO4", DVPTR(TILE_UFO4), 0, UFO4__STATICRR },
    { "UFO5", DVPTR(TILE_UFO5), 0, UFO5__STATICRR },
    { "MINJIBA", DVPTR(TILE_MINJIBA), 0, MINJIBA__STATICRR },
    { "MINJIBB", DVPTR(TILE_MINJIBB), 0, MINJIBB__STATICRR },
    { "MINJIBC", DVPTR(TILE_MINJIBC), 0, MINJIBC__STATICRR },
    { "COW", DVPTR(TILE_COW), 0, COW__STATICRR },
    { "COOT", DVPTR(TILE_COOT), 0, COOT__STATICRR },
    { "COOTSTAYPUT", DVPTR(TILE_COOTSTAYPUT), 0, COOTSTAYPUT__STATICRR },
    { "COOTSHOOT", DVPTR(TILE_COOTSHOOT), 0, COOTSHOOT__STATICRR },
    { "COOTDIE", DVPTR(TILE_COOTDIE), 0, COOTDIE__STATICRR },
    { "COOTDUCK", DVPTR(TILE_COOTDUCK), 0, COOTDUCK__STATICRR },
    { "COOTPAIN", DVPTR(TILE_COOTPAIN), 0, COOTPAIN__STATICRR },
    { "COOTTRANS", DVPTR(TILE_COOTTRANS), 0, COOTTRANS__STATICRR },
    { "COOTGETUP", DVPTR(TILE_COOTGETUP), 0, COOTGETUP__STATICRR },
    { "COOTJIBA", DVPTR(TILE_COOTJIBA), 0, COOTJIBA__STATICRR },
    { "COOTJIBB", DVPTR(TILE_COOTJIBB), 0, COOTJIBB__STATICRR },
    { "COOTJIBC", DVPTR(TILE_COOTJIBC), 0, COOTJIBC__STATICRR },
    { "VIXEN", DVPTR(TILE_VIXEN), 0, VIXEN__STATICRR },
    { "VIXENPAIN", DVPTR(TILE_VIXENPAIN), 0, VIXENPAIN__STATICRR },
    { "VIXENDIE", DVPTR(TILE_VIXENDIE), 0, VIXENDIE__STATICRR },
    { "VIXENSHOOT", DVPTR(TILE_VIXENSHOOT), 0, VIXENSHOOT__STATICRR },
    { "VIXENWDN", DVPTR(TILE_VIXENWDN), 0, VIXENWDN__STATICRR },
    { "VIXENWUP", DVPTR(TILE_VIXENWUP), 0, VIXENWUP__STATICRR },
    { "VIXENKICK", DVPTR(TILE_VIXENKICK), 0, VIXENKICK__STATICRR },
    { "VIXENTELE", DVPTR(TILE_VIXENTELE), 0, VIXENTELE__STATICRR },
    { "VIXENTEAT", DVPTR(TILE_VIXENTEAT), 0, VIXENTEAT__STATICRR },
    { "BIKEJIBA", DVPTR(TILE_BIKEJIBA), 0, BIKEJIBA__STATICRR },
    { "BIKEJIBB", DVPTR(TILE_BIKEJIBB), 0, BIKEJIBB__STATICRR },
    { "BIKEJIBC", DVPTR(TILE_BIKEJIBC), 0, BIKEJIBC__STATICRR },
    { "BIKERB", DVPTR(TILE_BIKERB), 0, BIKERB__STATICRR },
    { "BIKERBV2", DVPTR(TILE_BIKERBV2), 0, BIKERBV2__STATICRR },
    { "BIKER", DVPTR(TILE_BIKER), 0, BIKER__STATICRR },
    { "BIKERJIBA", DVPTR(TILE_BIKERJIBA), 0, BIKERJIBA__STATICRR },
    { "BIKERJIBB", DVPTR(TILE_BIKERJIBB), 0, BIKERJIBB__STATICRR },
    { "BIKERJIBC", DVPTR(TILE_BIKERJIBC), 0, BIKERJIBC__STATICRR },
    { "BIKERJIBD", DVPTR(TILE_BIKERJIBD), 0, BIKERJIBD__STATICRR },
    { "MAKEOUT", DVPTR(TILE_MAKEOUT), 0, MAKEOUT__STATICRR },
    { "CHEERB", DVPTR(TILE_CHEERB), 0, CHEERB__STATICRR },
    { "CHEER", DVPTR(TILE_CHEER), 0, CHEER__STATICRR },
    { "CHEERSTAYPUT", DVPTR(TILE_CHEERSTAYPUT), 0, CHEERSTAYPUT__STATICRR },
    { "CHEERJIBA", DVPTR(TILE_CHEERJIBA), 0, CHEERJIBA__STATICRR },
    { "CHEERJIBB", DVPTR(TILE_CHEERJIBB), 0, CHEERJIBB__STATICRR },
    { "CHEERJIBC", DVPTR(TILE_CHEERJIBC), 0, CHEERJIBC__STATICRR },
    { "CHEERJIBD", DVPTR(TILE_CHEERJIBD), 0, CHEERJIBD__STATICRR },
    { "FBOATJIBA", DVPTR(TILE_FBOATJIBA), 0, FBOATJIBA__STATICRR },
    { "FBOATJIBB", DVPTR(TILE_FBOATJIBB), 0, FBOATJIBB__STATICRR },
    { "COOTPLAY", DVPTR(TILE_COOTPLAY), 0, COOTPLAY__STATICRR },
    { "BILLYPLAY", DVPTR(TILE_BILLYPLAY), 0, BILLYPLAY__STATICRR },
    { "MINIONBOAT", DVPTR(TILE_MINIONBOAT), 0, MINIONBOAT__STATICRR },
    { "HULKBOAT", DVPTR(TILE_HULKBOAT), 0, HULKBOAT__STATICRR },
    { "CHEERBOAT", DVPTR(TILE_CHEERBOAT), 0, CHEERBOAT__STATICRR },
    { "RRTILE7274", DVPTR(TILE_RRTILE7274), 0, RRTILE7274__STATICRR },
    { "RABBIT", DVPTR(TILE_RABBIT), 0, RABBIT__STATICRR },
    { "RABBITJIBA", DVPTR(TILE_RABBITJIBA), 0, RABBITJIBA__STATICRR },
    { "RABBITJIBB", DVPTR(TILE_RABBITJIBB), 0, RABBITJIBB__STATICRR },
    { "RABBITJIBC", DVPTR(TILE_RABBITJIBC), 0, RABBITJIBC__STATICRR },
    { "ROCK", DVPTR(TILE_ROCK), 0, ROCK__STATICRR },
    { "ROCK2", DVPTR(TILE_ROCK2), 0, ROCK2__STATICRR },
    { "MAMACLOUD", DVPTR(TILE_MAMACLOUD), 0, MAMACLOUD__STATICRR },
    { "MAMA", DVPTR(TILE_MAMA), 0, MAMA__STATICRR },
    { "MAMAJIBA", DVPTR(TILE_MAMAJIBA), 0, MAMAJIBA__STATICRR },
    { "MAMAJIBB", DVPTR(TILE_MAMAJIBB), 0, MAMAJIBB__STATICRR },
    { NULL, NULL, -1, -1 },
};

int32_t TILE_SECTOREFFECTOR = SECTOREFFECTOR__STATIC;
int32_t TILE_ACTIVATOR = ACTIVATOR__STATIC;
int32_t TILE_TOUCHPLATE = TOUCHPLATE__STATIC;
int32_t TILE_ACTIVATORLOCKED = ACTIVATORLOCKED__STATIC;
int32_t TILE_MUSICANDSFX = MUSICANDSFX__STATIC;
int32_t TILE_LOCATORS = LOCATORS__STATIC;
int32_t TILE_CYCLER = CYCLER__STATIC;
int32_t TILE_MASTERSWITCH = MASTERSWITCH__STATIC;
int32_t TILE_RESPAWN = RESPAWN__STATIC;
int32_t TILE_GPSPEED = GPSPEED__STATIC;
int32_t TILE_FOF = FOF__STATIC;
int32_t TILE_ARROW = ARROW__STATIC;
int32_t TILE_FIRSTGUNSPRITE = FIRSTGUNSPRITE__STATIC;
int32_t TILE_CHAINGUNSPRITE = CHAINGUNSPRITE__STATIC;
int32_t TILE_RPGSPRITE = RPGSPRITE__STATIC;
int32_t TILE_FREEZESPRITE = FREEZESPRITE__STATIC;
int32_t TILE_SHRINKERSPRITE = SHRINKERSPRITE__STATIC;
int32_t TILE_HEAVYHBOMB = HEAVYHBOMB__STATIC;
int32_t TILE_TRIPBOMBSPRITE = TRIPBOMBSPRITE__STATIC;
int32_t TILE_SHOTGUNSPRITE = SHOTGUNSPRITE__STATIC;
int32_t TILE_DEVISTATORSPRITE = DEVISTATORSPRITE__STATIC;
int32_t TILE_HEALTHBOX = HEALTHBOX__STATIC;
int32_t TILE_AMMOBOX = AMMOBOX__STATIC;
int32_t TILE_GROWSPRITEICON = GROWSPRITEICON__STATIC;
int32_t TILE_INVENTORYBOX = INVENTORYBOX__STATIC;
int32_t TILE_FREEZEAMMO = FREEZEAMMO__STATIC;
int32_t TILE_AMMO = AMMO__STATIC;
int32_t TILE_BATTERYAMMO = BATTERYAMMO__STATIC;
int32_t TILE_DEVISTATORAMMO = DEVISTATORAMMO__STATIC;
int32_t TILE_RPGAMMO = RPGAMMO__STATIC;
int32_t TILE_GROWAMMO = GROWAMMO__STATIC;
int32_t TILE_CRYSTALAMMO = CRYSTALAMMO__STATIC;
int32_t TILE_HBOMBAMMO = HBOMBAMMO__STATIC;
int32_t TILE_AMMOLOTS = AMMOLOTS__STATIC;
int32_t TILE_SHOTGUNAMMO = SHOTGUNAMMO__STATIC;
int32_t TILE_COLA = COLA__STATIC;
int32_t TILE_SIXPAK = SIXPAK__STATIC;
int32_t TILE_FIRSTAID = FIRSTAID__STATIC;
int32_t TILE_SHIELD = SHIELD__STATIC;
int32_t TILE_STEROIDS = STEROIDS__STATIC;
int32_t TILE_AIRTANK = AIRTANK__STATIC;
int32_t TILE_JETPACK = JETPACK__STATIC;
int32_t TILE_HEATSENSOR = HEATSENSOR__STATIC;
int32_t TILE_ACCESSCARD = ACCESSCARD__STATIC;
int32_t TILE_BOOTS = BOOTS__STATIC;
int32_t TILE_MIRRORBROKE = MIRRORBROKE__STATIC;
int32_t TILE_CLOUDYOCEAN = CLOUDYOCEAN__STATIC;
int32_t TILE_CLOUDYSKIES = CLOUDYSKIES__STATIC;
int32_t TILE_MOONSKY1 = MOONSKY1__STATIC;
int32_t TILE_MOONSKY2 = MOONSKY2__STATIC;
int32_t TILE_MOONSKY3 = MOONSKY3__STATIC;
int32_t TILE_MOONSKY4 = MOONSKY4__STATIC;
int32_t TILE_BIGORBIT1 = BIGORBIT1__STATIC;
int32_t TILE_BIGORBIT2 = BIGORBIT2__STATIC;
int32_t TILE_BIGORBIT3 = BIGORBIT3__STATIC;
int32_t TILE_BIGORBIT4 = BIGORBIT4__STATIC;
int32_t TILE_BIGORBIT5 = BIGORBIT5__STATIC;
int32_t TILE_LA = LA__STATIC;
int32_t TILE_REDSKY1 = REDSKY1__STATIC;
int32_t TILE_REDSKY2 = REDSKY2__STATIC;
int32_t TILE_ATOMICHEALTH = ATOMICHEALTH__STATIC;
int32_t TILE_TECHLIGHT2 = TECHLIGHT2__STATIC;
int32_t TILE_TECHLIGHTBUST2 = TECHLIGHTBUST2__STATIC;
int32_t TILE_TECHLIGHT4 = TECHLIGHT4__STATIC;
int32_t TILE_TECHLIGHTBUST4 = TECHLIGHTBUST4__STATIC;
int32_t TILE_WALLLIGHT4 = WALLLIGHT4__STATIC;
int32_t TILE_WALLLIGHTBUST4 = WALLLIGHTBUST4__STATIC;
int32_t TILE_ACCESSSWITCH = ACCESSSWITCH__STATIC;
int32_t TILE_SLOTDOOR = SLOTDOOR__STATIC;
int32_t TILE_LIGHTSWITCH = LIGHTSWITCH__STATIC;
int32_t TILE_SPACEDOORSWITCH = SPACEDOORSWITCH__STATIC;
int32_t TILE_SPACELIGHTSWITCH = SPACELIGHTSWITCH__STATIC;
int32_t TILE_FRANKENSTINESWITCH = FRANKENSTINESWITCH__STATIC;
int32_t TILE_NUKEBUTTON = NUKEBUTTON__STATIC;
int32_t TILE_MULTISWITCH = MULTISWITCH__STATIC;
int32_t TILE_DOORTILE5 = DOORTILE5__STATIC;
int32_t TILE_DOORTILE6 = DOORTILE6__STATIC;
int32_t TILE_DOORTILE1 = DOORTILE1__STATIC;
int32_t TILE_DOORTILE2 = DOORTILE2__STATIC;
int32_t TILE_DOORTILE3 = DOORTILE3__STATIC;
int32_t TILE_DOORTILE4 = DOORTILE4__STATIC;
int32_t TILE_DOORTILE7 = DOORTILE7__STATIC;
int32_t TILE_DOORTILE8 = DOORTILE8__STATIC;
int32_t TILE_DOORTILE9 = DOORTILE9__STATIC;
int32_t TILE_DOORTILE10 = DOORTILE10__STATIC;
int32_t TILE_DOORSHOCK = DOORSHOCK__STATIC;
int32_t TILE_DIPSWITCH = DIPSWITCH__STATIC;
int32_t TILE_DIPSWITCH2 = DIPSWITCH2__STATIC;
int32_t TILE_TECHSWITCH = TECHSWITCH__STATIC;
int32_t TILE_DIPSWITCH3 = DIPSWITCH3__STATIC;
int32_t TILE_ACCESSSWITCH2 = ACCESSSWITCH2__STATIC;
int32_t TILE_REFLECTWATERTILE = REFLECTWATERTILE__STATIC;
int32_t TILE_FLOORSLIME = FLOORSLIME__STATIC;
int32_t TILE_BIGFORCE = BIGFORCE__STATIC;
int32_t TILE_EPISODE = EPISODE__STATIC;
int32_t TILE_MASKWALL9 = MASKWALL9__STATIC;
int32_t TILE_W_LIGHT = W_LIGHT__STATIC;
int32_t TILE_SCREENBREAK1 = SCREENBREAK1__STATIC;
int32_t TILE_SCREENBREAK2 = SCREENBREAK2__STATIC;
int32_t TILE_SCREENBREAK3 = SCREENBREAK3__STATIC;
int32_t TILE_SCREENBREAK4 = SCREENBREAK4__STATIC;
int32_t TILE_SCREENBREAK5 = SCREENBREAK5__STATIC;
int32_t TILE_SCREENBREAK6 = SCREENBREAK6__STATIC;
int32_t TILE_SCREENBREAK7 = SCREENBREAK7__STATIC;
int32_t TILE_SCREENBREAK8 = SCREENBREAK8__STATIC;
int32_t TILE_SCREENBREAK9 = SCREENBREAK9__STATIC;
int32_t TILE_SCREENBREAK10 = SCREENBREAK10__STATIC;
int32_t TILE_SCREENBREAK11 = SCREENBREAK11__STATIC;
int32_t TILE_SCREENBREAK12 = SCREENBREAK12__STATIC;
int32_t TILE_SCREENBREAK13 = SCREENBREAK13__STATIC;
int32_t TILE_MASKWALL1 = MASKWALL1__STATIC;
int32_t TILE_W_TECHWALL1 = W_TECHWALL1__STATIC;
int32_t TILE_W_TECHWALL2 = W_TECHWALL2__STATIC;
int32_t TILE_W_TECHWALL15 = W_TECHWALL15__STATIC;
int32_t TILE_W_TECHWALL3 = W_TECHWALL3__STATIC;
int32_t TILE_W_TECHWALL4 = W_TECHWALL4__STATIC;
int32_t TILE_W_TECHWALL10 = W_TECHWALL10__STATIC;
int32_t TILE_W_TECHWALL16 = W_TECHWALL16__STATIC;
int32_t TILE_WATERTILE2 = WATERTILE2__STATIC;
int32_t TILE_BPANNEL1 = BPANNEL1__STATIC;
int32_t TILE_PANNEL1 = PANNEL1__STATIC;
int32_t TILE_PANNEL2 = PANNEL2__STATIC;
int32_t TILE_WATERTILE = WATERTILE__STATIC;
int32_t TILE_STATIC = STATIC__STATIC;
int32_t TILE_W_SCREENBREAK = W_SCREENBREAK__STATIC;
int32_t TILE_W_HITTECHWALL3 = W_HITTECHWALL3__STATIC;
int32_t TILE_W_HITTECHWALL4 = W_HITTECHWALL4__STATIC;
int32_t TILE_W_HITTECHWALL2 = W_HITTECHWALL2__STATIC;
int32_t TILE_W_HITTECHWALL1 = W_HITTECHWALL1__STATIC;
int32_t TILE_MASKWALL10 = MASKWALL10__STATIC;
int32_t TILE_MASKWALL11 = MASKWALL11__STATIC;
int32_t TILE_DOORTILE22 = DOORTILE22__STATIC;
int32_t TILE_FANSPRITE = FANSPRITE__STATIC;
int32_t TILE_FANSPRITEBROKE = FANSPRITEBROKE__STATIC;
int32_t TILE_FANSHADOW = FANSHADOW__STATIC;
int32_t TILE_FANSHADOWBROKE = FANSHADOWBROKE__STATIC;
int32_t TILE_DOORTILE18 = DOORTILE18__STATIC;
int32_t TILE_DOORTILE19 = DOORTILE19__STATIC;
int32_t TILE_DOORTILE20 = DOORTILE20__STATIC;
int32_t TILE_SATELLITE = SATELLITE__STATIC;
int32_t TILE_VIEWSCREEN2 = VIEWSCREEN2__STATIC;
int32_t TILE_VIEWSCREENBROKE = VIEWSCREENBROKE__STATIC;
int32_t TILE_VIEWSCREEN = VIEWSCREEN__STATIC;
int32_t TILE_GLASS = GLASS__STATIC;
int32_t TILE_GLASS2 = GLASS2__STATIC;
int32_t TILE_STAINGLASS1 = STAINGLASS1__STATIC;
int32_t TILE_MASKWALL5 = MASKWALL5__STATIC;
int32_t TILE_SATELITE = SATELITE__STATIC;
int32_t TILE_FUELPOD = FUELPOD__STATIC;
int32_t TILE_SLIMEPIPE = SLIMEPIPE__STATIC;
int32_t TILE_CRACK1 = CRACK1__STATIC;
int32_t TILE_CRACK2 = CRACK2__STATIC;
int32_t TILE_CRACK3 = CRACK3__STATIC;
int32_t TILE_CRACK4 = CRACK4__STATIC;
int32_t TILE_FOOTPRINTS = FOOTPRINTS__STATIC;
int32_t TILE_DOMELITE = DOMELITE__STATIC;
int32_t TILE_CAMERAPOLE = CAMERAPOLE__STATIC;
int32_t TILE_CHAIR1 = CHAIR1__STATIC;
int32_t TILE_CHAIR2 = CHAIR2__STATIC;
int32_t TILE_BROKENCHAIR = BROKENCHAIR__STATIC;
int32_t TILE_MIRROR = MIRROR__STATIC;
int32_t TILE_WATERFOUNTAIN = WATERFOUNTAIN__STATIC;
int32_t TILE_WATERFOUNTAINBROKE = WATERFOUNTAINBROKE__STATIC;
int32_t TILE_FEMMAG1 = FEMMAG1__STATIC;
int32_t TILE_TOILET = TOILET__STATIC;
int32_t TILE_STALL = STALL__STATIC;
int32_t TILE_STALLBROKE = STALLBROKE__STATIC;
int32_t TILE_FEMMAG2 = FEMMAG2__STATIC;
int32_t TILE_REACTOR2 = REACTOR2__STATIC;
int32_t TILE_REACTOR2BURNT = REACTOR2BURNT__STATIC;
int32_t TILE_REACTOR2SPARK = REACTOR2SPARK__STATIC;
int32_t TILE_GRATE1 = GRATE1__STATIC;
int32_t TILE_BGRATE1 = BGRATE1__STATIC;
int32_t TILE_SOLARPANNEL = SOLARPANNEL__STATIC;
int32_t TILE_NAKED1 = NAKED1__STATIC;
int32_t TILE_ANTENNA = ANTENNA__STATIC;
int32_t TILE_MASKWALL12 = MASKWALL12__STATIC;
int32_t TILE_TOILETBROKE = TOILETBROKE__STATIC;
int32_t TILE_PIPE2 = PIPE2__STATIC;
int32_t TILE_PIPE1B = PIPE1B__STATIC;
int32_t TILE_PIPE3 = PIPE3__STATIC;
int32_t TILE_PIPE1 = PIPE1__STATIC;
int32_t TILE_CAMERA1 = CAMERA1__STATIC;
int32_t TILE_BRICK = BRICK__STATIC;
int32_t TILE_SPLINTERWOOD = SPLINTERWOOD__STATIC;
int32_t TILE_PIPE2B = PIPE2B__STATIC;
int32_t TILE_BOLT1 = BOLT1__STATIC;
int32_t TILE_W_NUMBERS = W_NUMBERS__STATIC;
int32_t TILE_WATERDRIP = WATERDRIP__STATIC;
int32_t TILE_WATERBUBBLE = WATERBUBBLE__STATIC;
int32_t TILE_WATERBUBBLEMAKER = WATERBUBBLEMAKER__STATIC;
int32_t TILE_W_FORCEFIELD = W_FORCEFIELD__STATIC;
int32_t TILE_VACUUM = VACUUM__STATIC;
int32_t TILE_FOOTPRINTS2 = FOOTPRINTS2__STATIC;
int32_t TILE_FOOTPRINTS3 = FOOTPRINTS3__STATIC;
int32_t TILE_FOOTPRINTS4 = FOOTPRINTS4__STATIC;
int32_t TILE_EGG = EGG__STATIC;
int32_t TILE_SCALE = SCALE__STATIC;
int32_t TILE_CHAIR3 = CHAIR3__STATIC;
int32_t TILE_CAMERALIGHT = CAMERALIGHT__STATIC;
int32_t TILE_MOVIECAMERA = MOVIECAMERA__STATIC;
int32_t TILE_IVUNIT = IVUNIT__STATIC;
int32_t TILE_POT1 = POT1__STATIC;
int32_t TILE_POT2 = POT2__STATIC;
int32_t TILE_POT3 = POT3__STATIC;
int32_t TILE_PIPE3B = PIPE3B__STATIC;
int32_t TILE_WALLLIGHT3 = WALLLIGHT3__STATIC;
int32_t TILE_WALLLIGHTBUST3 = WALLLIGHTBUST3__STATIC;
int32_t TILE_WALLLIGHT1 = WALLLIGHT1__STATIC;
int32_t TILE_WALLLIGHTBUST1 = WALLLIGHTBUST1__STATIC;
int32_t TILE_WALLLIGHT2 = WALLLIGHT2__STATIC;
int32_t TILE_WALLLIGHTBUST2 = WALLLIGHTBUST2__STATIC;
int32_t TILE_LIGHTSWITCH2 = LIGHTSWITCH2__STATIC;
int32_t TILE_WAITTOBESEATED = WAITTOBESEATED__STATIC;
int32_t TILE_DOORTILE14 = DOORTILE14__STATIC;
int32_t TILE_STATUE = STATUE__STATIC;
int32_t TILE_MIKE = MIKE__STATIC;
int32_t TILE_VASE = VASE__STATIC;
int32_t TILE_SUSHIPLATE1 = SUSHIPLATE1__STATIC;
int32_t TILE_SUSHIPLATE2 = SUSHIPLATE2__STATIC;
int32_t TILE_SUSHIPLATE3 = SUSHIPLATE3__STATIC;
int32_t TILE_SUSHIPLATE4 = SUSHIPLATE4__STATIC;
int32_t TILE_DOORTILE16 = DOORTILE16__STATIC;
int32_t TILE_SUSHIPLATE5 = SUSHIPLATE5__STATIC;
int32_t TILE_OJ = OJ__STATIC;
int32_t TILE_MASKWALL13 = MASKWALL13__STATIC;
int32_t TILE_HURTRAIL = HURTRAIL__STATIC;
int32_t TILE_POWERSWITCH1 = POWERSWITCH1__STATIC;
int32_t TILE_LOCKSWITCH1 = LOCKSWITCH1__STATIC;
int32_t TILE_POWERSWITCH2 = POWERSWITCH2__STATIC;
int32_t TILE_ATM = ATM__STATIC;
int32_t TILE_STATUEFLASH = STATUEFLASH__STATIC;
int32_t TILE_ATMBROKE = ATMBROKE__STATIC;
int32_t TILE_BIGHOLE2 = BIGHOLE2__STATIC;
int32_t TILE_STRIPEBALL = STRIPEBALL__STATIC;
int32_t TILE_QUEBALL = QUEBALL__STATIC;
int32_t TILE_POCKET = POCKET__STATIC;
int32_t TILE_WOODENHORSE = WOODENHORSE__STATIC;
int32_t TILE_TREE1 = TREE1__STATIC;
int32_t TILE_TREE2 = TREE2__STATIC;
int32_t TILE_CACTUS = CACTUS__STATIC;
int32_t TILE_MASKWALL2 = MASKWALL2__STATIC;
int32_t TILE_MASKWALL3 = MASKWALL3__STATIC;
int32_t TILE_MASKWALL4 = MASKWALL4__STATIC;
int32_t TILE_FIREEXT = FIREEXT__STATIC;
int32_t TILE_TOILETWATER = TOILETWATER__STATIC;
int32_t TILE_NEON1 = NEON1__STATIC;
int32_t TILE_NEON2 = NEON2__STATIC;
int32_t TILE_CACTUSBROKE = CACTUSBROKE__STATIC;
int32_t TILE_BOUNCEMINE = BOUNCEMINE__STATIC;
int32_t TILE_BROKEFIREHYDRENT = BROKEFIREHYDRENT__STATIC;
int32_t TILE_BOX = BOX__STATIC;
int32_t TILE_BULLETHOLE = BULLETHOLE__STATIC;
int32_t TILE_BOTTLE1 = BOTTLE1__STATIC;
int32_t TILE_BOTTLE2 = BOTTLE2__STATIC;
int32_t TILE_BOTTLE3 = BOTTLE3__STATIC;
int32_t TILE_BOTTLE4 = BOTTLE4__STATIC;
int32_t TILE_FEMPIC5 = FEMPIC5__STATIC;
int32_t TILE_FEMPIC6 = FEMPIC6__STATIC;
int32_t TILE_FEMPIC7 = FEMPIC7__STATIC;
int32_t TILE_HYDROPLANT = HYDROPLANT__STATIC;
int32_t TILE_OCEANSPRITE1 = OCEANSPRITE1__STATIC;
int32_t TILE_OCEANSPRITE2 = OCEANSPRITE2__STATIC;
int32_t TILE_OCEANSPRITE3 = OCEANSPRITE3__STATIC;
int32_t TILE_OCEANSPRITE4 = OCEANSPRITE4__STATIC;
int32_t TILE_OCEANSPRITE5 = OCEANSPRITE5__STATIC;
int32_t TILE_GENERICPOLE = GENERICPOLE__STATIC;
int32_t TILE_CONE = CONE__STATIC;
int32_t TILE_HANGLIGHT = HANGLIGHT__STATIC;
int32_t TILE_HYDRENT = HYDRENT__STATIC;
int32_t TILE_MASKWALL14 = MASKWALL14__STATIC;
int32_t TILE_TIRE = TIRE__STATIC;
int32_t TILE_PIPE5 = PIPE5__STATIC;
int32_t TILE_PIPE6 = PIPE6__STATIC;
int32_t TILE_PIPE4 = PIPE4__STATIC;
int32_t TILE_PIPE4B = PIPE4B__STATIC;
int32_t TILE_BROKEHYDROPLANT = BROKEHYDROPLANT__STATIC;
int32_t TILE_PIPE5B = PIPE5B__STATIC;
int32_t TILE_NEON3 = NEON3__STATIC;
int32_t TILE_NEON4 = NEON4__STATIC;
int32_t TILE_NEON5 = NEON5__STATIC;
int32_t TILE_BOTTLE5 = BOTTLE5__STATIC;
int32_t TILE_BOTTLE6 = BOTTLE6__STATIC;
int32_t TILE_BOTTLE8 = BOTTLE8__STATIC;
int32_t TILE_SPOTLITE = SPOTLITE__STATIC;
int32_t TILE_HANGOOZ = HANGOOZ__STATIC;
int32_t TILE_MASKWALL15 = MASKWALL15__STATIC;
int32_t TILE_BOTTLE7 = BOTTLE7__STATIC;
int32_t TILE_HORSEONSIDE = HORSEONSIDE__STATIC;
int32_t TILE_GLASSPIECES = GLASSPIECES__STATIC;
int32_t TILE_HORSELITE = HORSELITE__STATIC;
int32_t TILE_DONUTS = DONUTS__STATIC;
int32_t TILE_NEON6 = NEON6__STATIC;
int32_t TILE_MASKWALL6 = MASKWALL6__STATIC;
int32_t TILE_CLOCK = CLOCK__STATIC;
int32_t TILE_RUBBERCAN = RUBBERCAN__STATIC;
int32_t TILE_BROKENCLOCK = BROKENCLOCK__STATIC;
int32_t TILE_PLUG = PLUG__STATIC;
int32_t TILE_OOZFILTER = OOZFILTER__STATIC;
int32_t TILE_FLOORPLASMA = FLOORPLASMA__STATIC;
int32_t TILE_REACTOR = REACTOR__STATIC;
int32_t TILE_REACTORSPARK = REACTORSPARK__STATIC;
int32_t TILE_REACTORBURNT = REACTORBURNT__STATIC;
int32_t TILE_DOORTILE15 = DOORTILE15__STATIC;
int32_t TILE_HANDSWITCH = HANDSWITCH__STATIC;
int32_t TILE_CIRCLEPANNEL = CIRCLEPANNEL__STATIC;
int32_t TILE_CIRCLEPANNELBROKE = CIRCLEPANNELBROKE__STATIC;
int32_t TILE_PULLSWITCH = PULLSWITCH__STATIC;
int32_t TILE_MASKWALL8 = MASKWALL8__STATIC;
int32_t TILE_BIGHOLE = BIGHOLE__STATIC;
int32_t TILE_ALIENSWITCH = ALIENSWITCH__STATIC;
int32_t TILE_DOORTILE21 = DOORTILE21__STATIC;
int32_t TILE_HANDPRINTSWITCH = HANDPRINTSWITCH__STATIC;
int32_t TILE_BOTTLE10 = BOTTLE10__STATIC;
int32_t TILE_BOTTLE11 = BOTTLE11__STATIC;
int32_t TILE_BOTTLE12 = BOTTLE12__STATIC;
int32_t TILE_BOTTLE13 = BOTTLE13__STATIC;
int32_t TILE_BOTTLE14 = BOTTLE14__STATIC;
int32_t TILE_BOTTLE15 = BOTTLE15__STATIC;
int32_t TILE_BOTTLE16 = BOTTLE16__STATIC;
int32_t TILE_BOTTLE17 = BOTTLE17__STATIC;
int32_t TILE_BOTTLE18 = BOTTLE18__STATIC;
int32_t TILE_BOTTLE19 = BOTTLE19__STATIC;
int32_t TILE_DOORTILE17 = DOORTILE17__STATIC;
int32_t TILE_MASKWALL7 = MASKWALL7__STATIC;
int32_t TILE_JAILBARBREAK = JAILBARBREAK__STATIC;
int32_t TILE_DOORTILE11 = DOORTILE11__STATIC;
int32_t TILE_DOORTILE12 = DOORTILE12__STATIC;
int32_t TILE_VENDMACHINE = VENDMACHINE__STATIC;
int32_t TILE_VENDMACHINEBROKE = VENDMACHINEBROKE__STATIC;
int32_t TILE_COLAMACHINE = COLAMACHINE__STATIC;
int32_t TILE_COLAMACHINEBROKE = COLAMACHINEBROKE__STATIC;
int32_t TILE_CRANEPOLE = CRANEPOLE__STATIC;
int32_t TILE_CRANE = CRANE__STATIC;
int32_t TILE_BARBROKE = BARBROKE__STATIC;
int32_t TILE_BLOODPOOL = BLOODPOOL__STATIC;
int32_t TILE_NUKEBARREL = NUKEBARREL__STATIC;
int32_t TILE_NUKEBARRELDENTED = NUKEBARRELDENTED__STATIC;
int32_t TILE_NUKEBARRELLEAKED = NUKEBARRELLEAKED__STATIC;
int32_t TILE_CANWITHSOMETHING = CANWITHSOMETHING__STATIC;
int32_t TILE_MONEY = MONEY__STATIC;
int32_t TILE_BANNER = BANNER__STATIC;
int32_t TILE_EXPLODINGBARREL = EXPLODINGBARREL__STATIC;
int32_t TILE_EXPLODINGBARREL2 = EXPLODINGBARREL2__STATIC;
int32_t TILE_FIREBARREL = FIREBARREL__STATIC;
int32_t TILE_SEENINE = SEENINE__STATIC;
int32_t TILE_SEENINEDEAD = SEENINEDEAD__STATIC;
int32_t TILE_STEAM = STEAM__STATIC;
int32_t TILE_CEILINGSTEAM = CEILINGSTEAM__STATIC;
int32_t TILE_PIPE6B = PIPE6B__STATIC;
int32_t TILE_TRANSPORTERBEAM = TRANSPORTERBEAM__STATIC;
int32_t TILE_RAT = RAT__STATIC;
int32_t TILE_TRASH = TRASH__STATIC;
int32_t TILE_FEMPIC1 = FEMPIC1__STATIC;
int32_t TILE_FEMPIC2 = FEMPIC2__STATIC;
int32_t TILE_BLANKSCREEN = BLANKSCREEN__STATIC;
int32_t TILE_PODFEM1 = PODFEM1__STATIC;
int32_t TILE_FEMPIC3 = FEMPIC3__STATIC;
int32_t TILE_FEMPIC4 = FEMPIC4__STATIC;
int32_t TILE_FEM1 = FEM1__STATIC;
int32_t TILE_FEM2 = FEM2__STATIC;
int32_t TILE_FEM3 = FEM3__STATIC;
int32_t TILE_FEM5 = FEM5__STATIC;
int32_t TILE_BLOODYPOLE = BLOODYPOLE__STATIC;
int32_t TILE_FEM4 = FEM4__STATIC;
int32_t TILE_FEM6 = FEM6__STATIC;
int32_t TILE_FEM6PAD = FEM6PAD__STATIC;
int32_t TILE_FEM8 = FEM8__STATIC;
int32_t TILE_HELECOPT = HELECOPT__STATIC;
int32_t TILE_FETUSJIB = FETUSJIB__STATIC;
int32_t TILE_HOLODUKE = HOLODUKE__STATIC;
int32_t TILE_SPACEMARINE = SPACEMARINE__STATIC;
int32_t TILE_INDY = INDY__STATIC;
int32_t TILE_FETUS = FETUS__STATIC;
int32_t TILE_FETUSBROKE = FETUSBROKE__STATIC;
int32_t TILE_MONK = MONK__STATIC;
int32_t TILE_LUKE = LUKE__STATIC;
int32_t TILE_COOLEXPLOSION1 = COOLEXPLOSION1__STATIC;
int32_t TILE_WATERSPLASH2 = WATERSPLASH2__STATIC;
int32_t TILE_FIREVASE = FIREVASE__STATIC;
int32_t TILE_SCRATCH = SCRATCH__STATIC;
int32_t TILE_FEM7 = FEM7__STATIC;
int32_t TILE_APLAYERTOP = APLAYERTOP__STATIC;
int32_t TILE_APLAYER = APLAYER__STATIC;
int32_t TILE_PLAYERONWATER = PLAYERONWATER__STATIC;
int32_t TILE_DUKELYINGDEAD = DUKELYINGDEAD__STATIC;
int32_t TILE_DUKETORSO = DUKETORSO__STATIC;
int32_t TILE_DUKEGUN = DUKEGUN__STATIC;
int32_t TILE_DUKELEG = DUKELEG__STATIC;
int32_t TILE_SHARK = SHARK__STATIC;
int32_t TILE_BLOOD = BLOOD__STATIC;
int32_t TILE_FIRELASER = FIRELASER__STATIC;
int32_t TILE_TRANSPORTERSTAR = TRANSPORTERSTAR__STATIC;
int32_t TILE_SPIT = SPIT__STATIC;
int32_t TILE_LOOGIE = LOOGIE__STATIC;
int32_t TILE_FIST = FIST__STATIC;
int32_t TILE_FREEZEBLAST = FREEZEBLAST__STATIC;
int32_t TILE_DEVISTATORBLAST = DEVISTATORBLAST__STATIC;
int32_t TILE_SHRINKSPARK = SHRINKSPARK__STATIC;
int32_t TILE_TONGUE = TONGUE__STATIC;
int32_t TILE_MORTER = MORTER__STATIC;
int32_t TILE_SHRINKEREXPLOSION = SHRINKEREXPLOSION__STATIC;
int32_t TILE_RADIUSEXPLOSION = RADIUSEXPLOSION__STATIC;
int32_t TILE_FORCERIPPLE = FORCERIPPLE__STATIC;
int32_t TILE_LIZTROOP = LIZTROOP__STATIC;
int32_t TILE_LIZTROOPRUNNING = LIZTROOPRUNNING__STATIC;
int32_t TILE_LIZTROOPSTAYPUT = LIZTROOPSTAYPUT__STATIC;
int32_t TILE_LIZTOP = LIZTOP__STATIC;
int32_t TILE_LIZTROOPSHOOT = LIZTROOPSHOOT__STATIC;
int32_t TILE_LIZTROOPJETPACK = LIZTROOPJETPACK__STATIC;
int32_t TILE_LIZTROOPDSPRITE = LIZTROOPDSPRITE__STATIC;
int32_t TILE_LIZTROOPONTOILET = LIZTROOPONTOILET__STATIC;
int32_t TILE_LIZTROOPJUSTSIT = LIZTROOPJUSTSIT__STATIC;
int32_t TILE_LIZTROOPDUCKING = LIZTROOPDUCKING__STATIC;
int32_t TILE_HEADJIB1 = HEADJIB1__STATIC;
int32_t TILE_ARMJIB1 = ARMJIB1__STATIC;
int32_t TILE_LEGJIB1 = LEGJIB1__STATIC;
int32_t TILE_CANNONBALL = CANNONBALL__STATIC;
int32_t TILE_OCTABRAIN = OCTABRAIN__STATIC;
int32_t TILE_OCTABRAINSTAYPUT = OCTABRAINSTAYPUT__STATIC;
int32_t TILE_OCTATOP = OCTATOP__STATIC;
int32_t TILE_OCTADEADSPRITE = OCTADEADSPRITE__STATIC;
int32_t TILE_INNERJAW = INNERJAW__STATIC;
int32_t TILE_DRONE = DRONE__STATIC;
int32_t TILE_EXPLOSION2 = EXPLOSION2__STATIC;
int32_t TILE_COMMANDER = COMMANDER__STATIC;
int32_t TILE_COMMANDERSTAYPUT = COMMANDERSTAYPUT__STATIC;
int32_t TILE_RECON = RECON__STATIC;
int32_t TILE_TANK = TANK__STATIC;
int32_t TILE_PIGCOP = PIGCOP__STATIC;
int32_t TILE_PIGCOPSTAYPUT = PIGCOPSTAYPUT__STATIC;
int32_t TILE_PIGCOPDIVE = PIGCOPDIVE__STATIC;
int32_t TILE_PIGCOPDEADSPRITE = PIGCOPDEADSPRITE__STATIC;
int32_t TILE_PIGTOP = PIGTOP__STATIC;
int32_t TILE_LIZMAN = LIZMAN__STATIC;
int32_t TILE_LIZMANSTAYPUT = LIZMANSTAYPUT__STATIC;
int32_t TILE_LIZMANSPITTING = LIZMANSPITTING__STATIC;
int32_t TILE_LIZMANFEEDING = LIZMANFEEDING__STATIC;
int32_t TILE_LIZMANJUMP = LIZMANJUMP__STATIC;
int32_t TILE_LIZMANDEADSPRITE = LIZMANDEADSPRITE__STATIC;
int32_t TILE_FECES = FECES__STATIC;
int32_t TILE_LIZMANHEAD1 = LIZMANHEAD1__STATIC;
int32_t TILE_LIZMANARM1 = LIZMANARM1__STATIC;
int32_t TILE_LIZMANLEG1 = LIZMANLEG1__STATIC;
int32_t TILE_EXPLOSION2BOT = EXPLOSION2BOT__STATIC;
int32_t TILE_USERWEAPON = USERWEAPON__STATIC;
int32_t TILE_HEADERBAR = HEADERBAR__STATIC;
int32_t TILE_JIBS1 = JIBS1__STATIC;
int32_t TILE_JIBS2 = JIBS2__STATIC;
int32_t TILE_JIBS3 = JIBS3__STATIC;
int32_t TILE_JIBS4 = JIBS4__STATIC;
int32_t TILE_JIBS5 = JIBS5__STATIC;
int32_t TILE_BURNING = BURNING__STATIC;
int32_t TILE_FIRE = FIRE__STATIC;
int32_t TILE_JIBS6 = JIBS6__STATIC;
int32_t TILE_BLOODSPLAT1 = BLOODSPLAT1__STATIC;
int32_t TILE_BLOODSPLAT3 = BLOODSPLAT3__STATIC;
int32_t TILE_BLOODSPLAT2 = BLOODSPLAT2__STATIC;
int32_t TILE_BLOODSPLAT4 = BLOODSPLAT4__STATIC;
int32_t TILE_OOZ = OOZ__STATIC;
int32_t TILE_OOZ2 = OOZ2__STATIC;
int32_t TILE_WALLBLOOD1 = WALLBLOOD1__STATIC;
int32_t TILE_WALLBLOOD2 = WALLBLOOD2__STATIC;
int32_t TILE_WALLBLOOD3 = WALLBLOOD3__STATIC;
int32_t TILE_WALLBLOOD4 = WALLBLOOD4__STATIC;
int32_t TILE_WALLBLOOD5 = WALLBLOOD5__STATIC;
int32_t TILE_WALLBLOOD6 = WALLBLOOD6__STATIC;
int32_t TILE_WALLBLOOD7 = WALLBLOOD7__STATIC;
int32_t TILE_WALLBLOOD8 = WALLBLOOD8__STATIC;
int32_t TILE_BURNING2 = BURNING2__STATIC;
int32_t TILE_FIRE2 = FIRE2__STATIC;
int32_t TILE_CRACKKNUCKLES = CRACKKNUCKLES__STATIC;
int32_t TILE_SMALLSMOKE = SMALLSMOKE__STATIC;
int32_t TILE_SMALLSMOKEMAKER = SMALLSMOKEMAKER__STATIC;
int32_t TILE_FLOORFLAME = FLOORFLAME__STATIC;
int32_t TILE_ROTATEGUN = ROTATEGUN__STATIC;
int32_t TILE_GREENSLIME = GREENSLIME__STATIC;
int32_t TILE_WATERDRIPSPLASH = WATERDRIPSPLASH__STATIC;
int32_t TILE_SCRAP6 = SCRAP6__STATIC;
int32_t TILE_SCRAP1 = SCRAP1__STATIC;
int32_t TILE_SCRAP2 = SCRAP2__STATIC;
int32_t TILE_SCRAP3 = SCRAP3__STATIC;
int32_t TILE_SCRAP4 = SCRAP4__STATIC;
int32_t TILE_SCRAP5 = SCRAP5__STATIC;
int32_t TILE_ORGANTIC = ORGANTIC__STATIC;
int32_t TILE_BETAVERSION = BETAVERSION__STATIC;
int32_t TILE_PLAYERISHERE = PLAYERISHERE__STATIC;
int32_t TILE_PLAYERWASHERE = PLAYERWASHERE__STATIC;
int32_t TILE_SELECTDIR = SELECTDIR__STATIC;
int32_t TILE_F1HELP = F1HELP__STATIC;
int32_t TILE_NOTCHON = NOTCHON__STATIC;
int32_t TILE_NOTCHOFF = NOTCHOFF__STATIC;
int32_t TILE_GROWSPARK = GROWSPARK__STATIC;
int32_t TILE_DUKEICON = DUKEICON__STATIC;
int32_t TILE_BADGUYICON = BADGUYICON__STATIC;
int32_t TILE_FOODICON = FOODICON__STATIC;
int32_t TILE_GETICON = GETICON__STATIC;
int32_t TILE_MENUSCREEN = MENUSCREEN__STATIC;
int32_t TILE_MENUBAR = MENUBAR__STATIC;
int32_t TILE_KILLSICON = KILLSICON__STATIC;
int32_t TILE_FIRSTAID_ICON = FIRSTAID_ICON__STATIC;
int32_t TILE_HEAT_ICON = HEAT_ICON__STATIC;
int32_t TILE_BOTTOMSTATUSBAR = BOTTOMSTATUSBAR__STATIC;
int32_t TILE_BOOT_ICON = BOOT_ICON__STATIC;
int32_t TILE_FRAGBAR = FRAGBAR__STATIC;
int32_t TILE_JETPACK_ICON = JETPACK_ICON__STATIC;
int32_t TILE_AIRTANK_ICON = AIRTANK_ICON__STATIC;
int32_t TILE_STEROIDS_ICON = STEROIDS_ICON__STATIC;
int32_t TILE_HOLODUKE_ICON = HOLODUKE_ICON__STATIC;
int32_t TILE_ACCESS_ICON = ACCESS_ICON__STATIC;
int32_t TILE_DIGITALNUM = DIGITALNUM__STATIC;
int32_t TILE_DUKECAR = DUKECAR__STATIC;
int32_t TILE_CAMCORNER = CAMCORNER__STATIC;
int32_t TILE_CAMLIGHT = CAMLIGHT__STATIC;
int32_t TILE_LOGO = LOGO__STATIC;
int32_t TILE_TITLE = TITLE__STATIC;
int32_t TILE_NUKEWARNINGICON = NUKEWARNINGICON__STATIC;
int32_t TILE_MOUSECURSOR = MOUSECURSOR__STATIC;
int32_t TILE_SLIDEBAR = SLIDEBAR__STATIC;
int32_t TILE_DREALMS = DREALMS__STATIC;
int32_t TILE_BETASCREEN = BETASCREEN__STATIC;
int32_t TILE_WINDOWBORDER1 = WINDOWBORDER1__STATIC;
int32_t TILE_TEXTBOX = TEXTBOX__STATIC;
int32_t TILE_WINDOWBORDER2 = WINDOWBORDER2__STATIC;
int32_t TILE_DUKENUKEM = DUKENUKEM__STATIC;
int32_t TILE_THREEDEE = THREEDEE__STATIC;
int32_t TILE_INGAMEDUKETHREEDEE = INGAMEDUKETHREEDEE__STATIC;
int32_t TILE_TENSCREEN = TENSCREEN__STATIC;
int32_t TILE_PLUTOPAKSPRITE = PLUTOPAKSPRITE__STATIC;
int32_t TILE_DEVISTATOR = DEVISTATOR__STATIC;
int32_t TILE_KNEE = KNEE__STATIC;
int32_t TILE_CROSSHAIR = CROSSHAIR__STATIC;
int32_t TILE_FIRSTGUN = FIRSTGUN__STATIC;
int32_t TILE_FIRSTGUNRELOAD = FIRSTGUNRELOAD__STATIC;
int32_t TILE_FALLINGCLIP = FALLINGCLIP__STATIC;
int32_t TILE_CLIPINHAND = CLIPINHAND__STATIC;
int32_t TILE_HAND = HAND__STATIC;
int32_t TILE_SHELL = SHELL__STATIC;
int32_t TILE_SHOTGUNSHELL = SHOTGUNSHELL__STATIC;
int32_t TILE_CHAINGUN = CHAINGUN__STATIC;
int32_t TILE_RPGGUN = RPGGUN__STATIC;
int32_t TILE_RPGMUZZLEFLASH = RPGMUZZLEFLASH__STATIC;
int32_t TILE_FREEZE = FREEZE__STATIC;
int32_t TILE_CATLITE = CATLITE__STATIC;
int32_t TILE_SHRINKER = SHRINKER__STATIC;
int32_t TILE_HANDHOLDINGLASER = HANDHOLDINGLASER__STATIC;
int32_t TILE_TRIPBOMB = TRIPBOMB__STATIC;
int32_t TILE_LASERLINE = LASERLINE__STATIC;
int32_t TILE_HANDHOLDINGACCESS = HANDHOLDINGACCESS__STATIC;
int32_t TILE_HANDREMOTE = HANDREMOTE__STATIC;
int32_t TILE_HANDTHROW = HANDTHROW__STATIC;
int32_t TILE_TIP = TIP__STATIC;
int32_t TILE_GLAIR = GLAIR__STATIC;
int32_t TILE_SCUBAMASK = SCUBAMASK__STATIC;
int32_t TILE_SPACEMASK = SPACEMASK__STATIC;
int32_t TILE_FORCESPHERE = FORCESPHERE__STATIC;
int32_t TILE_SHOTSPARK1 = SHOTSPARK1__STATIC;
int32_t TILE_RPG = RPG__STATIC;
int32_t TILE_LASERSITE = LASERSITE__STATIC;
int32_t TILE_SHOTGUN = SHOTGUN__STATIC;
int32_t TILE_BOSS1 = BOSS1__STATIC;
int32_t TILE_BOSS1STAYPUT = BOSS1STAYPUT__STATIC;
int32_t TILE_BOSS1SHOOT = BOSS1SHOOT__STATIC;
int32_t TILE_BOSS1LOB = BOSS1LOB__STATIC;
int32_t TILE_BOSSTOP = BOSSTOP__STATIC;
int32_t TILE_BOSS2 = BOSS2__STATIC;
int32_t TILE_BOSS3 = BOSS3__STATIC;
int32_t TILE_SPINNINGNUKEICON = SPINNINGNUKEICON__STATIC;
int32_t TILE_BIGFNTCURSOR = BIGFNTCURSOR__STATIC;
int32_t TILE_SMALLFNTCURSOR = SMALLFNTCURSOR__STATIC;
int32_t TILE_STARTALPHANUM = STARTALPHANUM__STATIC;
int32_t TILE_ENDALPHANUM = ENDALPHANUM__STATIC;
int32_t TILE_BIGALPHANUM = BIGALPHANUM__STATIC;
int32_t TILE_BIGPERIOD = BIGPERIOD__STATIC;
int32_t TILE_BIGCOMMA = BIGCOMMA__STATIC;
// "BIGX" clashes on the Wii?
int32_t TILE_BIGX_ = BIGX__STATIC;
int32_t TILE_BIGQ = BIGQ__STATIC;
int32_t TILE_BIGSEMI = BIGSEMI__STATIC;
int32_t TILE_BIGCOLIN = BIGCOLIN__STATIC;
int32_t TILE_THREEBYFIVE = THREEBYFIVE__STATIC;
int32_t TILE_BIGAPPOS = BIGAPPOS__STATIC;
int32_t TILE_BLANK = BLANK__STATIC;
int32_t TILE_MINIFONT = MINIFONT__STATIC;
int32_t TILE_BUTTON1 = BUTTON1__STATIC;
int32_t TILE_GLASS3 = GLASS3__STATIC;
int32_t TILE_RESPAWNMARKERRED = RESPAWNMARKERRED__STATIC;
int32_t TILE_RESPAWNMARKERYELLOW = RESPAWNMARKERYELLOW__STATIC;
int32_t TILE_RESPAWNMARKERGREEN = RESPAWNMARKERGREEN__STATIC;
int32_t TILE_BONUSSCREEN = BONUSSCREEN__STATIC;
int32_t TILE_VIEWBORDER = VIEWBORDER__STATIC;
int32_t TILE_VICTORY1 = VICTORY1__STATIC;
int32_t TILE_ORDERING = ORDERING__STATIC;
int32_t TILE_TEXTSTORY = TEXTSTORY__STATIC;
int32_t TILE_LOADSCREEN = LOADSCREEN__STATIC;
int32_t TILE_BORNTOBEWILDSCREEN = BORNTOBEWILDSCREEN__STATIC;
int32_t TILE_BLIMP = BLIMP__STATIC;
int32_t TILE_FEM9 = FEM9__STATIC;
int32_t TILE_FOOTPRINT = FOOTPRINT__STATIC;
int32_t TILE_POOP = POOP__STATIC;
int32_t TILE_FRAMEEFFECT1 = FRAMEEFFECT1__STATIC;
int32_t TILE_PANNEL3 = PANNEL3__STATIC;
int32_t TILE_SCREENBREAK14 = SCREENBREAK14__STATIC;
int32_t TILE_SCREENBREAK15 = SCREENBREAK15__STATIC;
int32_t TILE_SCREENBREAK19 = SCREENBREAK19__STATIC;
int32_t TILE_SCREENBREAK16 = SCREENBREAK16__STATIC;
int32_t TILE_SCREENBREAK17 = SCREENBREAK17__STATIC;
int32_t TILE_SCREENBREAK18 = SCREENBREAK18__STATIC;
int32_t TILE_W_TECHWALL11 = W_TECHWALL11__STATIC;
int32_t TILE_W_TECHWALL12 = W_TECHWALL12__STATIC;
int32_t TILE_W_TECHWALL13 = W_TECHWALL13__STATIC;
int32_t TILE_W_TECHWALL14 = W_TECHWALL14__STATIC;
int32_t TILE_W_TECHWALL5 = W_TECHWALL5__STATIC;
int32_t TILE_W_TECHWALL6 = W_TECHWALL6__STATIC;
int32_t TILE_W_TECHWALL7 = W_TECHWALL7__STATIC;
int32_t TILE_W_TECHWALL8 = W_TECHWALL8__STATIC;
int32_t TILE_W_TECHWALL9 = W_TECHWALL9__STATIC;
int32_t TILE_BPANNEL3 = BPANNEL3__STATIC;
int32_t TILE_W_HITTECHWALL16 = W_HITTECHWALL16__STATIC;
int32_t TILE_W_HITTECHWALL10 = W_HITTECHWALL10__STATIC;
int32_t TILE_W_HITTECHWALL15 = W_HITTECHWALL15__STATIC;
int32_t TILE_W_MILKSHELF = W_MILKSHELF__STATIC;
int32_t TILE_W_MILKSHELFBROKE = W_MILKSHELFBROKE__STATIC;
int32_t TILE_PURPLELAVA = PURPLELAVA__STATIC;
int32_t TILE_LAVABUBBLE = LAVABUBBLE__STATIC;
int32_t TILE_DUKECUTOUT = DUKECUTOUT__STATIC;
int32_t TILE_TARGET = TARGET__STATIC;
int32_t TILE_GUNPOWDERBARREL = GUNPOWDERBARREL__STATIC;
int32_t TILE_DUCK = DUCK__STATIC;
int32_t TILE_HATRACK = HATRACK__STATIC;
int32_t TILE_DESKLAMP = DESKLAMP__STATIC;
int32_t TILE_COFFEEMACHINE = COFFEEMACHINE__STATIC;
int32_t TILE_CUPS = CUPS__STATIC;
int32_t TILE_GAVALS = GAVALS__STATIC;
int32_t TILE_GAVALS2 = GAVALS2__STATIC;
int32_t TILE_POLICELIGHTPOLE = POLICELIGHTPOLE__STATIC;
int32_t TILE_FLOORBASKET = FLOORBASKET__STATIC;
int32_t TILE_PUKE = PUKE__STATIC;
int32_t TILE_DOORTILE23 = DOORTILE23__STATIC;
int32_t TILE_TOPSECRET = TOPSECRET__STATIC;
int32_t TILE_SPEAKER = SPEAKER__STATIC;
int32_t TILE_TEDDYBEAR = TEDDYBEAR__STATIC;
int32_t TILE_ROBOTDOG = ROBOTDOG__STATIC;
int32_t TILE_ROBOTPIRATE = ROBOTPIRATE__STATIC;
int32_t TILE_ROBOTMOUSE = ROBOTMOUSE__STATIC;
int32_t TILE_MAIL = MAIL__STATIC;
int32_t TILE_MAILBAG = MAILBAG__STATIC;
int32_t TILE_HOTMEAT = HOTMEAT__STATIC;
int32_t TILE_COFFEEMUG = COFFEEMUG__STATIC;
int32_t TILE_DONUTS2 = DONUTS2__STATIC;
int32_t TILE_TRIPODCAMERA = TRIPODCAMERA__STATIC;
int32_t TILE_METER = METER__STATIC;
int32_t TILE_DESKPHONE = DESKPHONE__STATIC;
int32_t TILE_GUMBALLMACHINE = GUMBALLMACHINE__STATIC;
int32_t TILE_GUMBALLMACHINEBROKE = GUMBALLMACHINEBROKE__STATIC;
int32_t TILE_PAPER = PAPER__STATIC;
int32_t TILE_MACE = MACE__STATIC;
int32_t TILE_GENERICPOLE2 = GENERICPOLE2__STATIC;
int32_t TILE_XXXSTACY = XXXSTACY__STATIC;
int32_t TILE_WETFLOOR = WETFLOOR__STATIC;
int32_t TILE_BROOM = BROOM__STATIC;
int32_t TILE_MOP = MOP__STATIC;
int32_t TILE_LETTER = LETTER__STATIC;
int32_t TILE_PIRATE1A = PIRATE1A__STATIC;
int32_t TILE_PIRATE4A = PIRATE4A__STATIC;
int32_t TILE_PIRATE2A = PIRATE2A__STATIC;
int32_t TILE_PIRATE5A = PIRATE5A__STATIC;
int32_t TILE_PIRATE3A = PIRATE3A__STATIC;
int32_t TILE_PIRATE6A = PIRATE6A__STATIC;
int32_t TILE_PIRATEHALF = PIRATEHALF__STATIC;
int32_t TILE_CHESTOFGOLD = CHESTOFGOLD__STATIC;
int32_t TILE_SIDEBOLT1 = SIDEBOLT1__STATIC;
int32_t TILE_FOODOBJECT1 = FOODOBJECT1__STATIC;
int32_t TILE_FOODOBJECT2 = FOODOBJECT2__STATIC;
int32_t TILE_FOODOBJECT3 = FOODOBJECT3__STATIC;
int32_t TILE_FOODOBJECT4 = FOODOBJECT4__STATIC;
int32_t TILE_FOODOBJECT5 = FOODOBJECT5__STATIC;
int32_t TILE_FOODOBJECT6 = FOODOBJECT6__STATIC;
int32_t TILE_FOODOBJECT7 = FOODOBJECT7__STATIC;
int32_t TILE_FOODOBJECT8 = FOODOBJECT8__STATIC;
int32_t TILE_FOODOBJECT9 = FOODOBJECT9__STATIC;
int32_t TILE_FOODOBJECT10 = FOODOBJECT10__STATIC;
int32_t TILE_FOODOBJECT11 = FOODOBJECT11__STATIC;
int32_t TILE_FOODOBJECT12 = FOODOBJECT12__STATIC;
int32_t TILE_FOODOBJECT13 = FOODOBJECT13__STATIC;
int32_t TILE_FOODOBJECT14 = FOODOBJECT14__STATIC;
int32_t TILE_FOODOBJECT15 = FOODOBJECT15__STATIC;
int32_t TILE_FOODOBJECT16 = FOODOBJECT16__STATIC;
int32_t TILE_FOODOBJECT17 = FOODOBJECT17__STATIC;
int32_t TILE_FOODOBJECT18 = FOODOBJECT18__STATIC;
int32_t TILE_FOODOBJECT19 = FOODOBJECT19__STATIC;
int32_t TILE_FOODOBJECT20 = FOODOBJECT20__STATIC;
int32_t TILE_HEADLAMP = HEADLAMP__STATIC;
int32_t TILE_TAMPON = TAMPON__STATIC;
int32_t TILE_SKINNEDCHICKEN = SKINNEDCHICKEN__STATIC;
int32_t TILE_FEATHEREDCHICKEN = FEATHEREDCHICKEN__STATIC;
int32_t TILE_ROBOTDOG2 = ROBOTDOG2__STATIC;
int32_t TILE_JOLLYMEAL = JOLLYMEAL__STATIC;
int32_t TILE_DUKEBURGER = DUKEBURGER__STATIC;
int32_t TILE_SHOPPINGCART = SHOPPINGCART__STATIC;
int32_t TILE_CANWITHSOMETHING2 = CANWITHSOMETHING2__STATIC;
int32_t TILE_CANWITHSOMETHING3 = CANWITHSOMETHING3__STATIC;
int32_t TILE_CANWITHSOMETHING4 = CANWITHSOMETHING4__STATIC;
int32_t TILE_SNAKEP = SNAKEP__STATIC;
int32_t TILE_DOLPHIN1 = DOLPHIN1__STATIC;
int32_t TILE_DOLPHIN2 = DOLPHIN2__STATIC;
int32_t TILE_NEWBEAST = NEWBEAST__STATIC;
int32_t TILE_NEWBEASTSTAYPUT = NEWBEASTSTAYPUT__STATIC;
int32_t TILE_NEWBEASTJUMP = NEWBEASTJUMP__STATIC;
int32_t TILE_NEWBEASTHANG = NEWBEASTHANG__STATIC;
int32_t TILE_NEWBEASTHANGDEAD = NEWBEASTHANGDEAD__STATIC;
int32_t TILE_BOSS4 = BOSS4__STATIC;
int32_t TILE_BOSS4STAYPUT = BOSS4STAYPUT__STATIC;
int32_t TILE_FEM10 = FEM10__STATIC;
int32_t TILE_TOUGHGAL = TOUGHGAL__STATIC;
int32_t TILE_MAN = MAN__STATIC;
int32_t TILE_MAN2 = MAN2__STATIC;
int32_t TILE_WOMAN = WOMAN__STATIC;
int32_t TILE_PLEASEWAIT = PLEASEWAIT__STATIC;
int32_t TILE_NATURALLIGHTNING = NATURALLIGHTNING__STATIC;
int32_t TILE_WEATHERWARN = WEATHERWARN__STATIC;
int32_t TILE_DUKETAG = DUKETAG__STATIC;
int32_t TILE_SIGN1 = SIGN1__STATIC;
int32_t TILE_SIGN2 = SIGN2__STATIC;
int32_t TILE_JURYGUY = JURYGUY__STATIC;

int32_t TILE_RRTILE11 = 0;
int32_t TILE_RPG2SPRITE = 0;
int32_t TILE_RRTILE18 = 0;
int32_t TILE_RRTILE19 = 0;
int32_t TILE_RRTILE34 = 0;
int32_t TILE_RRTILE35 = 0;
int32_t TILE_DESTRUCTO = 0;
int32_t TILE_RRTILE38 = 0;
int32_t TILE_RRTILE43 = 0;
int32_t TILE_GUTMETER = 0;
int32_t TILE_RRTILE63 = 0;
int32_t TILE_RRTILE64 = 0;
int32_t TILE_RRTILE65 = 0;
int32_t TILE_RRTILE66 = 0;
int32_t TILE_RRTILE67 = 0;
int32_t TILE_RRTILE68 = 0;
int32_t TILE_SOUNDFX = 0;
int32_t TILE_MOTOAMMO = 0;
int32_t TILE_UFOBEAM = 0;
int32_t TILE_RRTILE280 = 0;
int32_t TILE_RRTILE281 = 0;
int32_t TILE_RRTILE282 = 0;
int32_t TILE_RRTILE283 = 0;
int32_t TILE_RRTILE285 = 0;
int32_t TILE_RRTILE286 = 0;
int32_t TILE_RRTILE287 = 0;
int32_t TILE_RRTILE288 = 0;
int32_t TILE_RRTILE289 = 0;
int32_t TILE_RRTILE290 = 0;
int32_t TILE_RRTILE291 = 0;
int32_t TILE_RRTILE292 = 0;
int32_t TILE_RRTILE293 = 0;
int32_t TILE_RRTILE295 = 0;
int32_t TILE_RRTILE296 = 0;
int32_t TILE_RRTILE297 = 0;
int32_t TILE_CDPLAYER = 0;
int32_t TILE_RRTILE380 = 0;
int32_t TILE_RRTILE403 = 0;
int32_t TILE_RRTILE409 = 0;
int32_t TILE_GUTMETER_LIGHT1 = 0;
int32_t TILE_GUTMETER_LIGHT2 = 0;
int32_t TILE_GUTMETER_LIGHT3 = 0;
int32_t TILE_GUTMETER_LIGHT4 = 0;
int32_t TILE_AMMO_ICON = 0;
int32_t TILE_RRTILE1076 = 0;
int32_t TILE_MUD = 0;
int32_t TILE_EXPLOSION3 = 0;
int32_t TILE_RRTILE1636 = 0;
int32_t TILE_WEAPONBAR = 0;
int32_t TILE_RRTILE1752 = 0;
int32_t TILE_RPG2 = 0;
int32_t TILE_RRTILE1790 = 0;
int32_t TILE_RRTILE1792 = 0;
int32_t TILE_RRTILE1801 = 0;
int32_t TILE_RRTILE1805 = 0;
int32_t TILE_RRTILE1807 = 0;
int32_t TILE_RRTILE1808 = 0;
int32_t TILE_RRTILE1812 = 0;
int32_t TILE_RRTILE1814 = 0;
int32_t TILE_RRTILE1817 = 0;
int32_t TILE_RRTILE1821 = 0;
int32_t TILE_RRTILE1824 = 0;
int32_t TILE_RRTILE1826 = 0;
int32_t TILE_RRTILE1850 = 0;
int32_t TILE_RRTILE1851 = 0;
int32_t TILE_RRTILE1856 = 0;
int32_t TILE_RRTILE1877 = 0;
int32_t TILE_RRTILE1878 = 0;
int32_t TILE_RRTILE1938 = 0;
int32_t TILE_RRTILE1939 = 0;
int32_t TILE_RRTILE1942 = 0;
int32_t TILE_RRTILE1944 = 0;
int32_t TILE_RRTILE1945 = 0;
int32_t TILE_RRTILE1947 = 0;
int32_t TILE_RRTILE1951 = 0;
int32_t TILE_RRTILE1952 = 0;
int32_t TILE_RRTILE1953 = 0;
int32_t TILE_RRTILE1961 = 0;
int32_t TILE_RRTILE1964 = 0;
int32_t TILE_RRTILE1973 = 0;
int32_t TILE_RRTILE1985 = 0;
int32_t TILE_RRTILE1986 = 0;
int32_t TILE_RRTILE1987 = 0;
int32_t TILE_RRTILE1988 = 0;
int32_t TILE_RRTILE1990 = 0;
int32_t TILE_RRTILE1995 = 0;
int32_t TILE_RRTILE1996 = 0;
int32_t TILE_RRTILE2004 = 0;
int32_t TILE_RRTILE2005 = 0;
int32_t TILE_POPCORN = 0;
int32_t TILE_RRTILE2022 = 0;
int32_t TILE_LANEPICS = 0;
int32_t TILE_RRTILE2025 = 0;
int32_t TILE_RRTILE2026 = 0;
int32_t TILE_RRTILE2027 = 0;
int32_t TILE_RRTILE2028 = 0;
int32_t TILE_RRTILE2034 = 0;
int32_t TILE_RRTILE2050 = 0;
int32_t TILE_RRTILE2052 = 0;
int32_t TILE_RRTILE2053 = 0;
int32_t TILE_RRTILE2056 = 0;
int32_t TILE_RRTILE2060 = 0;
int32_t TILE_RRTILE2072 = 0;
int32_t TILE_RRTILE2074 = 0;
int32_t TILE_RRTILE2075 = 0;
int32_t TILE_RRTILE2083 = 0;
int32_t TILE_RRTILE2097 = 0;
int32_t TILE_RRTILE2121 = 0;
int32_t TILE_RRTILE2122 = 0;
int32_t TILE_RRTILE2123 = 0;
int32_t TILE_RRTILE2124 = 0;
int32_t TILE_RRTILE2125 = 0;
int32_t TILE_RRTILE2126 = 0;
int32_t TILE_RRTILE2137 = 0;
int32_t TILE_RRTILE2132 = 0;
int32_t TILE_RRTILE2136 = 0;
int32_t TILE_RRTILE2139 = 0;
int32_t TILE_RRTILE2150 = 0;
int32_t TILE_RRTILE2151 = 0;
int32_t TILE_RRTILE2152 = 0;
int32_t TILE_RRTILE2156 = 0;
int32_t TILE_RRTILE2157 = 0;
int32_t TILE_RRTILE2158 = 0;
int32_t TILE_RRTILE2159 = 0;
int32_t TILE_RRTILE2160 = 0;
int32_t TILE_RRTILE2161 = 0;
int32_t TILE_RRTILE2175 = 0;
int32_t TILE_RRTILE2176 = 0;
int32_t TILE_RRTILE2178 = 0;
int32_t TILE_RRTILE2186 = 0;
int32_t TILE_RRTILE2214 = 0;
int32_t TILE_RRTILE2319 = 0;
int32_t TILE_RRTILE2321 = 0;
int32_t TILE_RRTILE2326 = 0;
int32_t TILE_RRTILE2329 = 0;
int32_t TILE_RRTILE2357 = 0;
int32_t TILE_RRTILE2382 = 0;
int32_t TILE_RRTILE2430 = 0;
int32_t TILE_RRTILE2431 = 0;
int32_t TILE_RRTILE2432 = 0;
int32_t TILE_RRTILE2437 = 0;
int32_t TILE_RRTILE2443 = 0;
int32_t TILE_RRTILE2445 = 0;
int32_t TILE_RRTILE2446 = 0;
int32_t TILE_RRTILE2450 = 0;
int32_t TILE_RRTILE2451 = 0;
int32_t TILE_RRTILE2455 = 0;
int32_t TILE_RRTILE2460 = 0;
int32_t TILE_RRTILE2465 = 0;
int32_t TILE_RRTILE2560 = 0;
int32_t TILE_RRTILE2562 = 0;
int32_t TILE_RRTILE2564 = 0;
int32_t TILE_RRTILE2573 = 0;
int32_t TILE_RRTILE2574 = 0;
int32_t TILE_RRTILE2577 = 0;
int32_t TILE_RRTILE2578 = 0;
int32_t TILE_RRTILE2581 = 0;
int32_t TILE_RRTILE2583 = 0;
int32_t TILE_RRTILE2604 = 0;
int32_t TILE_RRTILE2610 = 0;
int32_t TILE_RRTILE2613 = 0;
int32_t TILE_RRTILE2621 = 0;
int32_t TILE_RRTILE2622 = 0;
int32_t TILE_RRTILE2636 = 0;
int32_t TILE_RRTILE2637 = 0;
int32_t TILE_RRTILE2654 = 0;
int32_t TILE_RRTILE2656 = 0;
int32_t TILE_RRTILE2676 = 0;
int32_t TILE_RRTILE2689 = 0;
int32_t TILE_RRTILE2697 = 0;
int32_t TILE_RRTILE2702 = 0;
int32_t TILE_RRTILE2707 = 0;
int32_t TILE_RRTILE2732 = 0;
int32_t TILE_RRTILE2030 = 0;
int32_t TILE_RRTILE2831 = 0;
int32_t TILE_RRTILE2832 = 0;
int32_t TILE_RRTILE2842 = 0;
int32_t TILE_RRTILE2859 = 0;
int32_t TILE_RRTILE2876 = 0;
int32_t TILE_RRTILE2878 = 0;
int32_t TILE_RRTILE2879 = 0;
int32_t TILE_RRTILE2893 = 0;
int32_t TILE_RRTILE2894 = 0;
int32_t TILE_RRTILE2898 = 0;
int32_t TILE_RRTILE2899 = 0;
int32_t TILE_RRTILE2915 = 0;
int32_t TILE_RRTILE2940 = 0;
int32_t TILE_RRTILE2944 = 0;
int32_t TILE_RRTILE2945 = 0;
int32_t TILE_RRTILE2946 = 0;
int32_t TILE_RRTILE2947 = 0;
int32_t TILE_RRTILE2948 = 0;
int32_t TILE_RRTILE2949 = 0;
int32_t TILE_RRTILE2961 = 0;
int32_t TILE_RRTILE2970 = 0;
int32_t TILE_RRTILE2977 = 0;
int32_t TILE_RRTILE2978 = 0;
int32_t TILE_RRTILE2990 = 0;
int32_t TILE_RRTILE3073 = 0;
int32_t TILE_RRTILE3083 = 0;
int32_t TILE_RRTILE3100 = 0;
int32_t TILE_RRTILE3114 = 0;
int32_t TILE_RRTILE3115 = 0;
int32_t TILE_RRTILE3116 = 0;
int32_t TILE_RRTILE3117 = 0;
int32_t TILE_RRTILE3120 = 0;
int32_t TILE_RRTILE3121 = 0;
int32_t TILE_RRTILE3122 = 0;
int32_t TILE_RRTILE3123 = 0;
int32_t TILE_RRTILE3124 = 0;
int32_t TILE_RRTILE3132 = 0;
int32_t TILE_RRTILE3139 = 0;
int32_t TILE_RRTILE3144 = 0;
int32_t TILE_RRTILE3152 = 0;
int32_t TILE_RRTILE3153 = 0;
int32_t TILE_RRTILE3155 = 0;
int32_t TILE_RRTILE3171 = 0;
int32_t TILE_RRTILE3172 = 0;
int32_t TILE_RRTILE3190 = 0;
int32_t TILE_RRTILE3191 = 0;
int32_t TILE_RRTILE3192 = 0;
int32_t TILE_RRTILE3195 = 0;
int32_t TILE_RRTILE3200 = 0;
int32_t TILE_RRTILE3201 = 0;
int32_t TILE_RRTILE3202 = 0;
int32_t TILE_RRTILE3203 = 0;
int32_t TILE_RRTILE3204 = 0;
int32_t TILE_RRTILE3205 = 0;
int32_t TILE_RRTILE3206 = 0;
int32_t TILE_RRTILE3207 = 0;
int32_t TILE_RRTILE3208 = 0;
int32_t TILE_RRTILE3209 = 0;
int32_t TILE_RRTILE3216 = 0;
int32_t TILE_RRTILE3218 = 0;
int32_t TILE_RRTILE3219 = 0;
int32_t TILE_RRTILE3232 = 0;
int32_t TILE_SHOTGUNSHELLS = 0;
int32_t TILE_CIRCLESTUCK = 0;
int32_t TILE_RRTILE3410 = 0;
int32_t TILE_LUMBERBLADE = 0;
int32_t TILE_BOWLINGBALLH = 0;
int32_t TILE_BOWLINGBALL = 0;
int32_t TILE_BOWLINGBALLSPRITE = 0;
int32_t TILE_POWDERH = 0;
int32_t TILE_RRTILE3440 = 0;
int32_t TILE_RRTILE3462 = 0;
int32_t TILE_OWHIP = 0;
int32_t TILE_UWHIP = 0;
int32_t TILE_RPGGUN2 = 0;
int32_t TILE_RRTILE3497 = 0;
int32_t TILE_RRTILE3498 = 0;
int32_t TILE_RRTILE3499 = 0;
int32_t TILE_RRTILE3500 = 0;
int32_t TILE_SLINGBLADE = 0;
int32_t TILE_RRTILE3584 = 0;
int32_t TILE_RRTILE3586 = 0;
int32_t TILE_RRTILE3587 = 0;
int32_t TILE_RRTILE3600 = 0;
int32_t TILE_RRTILE3631 = 0;
int32_t TILE_RRTILE3635 = 0;
int32_t TILE_RRTILE3637 = 0;
int32_t TILE_RRTILE3643 = 0;
int32_t TILE_RRTILE3647 = 0;
int32_t TILE_RRTILE3652 = 0;
int32_t TILE_RRTILE3653 = 0;
int32_t TILE_RRTILE3668 = 0;
int32_t TILE_RRTILE3671 = 0;
int32_t TILE_RRTILE3673 = 0;
int32_t TILE_RRTILE3677 = 0;
int32_t TILE_RRTILE3684 = 0;
int32_t TILE_RRTILE3708 = 0;
int32_t TILE_RRTILE3714 = 0;
int32_t TILE_RRTILE3716 = 0;
int32_t TILE_RRTILE3720 = 0;
int32_t TILE_RRTILE3723 = 0;
int32_t TILE_RRTILE3725 = 0;
int32_t TILE_RRTILE3737 = 0;
int32_t TILE_RRTILE3754 = 0;
int32_t TILE_RRTILE3762 = 0;
int32_t TILE_RRTILE3763 = 0;
int32_t TILE_RRTILE3764 = 0;
int32_t TILE_RRTILE3765 = 0;
int32_t TILE_RRTILE3767 = 0;
int32_t TILE_RRTILE3773 = 0;
int32_t TILE_RRTILE3774 = 0;
int32_t TILE_RRTILE3793 = 0;
int32_t TILE_RRTILE3795 = 0;
int32_t TILE_RRTILE3804 = 0;
int32_t TILE_RRTILE3814 = 0;
int32_t TILE_RRTILE3815 = 0;
int32_t TILE_RRTILE3819 = 0;
int32_t TILE_RRTILE3827 = 0;
int32_t TILE_RRTILE3837 = 0;
int32_t TILE_RRTILE5014 = 0;
int32_t TILE_RRTILE5016 = 0;
int32_t TILE_RRTILE5017 = 0;
int32_t TILE_RRTILE5018 = 0;
int32_t TILE_RRTILE5019 = 0;
int32_t TILE_RRTILE5020 = 0;
int32_t TILE_RRTILE5021 = 0;
int32_t TILE_RRTILE5022 = 0;
int32_t TILE_RRTILE5023 = 0;
int32_t TILE_RRTILE5024 = 0;
int32_t TILE_RRTILE5025 = 0;
int32_t TILE_RRTILE5026 = 0;
int32_t TILE_RRTILE5027 = 0;
int32_t TILE_RRTILE5029 = 0;
int32_t TILE_RRTILE5030 = 0;
int32_t TILE_RRTILE5031 = 0;
int32_t TILE_RRTILE5032 = 0;
int32_t TILE_RRTILE5033 = 0;
int32_t TILE_RRTILE5034 = 0;
int32_t TILE_RRTILE5035 = 0;
int32_t TILE_RRTILE5036 = 0;
int32_t TILE_RRTILE5037 = 0;
int32_t TILE_RRTILE5038 = 0;
int32_t TILE_RRTILE5039 = 0;
int32_t TILE_RRTILE5040 = 0;
int32_t TILE_RRTILE5041 = 0;
int32_t TILE_RRTILE5043 = 0;
int32_t TILE_RRTILE5044 = 0;
int32_t TILE_RRTILE5045 = 0;
int32_t TILE_RRTILE5046 = 0;
int32_t TILE_RRTILE5047 = 0;
int32_t TILE_RRTILE5048 = 0;
int32_t TILE_RRTILE5049 = 0;
int32_t TILE_RRTILE5050 = 0;
int32_t TILE_RRTILE5051 = 0;
int32_t TILE_RRTILE5052 = 0;
int32_t TILE_RRTILE5053 = 0;
int32_t TILE_RRTILE5054 = 0;
int32_t TILE_RRTILE5055 = 0;
int32_t TILE_RRTILE5056 = 0;
int32_t TILE_RRTILE5057 = 0;
int32_t TILE_RRTILE5058 = 0;
int32_t TILE_RRTILE5059 = 0;
int32_t TILE_RRTILE5061 = 0;
int32_t TILE_RRTILE5062 = 0;
int32_t TILE_RRTILE5063 = 0;
int32_t TILE_RRTILE5064 = 0;
int32_t TILE_RRTILE5065 = 0;
int32_t TILE_RRTILE5066 = 0;
int32_t TILE_RRTILE5067 = 0;
int32_t TILE_RRTILE5068 = 0;
int32_t TILE_RRTILE5069 = 0;
int32_t TILE_RRTILE5070 = 0;
int32_t TILE_RRTILE5071 = 0;
int32_t TILE_RRTILE5072 = 0;
int32_t TILE_RRTILE5073 = 0;
int32_t TILE_RRTILE5074 = 0;
int32_t TILE_RRTILE5075 = 0;
int32_t TILE_RRTILE5076 = 0;
int32_t TILE_RRTILE5077 = 0;
int32_t TILE_RRTILE5078 = 0;
int32_t TILE_RRTILE5079 = 0;
int32_t TILE_RRTILE5080 = 0;
int32_t TILE_RRTILE5081 = 0;
int32_t TILE_RRTILE5082 = 0;
int32_t TILE_RRTILE5083 = 0;
int32_t TILE_RRTILE5084 = 0;
int32_t TILE_RRTILE5085 = 0;
int32_t TILE_RRTILE5086 = 0;
int32_t TILE_RRTILE5087 = 0;
int32_t TILE_RRTILE5088 = 0;
int32_t TILE_RRTILE5090 = 0;
int32_t TILE_RRTILE6144 = 0;
int32_t TILE_RRTILE7110 = 0;
int32_t TILE_RRTILE7111 = 0;
int32_t TILE_RRTILE7112 = 0;
int32_t TILE_RRTILE7113 = 0;
int32_t TILE_MOTOGUN = 0;
int32_t TILE_RRTILE7169 = 0;
int32_t TILE_MOTOHIT = 0;
int32_t TILE_BOATHIT = 0;
int32_t TILE_RRTILE7184 = 0;
int32_t TILE_RRTILE7190 = 0;
int32_t TILE_RRTILE7191 = 0;
int32_t TILE_RRTILE7213 = 0;
int32_t TILE_RRTILE7219 = 0;
int32_t TILE_EMPTYBIKE = 0;
int32_t TILE_EMPTYBOAT = 0;
int32_t TILE_RRTILE7424 = 0;
int32_t TILE_RRTILE7430 = 0;
int32_t TILE_RRTILE7433 = 0;
int32_t TILE_RRTILE7441 = 0;
int32_t TILE_RRTILE7547 = 0;
int32_t TILE_RRTILE7467 = 0;
int32_t TILE_RRTILE7469 = 0;
int32_t TILE_RRTILE7470 = 0;
int32_t TILE_RRTILE7475 = 0;
int32_t TILE_RRTILE7478 = 0;
int32_t TILE_RRTILE7505 = 0;
int32_t TILE_RRTILE7506 = 0;
int32_t TILE_RRTILE7534 = 0;
int32_t TILE_RRTILE7540 = 0;
int32_t TILE_RRTILE7533 = 0;
int32_t TILE_RRTILE7545 = 0;
int32_t TILE_RRTILE7552 = 0;
int32_t TILE_RRTILE7553 = 0;
int32_t TILE_RRTILE7554 = 0;
int32_t TILE_RRTILE7555 = 0;
int32_t TILE_RRTILE7557 = 0;
int32_t TILE_RRTILE7558 = 0;
int32_t TILE_RRTILE7559 = 0;
int32_t TILE_RRTILE7561 = 0;
int32_t TILE_RRTILE7566 = 0;
int32_t TILE_RRTILE7568 = 0;
int32_t TILE_RRTILE7574 = 0;
int32_t TILE_RRTILE7575 = 0;
int32_t TILE_RRTILE7576 = 0;
int32_t TILE_RRTILE7578 = 0;
int32_t TILE_RRTILE7579 = 0;
int32_t TILE_RRTILE7580 = 0;
int32_t TILE_RRTILE7595 = 0;
int32_t TILE_RRTILE7629 = 0;
int32_t TILE_RRTILE7636 = 0;
int32_t TILE_RRTILE7638 = 0;
int32_t TILE_RRTILE7640 = 0;
int32_t TILE_RRTILE7644 = 0;
int32_t TILE_RRTILE7646 = 0;
int32_t TILE_RRTILE7648 = 0;
int32_t TILE_RRTILE7650 = 0;
int32_t TILE_RRTILE7653 = 0;
int32_t TILE_RRTILE7655 = 0;
int32_t TILE_RRTILE7657 = 0;
int32_t TILE_RRTILE7659 = 0;
int32_t TILE_RRTILE7691 = 0;
int32_t TILE_RRTILE7694 = 0;
int32_t TILE_RRTILE7696 = 0;
int32_t TILE_RRTILE7697 = 0;
int32_t TILE_RRTILE7700 = 0;
int32_t TILE_RRTILE7702 = 0;
int32_t TILE_RRTILE7704 = 0;
int32_t TILE_RRTILE7705 = 0;
int32_t TILE_RRTILE7711 = 0;
int32_t TILE_RRTILE7716 = 0;
int32_t TILE_RRTILE7756 = 0;
int32_t TILE_RRTILE7768 = 0;
int32_t TILE_RRTILE7806 = 0;
int32_t TILE_RRTILE7820 = 0;
int32_t TILE_RRTILE7859 = 0;
int32_t TILE_RRTILE7870 = 0;
int32_t TILE_RRTILE7873 = 0;
int32_t TILE_RRTILE7875 = 0;
int32_t TILE_RRTILE7876 = 0;
int32_t TILE_RRTILE7879 = 0;
int32_t TILE_RRTILE7881 = 0;
int32_t TILE_RRTILE7883 = 0;
int32_t TILE_RRTILE7885 = 0;
int32_t TILE_RRTILE7886 = 0;
int32_t TILE_RRTILE7887 = 0;
int32_t TILE_RRTILE7888 = 0;
int32_t TILE_RRTILE7889 = 0;
int32_t TILE_RRTILE7890 = 0;
int32_t TILE_RRTILE7900 = 0;
int32_t TILE_RRTILE7901 = 0;
int32_t TILE_RRTILE7906 = 0;
int32_t TILE_RRTILE7912 = 0;
int32_t TILE_RRTILE7913 = 0;
int32_t TILE_RRTILE7936 = 0;
int32_t TILE_RRTILE8047 = 0;
int32_t TILE_MULTISWITCH2 = 0;
int32_t TILE_RRTILE8059 = 0;
int32_t TILE_RRTILE8060 = 0;
int32_t TILE_RRTILE8063 = 0;
int32_t TILE_RRTILE8067 = 0;
int32_t TILE_RRTILE8076 = 0;
int32_t TILE_RRTILE8094 = 0;
int32_t TILE_RRTILE8096 = 0;
int32_t TILE_RRTILE8099 = 0;
int32_t TILE_RRTILE8106 = 0;
int32_t TILE_RRTILE8162 = 0;
int32_t TILE_RRTILE8163 = 0;
int32_t TILE_RRTILE8164 = 0;
int32_t TILE_RRTILE8165 = 0;
int32_t TILE_RRTILE8166 = 0;
int32_t TILE_RRTILE8167 = 0;
int32_t TILE_RRTILE8168 = 0;
int32_t TILE_RRTILE8192 = 0;
int32_t TILE_RRTILE8193 = 0;
int32_t TILE_RRTILE8215 = 0;
int32_t TILE_RRTILE8216 = 0;
int32_t TILE_RRTILE8217 = 0;
int32_t TILE_RRTILE8218 = 0;
int32_t TILE_RRTILE8220 = 0;
int32_t TILE_RRTILE8221 = 0;
int32_t TILE_RRTILE8222 = 0;
int32_t TILE_RRTILE8223 = 0;
int32_t TILE_RRTILE8224 = 0;
int32_t TILE_RRTILE8227 = 0;
int32_t TILE_RRTILE8312 = 0;
int32_t TILE_RRTILE8370 = 0;
int32_t TILE_RRTILE8371 = 0;
int32_t TILE_RRTILE8372 = 0;
int32_t TILE_RRTILE8373 = 0;
int32_t TILE_RRTILE8379 = 0;
int32_t TILE_RRTILE8380 = 0;
int32_t TILE_RRTILE8385 = 0;
int32_t TILE_RRTILE8386 = 0;
int32_t TILE_RRTILE8387 = 0;
int32_t TILE_RRTILE8388 = 0;
int32_t TILE_RRTILE8389 = 0;
int32_t TILE_RRTILE8390 = 0;
int32_t TILE_RRTILE8391 = 0;
int32_t TILE_RRTILE8392 = 0;
int32_t TILE_RRTILE8394 = 0;
int32_t TILE_RRTILE8395 = 0;
int32_t TILE_RRTILE8396 = 0;
int32_t TILE_RRTILE8397 = 0;
int32_t TILE_RRTILE8398 = 0;
int32_t TILE_RRTILE8399 = 0;
int32_t TILE_RRTILE8423 = 0;
int32_t TILE_RRTILE8448 = 0;
int32_t TILE_RRTILE8450 = 0;
int32_t TILE_BOATAMMO = 0;
int32_t TILE_RRTILE8461 = 0;
int32_t TILE_RRTILE8462 = 0;
int32_t TILE_RRTILE8464 = 0;
int32_t TILE_RRTILE8475 = 0;
int32_t TILE_RRTILE8487 = 0;
int32_t TILE_RRTILE8488 = 0;
int32_t TILE_RRTILE8489 = 0;
int32_t TILE_RRTILE8490 = 0;
int32_t TILE_RRTILE8496 = 0;
int32_t TILE_RRTILE8497 = 0;
int32_t TILE_RRTILE8498 = 0;
int32_t TILE_RRTILE8499 = 0;
int32_t TILE_RRTILE8503 = 0;
int32_t TILE_RRTILE8525 = 0;
int32_t TILE_RRTILE8537 = 0;
int32_t TILE_RRTILE8565 = 0;
int32_t TILE_RRTILE8567 = 0;
int32_t TILE_RRTILE8568 = 0;
int32_t TILE_RRTILE8569 = 0;
int32_t TILE_RRTILE8570 = 0;
int32_t TILE_RRTILE8571 = 0;
int32_t TILE_RRTILE8579 = 0;
int32_t TILE_RRTILE8588 = 0;
int32_t TILE_RRTILE8589 = 0;
int32_t TILE_RRTILE8590 = 0;
int32_t TILE_RRTILE8591 = 0;
int32_t TILE_RRTILE8592 = 0;
int32_t TILE_RRTILE8593 = 0;
int32_t TILE_RRTILE8594 = 0;
int32_t TILE_RRTILE8595 = 0;
int32_t TILE_RRTILE8596 = 0;
int32_t TILE_RRTILE8598 = 0;
int32_t TILE_RRTILE8605 = 0;
int32_t TILE_RRTILE8608 = 0;
int32_t TILE_RRTILE8609 = 0;
int32_t TILE_RRTILE8611 = 0;
int32_t TILE_RRTILE8617 = 0;
int32_t TILE_RRTILE8618 = 0;
int32_t TILE_RRTILE8620 = 0;
int32_t TILE_RRTILE8621 = 0;
int32_t TILE_RRTILE8622 = 0;
int32_t TILE_RRTILE8623 = 0;
int32_t TILE_RRTILE8624 = 0;
int32_t TILE_RRTILE8640 = 0;
int32_t TILE_RRTILE8651 = 0;
int32_t TILE_RRTILE8660 = 0;
int32_t TILE_RRTILE8677 = 0;
int32_t TILE_RRTILE8679 = 0;
int32_t TILE_RRTILE8680 = 0;
int32_t TILE_RRTILE8681 = 0;
int32_t TILE_RRTILE8682 = 0;
int32_t TILE_RRTILE8683 = 0;
int32_t TILE_RRTILE8704 = 0;
int32_t TILE_BOULDER = 0;
int32_t TILE_BOULDER1 = 0;
int32_t TILE_TORNADO = 0;
int32_t TILE_CHEERBOMB = 0;
int32_t TILE_CHEERBLADE = 0;
int32_t TILE_DOGATTACK = 0;
int32_t TILE_BILLYWALK = 0;
int32_t TILE_BILLYDIE = 0;
int32_t TILE_BILLYCOCK = 0;
int32_t TILE_BILLYRAY = 0;
int32_t TILE_BILLYRAYSTAYPUT = 0;
int32_t TILE_BILLYBUT = 0;
int32_t TILE_BILLYSCRATCH = 0;
int32_t TILE_BILLYSNIFF = 0;
int32_t TILE_BILLYWOUND = 0;
int32_t TILE_BILLYGORE = 0;
int32_t TILE_BILLYJIBA = 0;
int32_t TILE_BILLYJIBB = 0;
int32_t TILE_BRAYSNIPER = 0;
int32_t TILE_DOGRUN = 0;
int32_t TILE_DOGDIE = 0;
int32_t TILE_DOGDEAD = 0;
int32_t TILE_DOGBARK = 0;
int32_t TILE_LTH = 0;
int32_t TILE_LTHSTRAFE = 0;
int32_t TILE_HULKHANG = 0;
int32_t TILE_HULKHANGDEAD = 0;
int32_t TILE_HULKJUMP = 0;
int32_t TILE_LTHLOAD = 0;
int32_t TILE_LTHDIE = 0;
int32_t TILE_BUBBASCRATCH = 0;
int32_t TILE_BUBBANOSE = 0;
int32_t TILE_BUBBAPISS = 0;
int32_t TILE_BUBBASTAND = 0;
int32_t TILE_BUBBAOUCH = 0;
int32_t TILE_BUBBADIE = 0;
int32_t TILE_BUBBADEAD = 0;
int32_t TILE_HULK = 0;
int32_t TILE_HULKSTAYPUT = 0;
int32_t TILE_HULKA = 0;
int32_t TILE_HULKB = 0;
int32_t TILE_HULKC = 0;
int32_t TILE_HULKJIBA = 0;
int32_t TILE_HULKJIBB = 0;
int32_t TILE_HULKJIBC = 0;
int32_t TILE_SBSWIPE = 0;
int32_t TILE_SBPAIN = 0;
int32_t TILE_SBDIE = 0;
int32_t TILE_HEN = 0;
int32_t TILE_HENSTAYPUT = 0;
int32_t TILE_HENSTAND = 0;
int32_t TILE_PIG = 0;
int32_t TILE_PIGSTAYPUT = 0;
int32_t TILE_PIGEAT = 0;
int32_t TILE_SBMOVE = 0;
int32_t TILE_SBSPIT = 0;
int32_t TILE_SBDIP = 0;
int32_t TILE_MINION = 0;
int32_t TILE_MINIONSTAYPUT = 0;
int32_t TILE_UFO1 = 0;
int32_t TILE_UFO2 = 0;
int32_t TILE_UFO3 = 0;
int32_t TILE_UFO4 = 0;
int32_t TILE_UFO5 = 0;
int32_t TILE_MINJIBA = 0;
int32_t TILE_MINJIBB = 0;
int32_t TILE_MINJIBC = 0;
int32_t TILE_COW = 0;
int32_t TILE_COOT = 0;
int32_t TILE_COOTSTAYPUT = 0;
int32_t TILE_COOTSHOOT = 0;
int32_t TILE_COOTDIE = 0;
int32_t TILE_COOTDUCK = 0;
int32_t TILE_COOTPAIN = 0;
int32_t TILE_COOTTRANS = 0;
int32_t TILE_COOTGETUP = 0;
int32_t TILE_COOTJIBA = 0;
int32_t TILE_COOTJIBB = 0;
int32_t TILE_COOTJIBC = 0;
int32_t TILE_VIXEN = 0;
int32_t TILE_VIXENPAIN = 0;
int32_t TILE_VIXENDIE = 0;
int32_t TILE_VIXENSHOOT = 0;
int32_t TILE_VIXENWDN = 0;
int32_t TILE_VIXENWUP = 0;
int32_t TILE_VIXENKICK = 0;
int32_t TILE_VIXENTELE = 0;
int32_t TILE_VIXENTEAT = 0;
int32_t TILE_BIKEJIBA = 0;
int32_t TILE_BIKEJIBB = 0;
int32_t TILE_BIKEJIBC = 0;
int32_t TILE_BIKERB = 0;
int32_t TILE_BIKERBV2 = 0;
int32_t TILE_BIKER = 0;
int32_t TILE_BIKERJIBA = 0;
int32_t TILE_BIKERJIBB = 0;
int32_t TILE_BIKERJIBC = 0;
int32_t TILE_BIKERJIBD = 0;
int32_t TILE_MAKEOUT = 0;
int32_t TILE_CHEERB = 0;
int32_t TILE_CHEER = 0;
int32_t TILE_CHEERSTAYPUT = 0;
int32_t TILE_CHEERJIBA = 0;
int32_t TILE_CHEERJIBB = 0;
int32_t TILE_CHEERJIBC = 0;
int32_t TILE_CHEERJIBD = 0;
int32_t TILE_FBOATJIBA = 0;
int32_t TILE_FBOATJIBB = 0;
int32_t TILE_COOTPLAY = 0;
int32_t TILE_BILLYPLAY = 0;
int32_t TILE_MINIONBOAT = 0;
int32_t TILE_HULKBOAT = 0;
int32_t TILE_CHEERBOAT = 0;
int32_t TILE_RRTILE7274 = 0;
int32_t TILE_RABBIT = 0;
int32_t TILE_RABBITJIBA = 0;
int32_t TILE_RABBITJIBB = 0;
int32_t TILE_RABBITJIBC = 0;
int32_t TILE_ROCK = 0;
int32_t TILE_ROCK2 = 0;
int32_t TILE_MAMACLOUD = 0;
int32_t TILE_MAMA = 0;
int32_t TILE_MAMAJIBA = 0;
int32_t TILE_MAMAJIBB = 0;

#if 0
static hashtable_t h_names = {512, NULL};

void G_ProcessDynamicTileMapping(const char *szLabel, int32_t lValue)
{
    int32_t i;

    if ((unsigned)lValue >= MAXTILES || !szLabel)
        return;

    i = hash_find(&h_names,szLabel);
    if (i>=0)
    {
        struct dynitem *di = &g_dynTileList[i];
#ifdef DEBUGGINGAIDS
        if (g_scriptDebug && di->staticval != lValue)
            Printf("REMAP %s (%d) --> %d\n", di->str, di->staticval, lValue);
#endif
        *di->dynvalptr = lValue;
    }
}

void inithashnames(void)
{
    int32_t i;

    hash_init(&h_names);

    for (i=0; g_dynTileList[i].staticval >= 0; i++)
        hash_add(&h_names, g_dynTileList[i].str, i, 0);
}

void freehashnames(void)
{
    hash_free(&h_names);
}
#endif

int APLAYER, CAMERA1;

// This is run after all CON define's have been processed to set up the
// dynamic->static tile mapping.
void G_InitDynamicTiles(void)
{
    int32_t i;

    // this cannot include the constant headers so it has to use literal numbers
    // Not sure yet about the weapons - this should probably be done by using the real RR constants where applicable
    if (!isRR())
    {
        APLAYER = 3845;
        CAMERA1 = 1134;
    }
    else
    {
        APLAYER = 1405;
        CAMERA1 = 621;
    }



    Bmemset(DynamicTileMap, 0, sizeof(DynamicTileMap));

    if (RR)
    {
        for (i = 0; g_dynTileList[i].staticval >= 0; i++) {
            *(g_dynTileList[i].dynvalptr) = -g_dynTileList[i].staticval_rr;
            if (RRRA && *(g_dynTileList[i].dynvalptr) == -UFO1__STATICRR) *(g_dynTileList[i].dynvalptr) = -UFO1__STATICRRRA;
            DynamicTileMap[*(g_dynTileList[i].dynvalptr)] = g_dynTileList[i].staticval_rr;
			NameToTileIndex.Insert(g_dynTileList[i].str, *(g_dynTileList[i].dynvalptr));
        }
        PHEIGHT = PHEIGHT_RR;
    }

    for (i=0; g_dynTileList[i].staticval >= 0; i++)
        if (g_dynTileList[i].staticval > 0)
		{
            DynamicTileMap[*(g_dynTileList[i].dynvalptr)] = g_dynTileList[i].staticval;
			NameToTileIndex.Insert(g_dynTileList[i].str, *(g_dynTileList[i].dynvalptr));
		}

    DynamicTileMap[0] = 0;

    g_blimpSpawnItems[0] = TILE_RPGSPRITE;
    g_blimpSpawnItems[1] = TILE_CHAINGUNSPRITE;
    g_blimpSpawnItems[2] = TILE_DEVISTATORAMMO;
    g_blimpSpawnItems[3] = TILE_RPGAMMO;
    g_blimpSpawnItems[4] = TILE_RPGAMMO;
    g_blimpSpawnItems[5] = TILE_JETPACK;
    g_blimpSpawnItems[6] = TILE_SHIELD;
    g_blimpSpawnItems[7] = TILE_FIRSTAID;
    g_blimpSpawnItems[8] = TILE_STEROIDS;
    g_blimpSpawnItems[9] = TILE_RPGAMMO;
    g_blimpSpawnItems[10] = TILE_RPGAMMO;
    g_blimpSpawnItems[11] = TILE_RPGSPRITE;
    g_blimpSpawnItems[12] = TILE_RPGAMMO;
    g_blimpSpawnItems[13] = TILE_FREEZESPRITE;
    g_blimpSpawnItems[14] = TILE_FREEZEAMMO;

    if (RR)
    {
        WeaponPickupSprites[0] = TILE_KNEE;
        WeaponPickupSprites[1] = TILE_FIRSTGUNSPRITE;
        WeaponPickupSprites[2] = TILE_SHOTGUNSPRITE;
        WeaponPickupSprites[3] = TILE_CHAINGUNSPRITE;
        WeaponPickupSprites[4] = TILE_RPGSPRITE;
        WeaponPickupSprites[5] = TILE_HEAVYHBOMB;
        WeaponPickupSprites[6] = TILE_SHRINKERSPRITE;
        WeaponPickupSprites[7] = TILE_DEVISTATORSPRITE;
        WeaponPickupSprites[8] = TILE_TRIPBOMBSPRITE;
        WeaponPickupSprites[9] = TILE_BOWLINGBALLSPRITE;
        WeaponPickupSprites[10] = TILE_FREEZEBLAST;
        WeaponPickupSprites[11] = TILE_HEAVYHBOMB;
    }
    else
    {
        WeaponPickupSprites[0] = TILE_KNEE;
        WeaponPickupSprites[1] = TILE_FIRSTGUNSPRITE;
        WeaponPickupSprites[2] = TILE_SHOTGUNSPRITE;
        WeaponPickupSprites[3] = TILE_CHAINGUNSPRITE;
        WeaponPickupSprites[4] = TILE_RPGSPRITE;
        WeaponPickupSprites[5] = TILE_HEAVYHBOMB;
        WeaponPickupSprites[6] = TILE_SHRINKERSPRITE;
        WeaponPickupSprites[7] = TILE_DEVISTATORSPRITE;
        WeaponPickupSprites[8] = TILE_TRIPBOMBSPRITE;
        WeaponPickupSprites[9] = TILE_FREEZESPRITE;
        WeaponPickupSprites[10] = TILE_HEAVYHBOMB;
        WeaponPickupSprites[11] = TILE_SHRINKERSPRITE;
    }

    // ouch... the big background image takes up a fuckload of memory and takes a second to load!
#ifdef EDUKE32_GLES
    TILE_MENUSCREEN = TILE_LOADSCREEN = TILE_BETASCREEN;
#endif
}
END_DUKE_NS
